Gamma.AimListContains = function(tb, mobj)
	if (type(tb) != "table")
		return false;
	end

	for _,v in pairs(tb) do
		if (v != nil)
		and (v.mobj and v.mobj.valid)
		and (v.mobj == mobj)
			return true;
		end
	end

	return false;
end

Gamma.ComboAlreadyTargeting = function(combo, mobj)
	if (type(combo) != "table")
		return false;
	end

	if (Gamma.AimListContains(combo.shotList, mobj) == true)
	or (Gamma.AimListContains(combo.aimList, mobj) == true)
		return true;
	end

	return false;
end

Gamma.AlreadyTargeting = function(comboList, mobj)
	if (type(comboList) != "table")
		return false;
	end

	for _,v in pairs(comboList)
		if (Gamma.ComboAlreadyTargeting(v, mobj) == true)
			return true;
		end
	end

	return false;
end

Gamma.Targetable = function(mo, aimList, owner, checkUnidus)
	if (mo.health <= 0)
	or (mo.flags2 & MF2_FRET)
		return false;
	end

	local hookVal = Gamma.GetTargetValue(mo, aimList, owner);
	if (hookVal != nil)
		return hookVal;
	end

	if (checkUnidus == true) and (mo.unidusList != nil)
		// Only aim at when all of its projectiles are targetted first
		for _,ball in pairs(mo.unidusList)
			if not (ball and ball.valid)
				continue;
			end

			if (Gamma.Targetable(ball, aimList, owner, false) != true)
				continue;
			end

			if (Gamma.AimListContains(aimList, ball) == false)
				return false;
			end
		end
	end

	local canTarget = false;

	if (mo.flags & (MF_ENEMY|MF_BOSS|MF_MONITOR|MF_MISSILE|MF_PUSHABLE))
		canTarget = true;
	end

	if (mo.flags2 & MF2_INVERTAIMABLE)
		canTarget = (not canTarget);
	end

	return canTarget;
end

Gamma.FocusTarget = function(player)
	if (player.mo.gammaVars.focusTime > 0)
	and (player.mo.gammaVars.focusTarget and player.mo.gammaVars.focusTarget.valid)
		return player.mo.gammaVars.focusTarget;
	end

	return nil;
end

Gamma.GetLaserPos = function(player)
	local laserPos = Gamma.GetJointCoordsOfPart(player, Gamma.AnimPartIDs.LASER, 2);
	if (laserPos == nil)
		// fallback?
		laserPos = {
			x = player.mo.x,
			y = player.mo.y,
			z = player.mo.z + (player.mo.height/2),
			angle = player.mo.angle
		};
	end
	return laserPos;
end

Gamma.BeamAim = function(player, comboID)
	if (comboID == nil)
		return true;
	end

	local combo = player.mo.gammaVars.comboList[ comboID ];
	if (combo == nil)
		return true;
	end

	if (combo.expireTime >= Gamma.ComboExpire)
		// Lose the combo, we don't finish it
		Gamma.ComboDropSound(player);
		return true;
	end

	combo.expireTime = $1 + 1;
	player.mo.gammaVars.aiming = true;

	// Fade the sound out near the end of the timer
	local volume = 200;

	local volumeMul = FixedDiv(
		Gamma.ComboExpire - combo.expireTime,
		Gamma.ComboExpire - Gamma.ComboExpireWarning
	);

	if (volumeMul < FRACUNIT/2)
		volumeMul = FRACUNIT/2;
	end

	if (volumeMul < FRACUNIT)
		volume = FixedMul($1, volumeMul);
	end

	Gamma.PlaySoundOnPart(player, Gamma.AnimPartIDs.LASER, sfx_gmlasr, volume, true);

	local oldAngle = player.mo.gammaVars.prevLaserAngle;
	local newAngle = player.mo.gammaVars.laserAngle;
	local angleRange = (newAngle - oldAngle);
	local laserAngle = oldAngle + (angleRange / 2);

	local oldPitch = player.mo.gammaVars.prevLaserPitch;
	local newPitch = player.mo.gammaVars.laserPitch;
	local pitchRange = (newPitch - oldPitch);
	local laserPitch = oldPitch + (pitchRange / 2);

	local laserPos = Gamma.GetLaserPos(player);
	local focus = Gamma.FocusTarget(player);

	if (focus and focus.valid)
		player.mo.gammaVars.focusTime = $1 - 1;
	else
		// Search everything between your current angle angle and your old aim angle!
		local laserDist = FixedMul(player.mo.scale, Gamma.LASERDIST);
		local searchRadius = laserDist / 2;

		// At this close range, angle doesn't matter, just always target.
		local leniency = player.mo.scale * 128;

		local searchCenter = {
			x = player.mo.x + FixedMul(cos(laserAngle), searchRadius),
			y = player.mo.y + FixedMul(sin(laserAngle), searchRadius),
		};

		local possibleTargets = {};

		local function laserTarget(_, mobj)
			if (Gamma.AlreadyTargeting(player.mo.gammaVars.comboList, mobj) == true)
				return nil;
			end

			if (Gamma.Targetable(mobj, combo.aimList, player.mo, true) == false)
				return nil;
			end

			local pHalfHeight = player.mo.height / 2;
			local eHalfHeight = mobj.height / 2;

			local xydist = R_PointToDist2(mobj.x, mobj.y, player.mo.x, player.mo.y);
			local zdist = abs((mobj.z + eHalfHeight) - (player.mo.z + pHalfHeight));

			xydist = max($1 - mobj.radius - player.mo.radius, 0);
			zdist = max($1 - eHalfHeight - pHalfHeight, 0);

			local dist = R_PointToDist2(0, 0, xydist, zdist);
			if (dist > laserDist)
				// Too far away
				return nil;
			end

			local angleDiff = 0;
			local pitchDiff = 0;
			local delta = 0;

			if (dist > 0)
				// If the enemy is literally touching Gamma,
				// then don't do this, so they are allowed
				// to aim at them immediately.

				local hAngle = R_PointToAngle2(laserPos.x, laserPos.y, mobj.x, mobj.y);
				angleDiff = abs(hAngle - laserAngle);

				local lxydist = P_AproxDistance(mobj.x - laserPos.x, mobj.y - laserPos.y);
				local lzdist = (mobj.z + eHalfHeight) - laserPos.z;
				local vAngle = R_PointToAngle2(0, 0, lxydist, -lzdist);
				pitchDiff = abs(vAngle - laserPitch);

				delta = R_PointToDist2(0, 0, angleDiff, pitchDiff);
			end

			if (dist - leniency > 0)
				// Only do precise angle checks if the object
				// is not too close to you.

				local coneH = Gamma.LASERCONE + (abs(angleRange) / 2);
				if (angleDiff > coneH)
					return nil;
				end

				local coneV = Gamma.LASERCONE + (abs(pitchRange) / 2);
				if (pitchDiff > coneV)
					return nil;
				end

				if not (P_CheckSight(player.mo, mobj))
					// Can't see it.
					// Do last because sight checks can be intensive.
					return nil;
				end
			else
				if (angleDiff > ANGLE_90)
					// Ignore close-up leniency targets if they're
					// literally behind.
					return nil;
				end
			end

			// Can target this one.
			// Insert into a table of possible targets.
			local newTarget = {};

			newTarget.mobj = mobj;
			newTarget.dist = dist;
			newTarget.delta = delta;

			table.insert(possibleTargets, newTarget);
			return nil;
		end

		// Just to be safe, add a little bit past
		// the block that we would search.
		local MAXRADIUS = 64*FRACUNIT;

		searchBlockmap("objects", laserTarget, player.mo,
			searchCenter.x - searchRadius - MAXRADIUS, searchCenter.x + searchRadius + MAXRADIUS,
			searchCenter.y - searchRadius - MAXRADIUS, searchCenter.y + searchRadius + MAXRADIUS
		);

		local numTargets = #possibleTargets;
		if (numTargets > 0)
			// Sort the targets.
			table.sort(possibleTargets, function(a, b)
				local scoreA = (AngleFixed(a.delta) * 16) + a.dist;
				local scoreB = (AngleFixed(b.delta) * 16) + b.dist;
				return (scoreA < scoreB);
			end);

			for try = 1,numTargets do
				local setTarget = possibleTargets[try].mobj;
				if not (setTarget and setTarget.valid)
					continue;
				end

				// Successful target.
				S_StartSound(player.mo, sfx_gmtarg);

				local pts = Gamma.ComboPoints(setTarget);
				local order = 0;
				if (pts > 0)
					combo.numWithPoints = $1 + 1;
					order = combo.numWithPoints;
				end

				local newAimListEntry = {
					mobj = setTarget,
					points = pts,
					order = order,
				};

				table.insert(combo.aimList, newAimListEntry);

				player.mo.gammaVars.focusTarget = setTarget;
				player.mo.gammaVars.focusTime = 4;

				// Spawn the wireframes.
				Gamma.CreateTargetWireframe(setTarget, player, combo, comboID);

				// Got one, let's keep moving.
				break;
			end
		end

		// Did we get a new target?
		focus = Gamma.FocusTarget(player);
		if (focus and focus.valid)
			Gamma.UpdateLaserAngle(player);
		end
	end

	laserPos = Gamma.GetLaserPos(player);
	Gamma.DrawLaser(
		player, combo,
		laserPos,
		player.mo.gammaVars.laserAngle,
		player.mo.gammaVars.laserPitch
	);

	return false;
end

Gamma.UpdateAimAngle = function(player)
	/*
	local focus = Gamma.FocusTarget(player);
	if (focus and focus.valid)
		-- Focused on what we just targeted.
		player.mo.gammaVars.aimAngle = R_PointToAngle2(
			player.mo.x, player.mo.y,
			focus.x, focus.y
		);
		return;
	end
	*/

	if (P_PlayerInPain(player) == true)
		-- Reset to display angle
		player.mo.gammaVars.aimAngle = player.drawangle;
		return;
	end

	if not (Gamma.AllowAbilities(player, PF_STASIS, true))
		-- Keep old angle
		return;
	end

	local forwardMove = player.cmd.forwardmove;
	local sideMove = player.cmd.sidemove;

	if (player.pflags & PF_SLIDING)
		forwardMove = 50;
	end

	if (player.powers[pw_carry] == CR_MINECART)
		sideMove = player.mo.gammaVars.minecartSideMove;
	end

	local controlDir = player.mo.angle;
	local magnitude = 0;
	local twod = false;

	if (player.mo.currentaxis)
		-- Axis2D support
		magnitude = abs(sideMove);

		if not (sideMove)
			return;
		end

		if (player.mo.controlflip)
			sideMove = -sideMove;
		end

		local axisAngle = 0;
		if (player.mo.currentaxis.angle != nil)
			axisAngle = player.mo.currentaxis.angle - ANGLE_90;
		else
			axisAngle = R_PointToAngle2(
				player.mo.currentaxis.x, player.mo.currentaxis.y,
				player.mo.x, player.mo.y
			);
		end

		controlDir = axisAngle + R_PointToAngle2(0, 0, 0, sideMove * FRACUNIT);

		if (player.mo.currentaxis.flipped)
			controlDir = $1 + ANGLE_180;
		end
	elseif (twodlevel) or (player.mo.flags2 & MF2_TWOD)
		magnitude = abs(sideMove);

		if not (sideMove)
			return;
		end

		controlDir = R_PointToAngle2(0, 0, sideMove * FRACUNIT, 0);
	else
		magnitude = FixedHypot(player.cmd.forwardmove, -sideMove)

		if not (player.pflags & PF_ANALOGMODE)
			if not (player.cmd.forwardmove or sideMove)
				return;
			end

			controlDir = $1 + R_PointToAngle2(0, 0, forwardMove * FRACUNIT, -sideMove * FRACUNIT);
		end
	end

	local aimDiff = (controlDir - player.mo.gammaVars.aimAngle);

	if (Gamma.UsingBattleLaser(player) == true)
		local turnSpeed = ANG1 / 2;
		if (abs(aimDiff) <= turnSpeed)
			player.mo.gammaVars.aimAngle = controlDir;
		else
			if (aimDiff < 0)
				player.mo.gammaVars.aimAngle = $1 - turnSpeed;
			else
				player.mo.gammaVars.aimAngle = $1 + turnSpeed;
			end
		end

		return;
	end

	// 4 when moving, 16 when stationary
	// Basically, the speed of the rotation is based on how far the analog stick is tilted.
	// Is this noticeable? Probably not :P
	local factor = 4*FRACUNIT + (((50 - magnitude) * 12*FRACUNIT) / 50);

	player.mo.gammaVars.aimAngle = $1 + FixedDiv(aimDiff, factor);
end

Gamma.UpdateShootAngle = function(player, target)
	local newAngle = player.drawangle;

	if (player.mo.gammaVars.aiming == true)
	or (Gamma.UsingBattleLaser(player) == true)
		newAngle = player.mo.gammaVars.aimAngle;
	end

	if (target and target.valid)
		newAngle = R_PointToAngle2(player.mo.x, player.mo.y, target.x, target.y);
	end

	player.mo.gammaVars.shootAngle = newAngle;
end

Gamma.UpdateLaserAngle = function(player)
	player.mo.gammaVars.prevLaserAngle = player.mo.gammaVars.laserAngle;
	player.mo.gammaVars.prevLaserPitch = player.mo.gammaVars.laserPitch;

	local focus = Gamma.FocusTarget(player);

	if (focus and focus.valid)
	and (player.mo.gammaVars.focusTime > 1) // Give 1 tic where you don't lock-on, but the laser can return to your desired angle
		// Instantly focus in on our target.
		local laserPos = Gamma.GetLaserPos(player);

		player.mo.gammaVars.laserAngle = R_PointToAngle2(
			laserPos.x, laserPos.y,
			focus.x, focus.y
		);

		player.mo.gammaVars.laserPitch = R_PointToAngle2(
			0, laserPos.z,
			R_PointToDist2(
				laserPos.x, laserPos.y,
				focus.x, focus.y
			), focus.z + (focus.height / 2)
		);

		return;
	end

	// Lerp back to neutral.
	local newAngle = player.mo.gammaVars.aimAngle;
	local newPitch = 0; //player.aiming?
	local factor = 4*FRACUNIT;

	local angDiff = (newAngle - player.mo.gammaVars.laserAngle);
	player.mo.gammaVars.laserAngle = $1 + FixedDiv(angDiff, factor);

	local ptcDiff = (newPitch - player.mo.gammaVars.laserPitch);
	player.mo.gammaVars.laserPitch = $1 + FixedDiv(ptcDiff, factor);
end
