//
// All Emeralds:
// Create explosions from the shots,
// like the laser blaster upgrade.
//

Gamma.HaveBlasterUpgrade = function(player)
	if not (player and player.valid)
		// This function doesn't use player,
		// but it MIGHT later.
		return false;
	end

	return All7Emeralds(emeralds);
end

Gamma.CreateBlasterExplosion = function(shot)
	local center = P_SpawnMobjFromMobj(shot, 0, 0, 0, MT_GAMMA_BLASTEREXPLOSION_CENTER);

	// Set scaling
	center.scale = $1 / 2;
	center.scalespeed = $1 * 3;
	P_SetScale(center, center.scale / 8);

	// Set length
	center.fuse = TICRATE;

	// Copy the bullet's properties
	center.target = shot.target;
	center.gammaComboID = shot.gammaComboID;

	// Create visual
	Gamma.CreateBlasterExplosionPolygons(center);
end

Gamma.BlasterExplosionThink = function(center)
	// Update visual
	Gamma.UpdateBlasterExplosionPolygons(center);

	local owner = center.target;
	local combo = nil;

	if (owner and owner.valid and owner.gammaVars != nil)
		combo = owner.gammaVars.comboList[center.gammaComboID];
	end

	if (combo != nil)
		// Don't cash in the combo until the boom is gone.
		combo.cashIn = Gamma.CASHINTIME;
	end

	// Find stuff to destroy.
	local searchRadius = Gamma.RegularPolygonCircumradius(
		mobjinfo[MT_GAMMA_BLASTEREXPLOSION].radius << 1,
		Gamma.BlasterExplosionSides
	);
	searchRadius = FixedMul(searchRadius, center.scale);

	local function blasterExplosionSearch(ref, mobj)
		if (Gamma.ShotTargetAlive(mobj) == false)
			return; // continue
		end

		local xydist = P_AproxDistance(mobj.x - center.x, mobj.y - center.y);
		local zdist = (mobj.z + mobj.height/2) - center.z;
		local fullDist = FixedHypot(xydist, zdist);

		if (abs(zdist) < searchRadius + (mobj.height / 2))
		and (xydist < (searchRadius + mobj.radius))
			// Act exactly like bullets collding.
			center.gammaComboPoints = Gamma.ComboPoints(mobj);
			Gamma.ShotCollide(center, mobj, owner, combo);
		end
	end

	local extra = 32*FRACUNIT;

	searchBlockmap("objects", blasterExplosionSearch, center,
		center.x - searchRadius - extra, center.x + searchRadius + extra,
		center.y - searchRadius - extra, center.y + searchRadius + extra
	);
end

Gamma.BlasterExplosionRemove = function(center)
	// Remove visual
	Gamma.DeleteBlasterExplosionPolygons(center);
end
