Gamma.DriveStats = function(player)
	local skin = skins[player.mo.skin];
	player.height = skin.spinheight;
	player.normalspeed = 6 * skin.normalspeed / 5; // x1.2
end

Gamma.NormalStats = function(player)
	local skin = skins[player.mo.skin];
	player.height = skin.height;
	player.normalspeed = skin.normalspeed;
end

Gamma.SetDriveMode = function(player)
	if not (player.mo.gammaVars.driveMode)
		S_StartSound(player.mo, sfx_gmmoch);
	end

	player.mo.gammaVars.driveTimer = 0;
	player.mo.gammaVars.driveMode = true;
	player.mo.state = S_PLAY_ROLL;

	Gamma.DriveStats(player);
end

Gamma.StopDriveMode = function(player)
	if (player.mo.gammaVars.keepDrive == true)
		-- Don't remove drive mode... yet.
		return;
	end

	player.mo.gammaVars.driveMode = false;
	player.mo.gammaVars.driving = false;

	if (player.mo.state == S_PLAY_ROLL)
		if (player.mo.momz > 0)
			player.mo.state = S_PLAY_SPRING;
		else
			player.mo.state = S_PLAY_FALL;
		end
	end

	Gamma.NormalStats(player);
end

Gamma.DriveModeLogic = function(player)
	local allowAbility = Gamma.AllowAbilities(player, PF_STASIS, false);
	local onGround = P_IsObjectOnGround(player.mo);
	local moving = false;

	if (player.mo.gammaVars.prevOnGround == true)
		onGround = true;
	end

	if (R_PointToDist2(0, 0, player.cmd.sidemove, player.cmd.forwardmove) > 35)
		moving = true;
	end

	if (player.pflags & PF_SPINNING)
	or ((player.pflags & (PF_JUMPED|PF_NOJUMPDAMAGE)) == PF_JUMPED)
		// Replace rolling with drive mode
		Gamma.SetDriveMode(player);
		player.pflags = ($1|PF_NOJUMPDAMAGE) & ~(PF_JUMPED|PF_SPINNING);
	end

	if (player.mo.gammaVars.driveMode == true)
		if (player.powers[pw_carry] != CR_NONE)
		or (player.pflags & PF_SLIDING)
			-- In a "transition state", set normal stats, but keep the drive mode flag.
			Gamma.NormalStats(player);
			player.mo.gammaVars.keepDrive = true;
		else
			if (allowAbility == true)
			and not (player.pflags & PF_STARTJUMP)
				Gamma.DriveStats(player); -- Brings the stats back if you were in a transition state.

				if (moving == true)
					local prevDrive = player.mo.gammaVars.driving;

					// No friction while moving.
					// This allows Gamma to get speed from slopes in this mode!
					player.mo.friction = FRACUNIT;

					// Flag to play sounds / do effects
					player.mo.gammaVars.driving = onGround;

					if (player.mo.gammaVars.driving == true)
						if (prevDrive == false)
							-- Kickstart dust
							Gamma.DriveDust(player, Gamma.AnimPartIDs.LOWERLEG_L, true);
							Gamma.DriveDust(player, Gamma.AnimPartIDs.LOWERLEG_R, true);
						end

						local counter = leveltime;
						if (counter & 1)
							local leg = Gamma.AnimPartIDs.LOWERLEG_L;
							if ((counter / 2) & 1)
								leg = Gamma.AnimPartIDs.LOWERLEG_R;
							end

							Gamma.DriveDust(player, leg, false);
						end
					end
				else
					player.mo.gammaVars.driving = false;
				end

				player.mo.state = S_PLAY_ROLL;
			else
				-- Remove fully.
				Gamma.StopDriveMode(player);
			end

			player.mo.gammaVars.keepDrive = false;
		end

		player.mo.gammaVars.driveTimer = 0;
		return;
	end

	player.mo.gammaVars.keepDrive = false;

	if (allowAbility == false)
		player.mo.gammaVars.driveTimer = 0;
		return;
	end

	if (onGround == true)
	and (moving == true)
	and not (player.pflags & PF_SLIDING)
		local addTime = 3;
		local waitTime = (3*TICRATE) * addTime;

		if (player.powers[pw_sneakers] > 0)
			waitTime = $1 / 3;
		end

		if (player.mo.standingslope and player.mo.standingslope.valid)
			// Increment more quickly when going up/down slopes,
			// so that you can experience slope fun more often,
			// and get up steep slopes in maps better.

			local slopeAngle = abs(player.mo.standingslope.zangle);
			local slopeMul = FixedDiv(slopeAngle, ANGLE_45);

			addTime = FixedMul(
				$1 << FRACBITS,
				(5*FRACUNIT/4) + slopeMul // the extra 1/4 helps fudge the numbers to be closer how I wanted it to work
			);
			addTime = FixedFloor($1) >> FRACBITS; // Floor the result
		end

		player.mo.gammaVars.driveTimer = $1 + addTime;

		// if you've moved for 3 seconds
		// OR if you've been pushed to thok speed
		if (player.mo.gammaVars.driveTimer >= waitTime)
		or (player.speed >= 60*player.mo.scale)
			Gamma.SetDriveMode(player);
		end
	elseif (player.mo.gammaVars.driveTimer > 0)
		local subTime = 2;
		player.mo.gammaVars.driveTimer = $1 - subTime;

		if (player.mo.gammaVars.driveTimer < 0)
			player.mo.gammaVars.driveTimer = 0;
		end
	end
end

local function v3rotate(rotVec, axisVec, angle)
	local ux = FixedMul(axisVec.x, rotVec.x);
	local uy = FixedMul(axisVec.x, rotVec.y);
	local uz = FixedMul(axisVec.x, rotVec.z);
	local vx = FixedMul(axisVec.y, rotVec.x);
	local vy = FixedMul(axisVec.y, rotVec.y);
	local vz = FixedMul(axisVec.y, rotVec.z);
	local wx = FixedMul(axisVec.z, rotVec.x);
	local wy = FixedMul(axisVec.z, rotVec.y);
	local wz = FixedMul(axisVec.z, rotVec.z);
	local sa = sin(angle);
	local ca = cos(angle);
	local ua = ux+vy+wz;
	local ax = FixedMul(axisVec.x,ua);
	local ay = FixedMul(axisVec.y,ua);
	local az = FixedMul(axisVec.z,ua);
	local xs = FixedMul(axisVec.x,axisVec.x);
	local ys = FixedMul(axisVec.y,axisVec.y);
	local zs = FixedMul(axisVec.z,axisVec.z);
	local bx = FixedMul(rotVec.x,ys+zs);
	local by = FixedMul(rotVec.y,xs+zs);
	local bz = FixedMul(rotVec.z,xs+ys);
	local cx = FixedMul(axisVec.x,vy+wz);
	local cy = FixedMul(axisVec.y,ux+wz);
	local cz = FixedMul(axisVec.z,ux+vy);
	local dx = FixedMul(bx-cx, ca);
	local dy = FixedMul(by-cy, ca);
	local dz = FixedMul(bz-cz, ca);
	local ex = FixedMul(vz-wy, sa);
	local ey = FixedMul(wx-uz, sa);
	local ez = FixedMul(uy-vx, sa);

	return {
		x = ax+dx+ex,
		y = ay+dy+ey,
		z = az+dz+ez,
	};
end

addHook("MobjMoveBlocked", function(mo, thing, line)
	if (mo.skin == "gamma")
	and (mo.gammaVars != nil)
	and (mo.gammaVars.driving == true)
		if (thing and thing.valid and (thing.flags & MF_PUSHABLE))
			// Not if we want to push a gargoyle.
			return;
		end

		if (R_PointToDist2(0, 0, mo.momx, mo.momy) < mo.scale)
			// Not if we have very low speed.
			return;
		end

		local transferPossible = false;
		if (mo.standingslope and mo.standingslope.valid)
			local slope = mo.standingslope;
			local steepEnough = (abs(slope.zangle) >= ANGLE_45);

			if (steepEnough == true and (slope.flags & SL_NOPHYSICS) == 0)
				local momentum = {
					x = mo.momx,
					y = mo.momy,
					z = 0,
				};
				local axis = {
					x = -slope.d.y,
					y = slope.d.x,
					z = 0,
				};
				momentum = v3rotate(momentum, axis, slope.zangle);
				transferPossible = (momentum.z * P_MobjFlip(mo) > 0);
			end
		end

		if (transferPossible == true)
			// If we're trying to wall-transfer,
			// then don't bump against it!
			return;
		end

		P_BounceMove(mo);

		local factor = 4;
		local minSpeed = 5 * mo.scale;
		local newSpeed = R_PointToDist2(0, 0, mo.momx, mo.momy) / factor;
		if (newSpeed < minSpeed)
			// Bump is too small, force a certain thrust
			local a = R_PointToAngle2(0, 0, mo.momx, mo.momy);
			mo.momx = FixedMul(minSpeed, cos(a));
			mo.momy = FixedMul(minSpeed, sin(a));
		else
			mo.momx = $1 / factor;
			mo.momy = $1 / factor;
		end

		S_StartSound(mo, sfx_gmdrvi);
		S_StopSoundByID(mo, sfx_gmdrvs);
		S_StopSoundByID(mo, sfx_gmdrvl);
		mo.gammaVars.driveSound = false;
	end
end, MT_PLAYER);
