Gamma.HoverSpeed = function(player)
	return abs(P_GetMobjGravity(player.mo)) * 8;
end

Gamma.GravityDir = function(player)
	local g = P_GetMobjGravity(player.mo);

	if (g > 0)
		return 1;
	elseif (g < 0)
		return -1;
	end

	return 0;
end

Gamma.JetBoosterLogic = function(player)
	local DJUMPTIME = TICRATE/2;

	local jetMobj = Gamma.GetPartMobj(player, Gamma.AnimPartIDs.JET);

	local jetSpeed = player.mo.scale / 8;
	local hoverSpeed = Gamma.HoverSpeed(player);
	local flip = P_MobjFlip(player.mo);

	if (player.mo.gammaVars.jetDJTime > 0)
		local thrust = player.mo.gammaVars.jetDJTime * jetSpeed * flip;
		player.mo.momz = $1 + thrust;
		player.mo.gammaVars.jetDJTime = $1 - 1;

		if (player.mo.gammaVars.jetDJTime > 0)
			// reduces acceleration
			local accelMul = (DJUMPTIME - player.mo.gammaVars.jetDJTime) * (FRACUNIT / DJUMPTIME);
			player.mo.movefactor = FixedMul($1, accelMul);
		end

		Gamma.SpawnJetSmoke(player, jetMobj, FixedMul(abs(thrust), player.mo.scale) * 8)
	end

	if (player.mo.gammaVars.jetHovering == true)
		local STARTUPLENGTH = TICRATE/3

		if (jetMobj and jetMobj.valid)
			if not (player.mo.gammaVars.jetOverlay and player.mo.gammaVars.jetOverlay.valid)
				player.mo.gammaVars.jetOverlay = P_SpawnMobjFromMobj(jetMobj, 0, 0, 0, MT_GAMMA_JETOVERLAY)
			end

			local startupTime = max(0, STARTUPLENGTH - player.mo.gammaVars.jetStart)
			local startupScale = FRACUNIT + ((startupTime * startupTime) * (FRACUNIT >> 6))

			local dJumpScale = (player.mo.gammaVars.jetDJTime * player.mo.gammaVars.jetDJTime) * (FRACUNIT >> 5)

			player.mo.gammaVars.jetOverlay.spriteyscale = startupScale + dJumpScale;
			player.mo.gammaVars.jetOverlay.spritexscale = startupScale + (dJumpScale >> 3);
		end

		if (player.mo.gammaVars.jetStart < STARTUPLENGTH)
			player.mo.gammaVars.jetStart = $1 + 1
		end

		if (leveltime & 1)
			Gamma.SpawnJetSmoke(player, jetMobj, abs(hoverSpeed))
		end

		if (player.mo.state != S_PLAY_FLOAT)
		and (player.mo.gammaVars.driveMode != true)
			player.mo.state = S_PLAY_FLOAT
		end

		if (player.mo.gammaVars.jetDJTime > 0)
			return
		end

		if (player.cmd.buttons & BT_JUMP)
			if (player.mo.momz * flip < -hoverSpeed)
				player.mo.momz = -hoverSpeed * flip
			end
		else
			Gamma.StopHover(player)
		end
	elseif (player.cmd.buttons & BT_JUMP)
	and (player.mo.gammaVars.prevJump == false)
		if not (player.pflags & PF_THOKKED)
			player.mo.gammaVars.jetDJTime = DJUMPTIME
			player.pflags = $1|PF_THOKKED

			player.mo.momx = $1 / 2
			player.mo.momy = $1 / 2

			local jumpZ = FixedMul(player.jumpfactor, FixedMul(player.mo.scale, 39*(FRACUNIT/4))); // Default jump momentum.
			player.mo.momz = $1 - (flip * jumpZ); // Don't give bonus for having normal jump momentum
			if (player.mo.momz * flip < 0)
				player.mo.momz = 0; // Don't nerf when falling
			end
		end

		Gamma.PlaySoundOnPart(player, Gamma.AnimPartIDs.JET, sfx_gmhovr)
		player.mo.gammaVars.jetHovering = true
	end
end

Gamma.StopHover = function(player)
	if (player.mo.gammaVars.jetHovering == true)
	and not (P_PlayerInPain(player))
		player.mo.state = S_PLAY_FALL;
	end

	player.mo.gammaVars.jetHovering = false;
	player.mo.gammaVars.jetStart = 0;

	if (player.mo.gammaVars.jetOverlay and player.mo.gammaVars.jetOverlay.valid)
		P_RemoveMobj(player.mo.gammaVars.jetOverlay);
	end

	Gamma.PlaySoundOnPart(player, Gamma.AnimPartIDs.JET, nil);
end

Gamma.RestoreDoubleJump = function(player)
	player.pflags = $1 & ~PF_THOKKED;
	player.mo.gammaVars.jetDJTime = 0;
end

Gamma.LiquidFOF = function(rover)
	if not (rover.flags & FF_EXISTS)
		return false;
	end

	if (rover.flags & FF_SWIMMABLE)
	or (rover.flags & FF_QUICKSAND)
	or (rover.master.frontsector.damagetype != SD_NONE)
		return true;
	end

	return false;
end

Gamma.SolidFOF = function(rover, flip)
	if not (rover.flags & FF_EXISTS)
		return false;
	end

	if (flip == -1)
		if (rover.flags & FF_PLATFORM)
			return false;
		end
	else
		if (rover.flags & FF_REVERSEPLATFORM)
			return false;
		end
	end

	if (rover.flags & FF_BLOCKPLAYER)
		return true;
	end

	return false;
end

Gamma.LiquidSector = function(sector)
	if (sector.damagetype != SD_NONE)
		return true;
	end

	return false;
end

Gamma.TryPropeller = function(player, eekImWet, nearSurface, gravityDir)
	local ourSpeed = player.mo.momz * gravityDir;

	if (player.pflags & PF_STARTJUMP)
	and (ourSpeed < 0)
		// Allow an initial jump off of the ground.
		return false;
	end

	if (eekImWet == true)
		// Always allow when touching water.
		return true;
	end

	if (nearSurface == true)
		// Conditions are a little trickier when
		// near water, but not touching it.

		local hoverSpeed = Gamma.HoverSpeed(player);
		local minSpeed = hoverSpeed; // (7 * hoverSpeed) / 8;

		if (ourSpeed < minSpeed)
		or (player.mo.gammaVars.prevJump == false)
			// Allows for an oppritunity to use Jet Booster.
			return false;
		end

		return true;
	end

	return false;
end

Gamma.PropellerLogic = function(player)
	local hoverSpeed = Gamma.HoverSpeed(player);
	local flip = P_MobjFlip(player.mo);
	local gravityDir = Gamma.GravityDir(player);

	local eekImWet = ((player.mo.eflags & (MFE_UNDERWATER|MFE_TOUCHWATER)) != 0);

	local validWater = false;
	local blocked = false;

	local waterSurface = INT32_MIN;
	local waterCap = INT32_MAX;

	local nearSurface = false;
	local insideFOF = false;

	if (player.cmd.buttons & BT_JUMP)
		local height = player.mo.height;
		local waterSpacing = 48 * player.mo.scale;

		local thingBottom = player.mo.z;
		local thingTop = player.mo.z + height;
		local thingMid = player.mo.z + (height >> 1);

		local sec = player.mo.subsector.sector;
		if (sec and sec.valid)
			local liquidSec = Gamma.LiquidSector(sec);
			if (liquidSec == true)
				if (flip == -1)
					waterSurface = sec.ceilingheight;
					waterCap = INT32_MAX;
					validWater = true;
				else
					waterSurface = sec.floorheight;
					waterCap = INT32_MIN;
					validWater = true;
				end
			end

			// Please expose these to lua.
			local GetFFloorTopZAt = function(ffloor, x, y)
				if (ffloor.t_slope != nil)
					return P_GetZAt(ffloor.t_slope, x, y);
				else
					return ffloor.topheight;
				end
			end

			local GetFFloorBottomZAt = function(ffloor, x, y)
				if (ffloor.b_slope != nil)
					return P_GetZAt(ffloor.b_slope, x, y);
				else
					return ffloor.bottomheight;
				end
			end

			for rover in sec.ffloors() do
				local liquid = Gamma.LiquidFOF(rover);
				local solid = Gamma.SolidFOF(rover, flip);

				if (liquid == false) and (solid == false)
					// None of our concern.
					continue;
				end

				local topheight = GetFFloorTopZAt(rover, player.mo.x, player.mo.y);
				local bottomheight = GetFFloorBottomZAt(rover, player.mo.x, player.mo.y);

				// We are looking for any water that we're INSIDE of, or ABOVE.
				if (flip == -1)
					if (topheight < thingBottom)
						continue;
					end

					if (validWater == false)
					or (bottomheight < waterSurface)
						waterSurface = bottomheight;
						waterCap = topheight;
						validWater = true;

						if (liquid == true)
							// We want this water.
							blocked = false;
						elseif (solid == true)
							// Platform is blocking the water.
							blocked = true;
						end
					end
				else
					if (bottomheight > thingTop)
						continue;
					end

					if (validWater == false)
					or (topheight > waterSurface)
						waterSurface = topheight;
						waterCap = bottomheight;
						validWater = true;

						if (liquid == true)
							// We want this water.
							blocked = false;
						elseif (solid == true)
							// Platform is blocking the water.
							blocked = true;
						end
					end
				end
			end
		end

		if (validWater == true)
			if (flip == -1)
				insideFOF = (thingMid >= waterSurface and thingMid <= waterCap);
				waterSurface = $1 - height - waterSpacing;
			else
				insideFOF = (thingMid <= waterSurface and thingMid >= waterCap);
				waterSurface = $1 + waterSpacing;
			end

			if (blocked == false)
				if (flip == -1)
					nearSurface = (player.mo.z >= waterSurface);
				else
					nearSurface = (player.mo.z <= waterSurface);
				end
			end
		end

		if (player.mo.gammaVars.propellerMode == false)
		and (Gamma.TryPropeller(player, eekImWet, nearSurface, gravityDir) == true)
			player.mo.gammaVars.propellerMode = true;
			player.mo.gammaVars.propellerDelay = Gamma.PROPELLER_DELAY;

			if (nearSurface)
			and not (eekImWet or insideFOF)
			and (player.mo.momz * flip <= 0)
				// Falling onto the surface.
				// Snap to it to fix a bug with the propeller delay.
				player.mo.z = waterSurface;
			end
		end
	else
		Gamma.StopPropeller(player);
	end

	if (player.mo.gammaVars.propellerMode == true)
		player.pflags = $1 & ~PF_STARTJUMP;

		Gamma.StopHover(player);
		Gamma.RestoreDoubleJump(player);
		Gamma.SetDriveMode(player);

		local skin = skins[player.mo.skin];
		player.height = skin.spinheight;

		if (player.mo.state != S_PLAY_FLY)
			player.mo.state = S_PLAY_FLY;
		end

		// Counter lowered acceleration in the air
		player.mo.movefactor = $1 << 1;

		local underwater = ((player.mo.eflags & MFE_UNDERWATER) == MFE_UNDERWATER);
		if (underwater or insideFOF)
			if (flip == gravityDir)
				// Go really fast in goop.
				player.mo.momz = $1 + (hoverSpeed * flip);
			else
				// Float up
				local swimSpeed = hoverSpeed * 4;

				if not (underwater)
					swimSpeed = $1 / 3;
				end

				if (player.mo.momz * flip < swimSpeed)
					player.mo.momz = swimSpeed * flip;
				end
			end

			if (underwater)
				Gamma.PropellerBubbles(player);
			end

			player.mo.gammaVars.propellerDelay = 0;
		elseif (nearSurface)
			// Ease above the surface
			player.mo.momz = (waterSurface - player.mo.z) / 16;
			player.mo.gammaVars.propellerDelay = 0;
		elseif (player.mo.momz * flip <= 0)
			if (player.mo.gammaVars.propellerDelay >= Gamma.PROPELLER_DELAY)
				// Slowly float down.
				player.mo.momz = -hoverSpeed * flip / 2;
			else
				// Stay where you are
				player.mo.momz = 0;

				if (player.mo.gammaVars.propellerDelay < Gamma.PROPELLER_DELAY)
					player.mo.gammaVars.propellerDelay = $1 + 1;
				end
			end
		end

		return true;
	end

	return false;
end

Gamma.StopPropeller = function(player)
	if (player.mo.gammaVars.propellerMode == true)
	and not (P_PlayerInPain(player))
		player.mo.state = S_PLAY_FALL;
	end

	player.mo.gammaVars.propellerMode = false;
	player.mo.gammaVars.propellerDelay = 0;

	local skin = skins[player.mo.skin];
	player.height = skin.height;
end

Gamma.AbilityHandler = function(player)
	if (Gamma.BattleAdded() == true)
		if (player.exhaustmeter != nil)
		and (player.exhaustmeter <= 0)
			return;
		end
	end

	local allowAbility = Gamma.AllowAbilities(player, PF_JUMPSTASIS, false);

	local onGround = P_IsObjectOnGround(player.mo)
	if not (onGround)
		onGround = player.mo.gammaVars.prevOnGround
	end

	if (onGround == true)
	or (player.mo.eflags & MFE_SPRUNG)
	or (player.powers[pw_carry] != CR_NONE)
		Gamma.RestoreDoubleJump(player);
	end

	if (onGround == true)
	or (allowAbility == false)
		Gamma.StopHover(player);
		Gamma.StopPropeller(player);
		return;
	end

	if (player.pflags & PF_SLIDING)
		Gamma.StopPropeller(player);
	elseif (Gamma.PropellerLogic(player) == true)
		return;
	end

	Gamma.JetBoosterLogic(player);
end
