Gamma.RemoveFromShotList = function(tb, mobj, bullet)
	if (type(tb) != "table")
		return;
	end

	for k,v in pairs(tb) do
		if (v != nil)
		and (v.mobj and v.mobj.valid)
		and (v.mobj == mobj)
			if (v.bullets != nil)
			and (bullet and bullet.valid)
				for kb,vb in pairs(v.bullets) do
					if (bullet == vb)
						table.remove(v.bullets, kb);
					end
				end
			end

			if (v.bullets == nil)
			or (#(v.bullets) == 0)
				table.remove(tb, k);
			end
		end
	end
end

Gamma.ShotExplode = function(mo)
	if (mo.type != MT_GAMMA_SHOT)
		// Lazy workaround.
		return;
	end

	if (mo.health <= 0)
		// Already used an explosion effect.
		return;
	end

	P_KillMobj(mo);
	//S_StartSound(mo, sfx_gmkill); //sfx_gmshex

	mo.state = mo.info.xdeathstate;

	local newScale = mo.scale * 3;
	P_SetScale(mo, newScale);
	mo.destscale = newScale;

	mo.momx = 0;
	mo.momy = 0;
	mo.momz = 0;

	if (mo.flags2 & MF2_AMBUSH) // has blaster upgrade
		Gamma.CreateBlasterExplosion(mo);
	end
end

Gamma.ShotFizzle = function(mo, fromFuse)
	if (mo.type != MT_GAMMA_SHOT)
		// Lazy workaround.
		return;
	end

	if (mo.health <= 0)
	and (fromFuse != true)
		// Already used an explosion effect.
		return;
	end

	P_KillMobj(mo);
	S_StartSound(mo, sfx_gmshfz);

	local newScale = 2 * mo.scale;
	P_SetScale(mo, newScale);
	mo.destscale = newScale;

	mo.momx = 0;
	mo.momy = 0;
	mo.momz = 0;
end

Gamma.ShotReflect = function(mo, dest)
	if (mo.type != MT_GAMMA_SHOT)
		// Lazy workaround.
		return;
	end

	if (mo.health <= 0)
		// Already used an explosion effect.
		return;
	end

	mo.health = 0;
	mo.flags = $1 | MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT|MF_NOCLIPTHING;
	mo.flags = $1 & ~(MF_SHOOTABLE|MF_FLOAT|MF_SPECIAL|MF_MISSILE);
	Gamma.ShotRemove(mo);

	S_StartSound(mo, sfx_gmshre);

	mo.fuse = TICRATE;

	local hAngle = R_PointToAngle2(mo.momx, mo.momy, 0, 0) + ANGLE_180;
	local vAngle = R_PointToAngle2(R_PointToDist2(0, 0, mo.momx, mo.momy), mo.momz, 0, 0);

	local speed = FixedMul(mo.info.speed, mo.scale) >> 1;

	mo.momx = FixedMul(FixedMul(speed, cos(hAngle)), cos(vAngle));
	mo.momy = FixedMul(FixedMul(speed, sin(hAngle)), cos(vAngle));
	mo.momz = FixedMul(speed, sin(vAngle));
end

Gamma.ShotPush = function(gargoyle, shot)
	local scaling = FixedDiv(shot.scale, gargoyle.scale);
	local thrust = FixedMul(gargoyle.info.speed, scaling);
	local minThrust = 4 * scaling;

	if (thrust < minThrust)
		thrust = minThrust;
	end

	local dir = R_PointToAngle2(
		shot.x - shot.momx,
		shot.y - shot.momy,
		gargoyle.x - gargoyle.momx,
		gargoyle.y - gargoyle.momy
	);

	gargoyle.momx = FixedMul(thrust, cos(dir)); -- + $1
	gargoyle.momy = FixedMul(thrust, sin(dir)); -- + $1

	S_StartSound(gargoyle, gargoyle.info.activesound);
	S_StartSound(gargoyle, sfx_s3k69);
	gargoyle.target = shot; -- Might not be needed?
end

Gamma.ShotCollide = function(mo, dest, owner, combo)
	local oldScoreAdd = owner.player.scoreadd;
	if (combo != nil)
		owner.player.scoreadd = combo.numGotten;
	end

	local result = nil;
	local hookVal = Gamma.GetDamageValue(dest, mo, owner);

	if (hookVal != nil)
		result = hookVal;
	else
		if (dest.flags & MF_PUSHABLE)
		or ((mobjinfo[dest.type].flags & MF_PUSHABLE) and dest.fuse) -- Allows pushing rollout rocks while they're being ridden
			Gamma.ShotPush(dest, mo);
			result = Gamma.SHOT_EXPLODE;
		end

		-- Attempt to cause damage!!
		if (dest.flags & MF_SHOOTABLE)
			-- Standard damage
			P_DamageMobj(dest, mo, owner);
			result = Gamma.SHOT_EXPLODE;
		elseif (dest.flags & MF_MISSILE)
			-- Projectile clashing
			P_ExplodeMissile(dest);
			result = Gamma.SHOT_EXPLODE;
		end
	end

	if (result == Gamma.SHOT_EXPLODE)
		// Successful damage!!

		if (Gamma.ShotTargetAlive(dest) == false)
			// If it died, then we get extra time.
			Gamma.GiveTimeFromMobj(dest, mo.gammaComboPoints, combo, owner.player);

			/*
			if (mo.gammaComboPoints > 0)
				S_StartSound(mo, sfx_gmkill);
			end
			*/
		end

		// Shot caused damage, so it uses the explosion effect.
		Gamma.ShotExplode(mo);
	elseif (result == Gamma.SHOT_ABSORB)
		// Shot did not cause damage due to being absorbed.
		Gamma.ShotFizzle(mo);
	else
		// Shot did not cause damage due to being reflected.
		Gamma.ShotReflect(mo);
	end

	owner.player.scoreadd = oldScoreAdd;
end

Gamma.ShotTargetAlive = function(dest)
	-- This function's mostly just a hack for TNT barrels.
	local destAlive = false;

	if (dest and dest.valid and dest.health > 0)
		destAlive = true;

		if (dest.type == MT_TNTBARREL)
		and not ((dest.state == dest.info.spawnstate) or (dest.state == dest.info.missilestate))
			-- Pretend this is 0 health to remove lock-on.
			destAlive = false;
		end
	end

	return destAlive;
end

Gamma.ShotUpdateSpeed = function(mo, speed)
	local pitch = Gamma.CappedPitch(mo.movedir);
	mo.momx = FixedMul(FixedMul(speed, cos(mo.angle)), cos(pitch));
	mo.momy = FixedMul(FixedMul(speed, sin(mo.angle)), cos(pitch));
	mo.momz = FixedMul(speed, sin(pitch));
end

Gamma.ShotHandleTrail = function(mo)
	local xydist = P_AproxDistance(mo.momx, mo.momy);
	local zdist = mo.momz;
	local speed = FixedHypot(xydist, zdist);

	local hAngle = R_PointToAngle2(0, 0, mo.momx, mo.momy);
	local vAngle = R_PointToAngle2(0, 0, xydist, zdist);

	Gamma.SpawnTrail(mo, speed, hAngle, vAngle, 0, 0, 0);
end

Gamma.ShotThink = function(mo)
	local owner = mo.target;
	local dest = mo.tracer;

	if (mo.fuse > 0)
		Gamma.ShotHandleTrail(mo);
		mo.flags2 = $1 ^^ MF2_DONTDRAW;
		return;
	end

	if (mo.health <= 0)
		return;
	end

	if not (owner and owner.valid and owner.gammaVars != nil)
		// commit sudoku
		Gamma.ShotFizzle(mo);
		return;
	end

	if (owner.player and owner.player.valid)
		if (P_PlayerInPain(owner.player) == true)
			Gamma.ShotFizzle(mo);
			return;
		end
	end

	local combo = owner.gammaVars.comboList[mo.gammaComboID];
	if (Gamma.ShotTargetAlive(dest) == false)
		// Your target died before you got to it?
		// That's OK, we'll still give you credit for locking onto it at least!
		Gamma.GiveTimeFromMobj(dest, mo.gammaComboPoints, combo, owner.player);
		Gamma.ShotFizzle(mo);
		return;
	end

	if (combo != nil)
		// Don't cash in the combo until every shot is gone
		combo.cashIn = Gamma.CASHINTIME;
	end

	local oldSpeed = R_PointToDist2(0, 0, R_PointToDist2(0, 0, mo.momx, mo.momy), mo.momz);
	local hSpeed = oldSpeed / 4; // Fix for INSANELY small hitboxes occasionally catching into infinite loop

	local hHeight = dest.height / 2;
	local destZ = (dest.z + hHeight);

	local xydist = R_PointToDist2(mo.x, mo.y, dest.x, dest.y);
	local zdist = (destZ - mo.z);
	local fullDist = R_PointToDist2(0, 0, xydist, zdist);

	local heightCheck = abs(zdist) - (mo.height + hHeight + hSpeed);
	if (heightCheck <= 0)
		// close distance by the size of the objects
		local radiusCheck = xydist - (mo.radius + dest.radius + hSpeed);

		if (radiusCheck <= 0)
			Gamma.ShotCollide(mo, dest, owner, combo);
			return;
		end
	end

	if (mo.extravalue1 > 0)
		// Keep starting direction for a bit.
		mo.extravalue1 = $1 - 1;
		Gamma.ShotHandleTrail(mo);
		return;
	end

	local hAngle = R_PointToAngle2(mo.x, mo.y, dest.x, dest.y);
	local vAngle = R_PointToAngle2(0, 0, xydist, zdist);

	local speed = FixedMul(mo.info.speed, mo.scale);
	if (speed > fullDist)
		speed = fullDist;
	end

	local slowShot = false;
	if (dest.player and dest.player.valid)
		slowShot = (not (gametyperules & GTR_FRIENDLY));

		if (owner.player and owner.player.valid)
			slowShot = (not Gamma.PlayerIsFriendly(owner.player, dest.player));
		end
	else
		slowShot = ((dest.flags & MF_BOSS) == MF_BOSS);
	end

	local lerp = FRACUNIT;
	local timeout = 10*TICRATE;

	if (slowShot == true)
		// If going after an enemy player,
		// then it needs to slow down to be out-runnable,
		// and to fizzle out only after a few seconds.

		local normalSpeed = 27*FRACUNIT; // Speed used for bosses
		if (dest.player and dest.player.valid)
			normalSpeed = dest.player.normalspeed;
		end

		local maxSpeed = FixedMul(normalSpeed * 3 / 4, dest.scale);
		speed = min($1, maxSpeed);
		lerp = $1 + (fullDist / 256);

		timeout = mo.info.reactiontime;
	else
		lerp = $1 + (fullDist / 128);
	end

	mo.threshold = $1 + 1;
	if (mo.threshold > timeout)
		// Flicker away and lose target.
		local oldX = mo.momx;
		local oldY = mo.momy;
		local oldZ = mo.momz;

		Gamma.ShotReflect(mo);

		mo.momx = oldX;
		mo.momy = oldY;
		mo.momz = oldZ;
		return;
	end

	mo.angle = $1 + FixedDiv(hAngle - mo.angle, lerp);
	mo.movedir = $1 + FixedDiv(vAngle - mo.movedir, lerp);

	Gamma.ShotUpdateSpeed(mo, speed);
	Gamma.ShotHandleTrail(mo);
end

Gamma.ShotRemove = function(mo)
	local owner = mo.target;
	local follow = mo.tracer;

	if (owner and owner.valid and owner.gammaVars != nil)
	and (follow and follow.valid)
		local combo = owner.gammaVars.comboList[mo.gammaComboID];

		if (combo != nil)
			Gamma.RemoveFromShotList(combo.shotList, follow, mo);
		end
	end
end

Gamma.ShotFuse = function(mo)
	Gamma.ShotFizzle(mo, true);
	return true;
end

Gamma.SpawnBullet = function(pmo, comboID, targetData, shotNum, shotOffset, hAngle, vAngle)
	local shot = P_SpawnMobjFromMobj(pmo, shotOffset.x, shotOffset.y, shotOffset.z, MT_GAMMA_SHOT);

	shot.target = pmo;
	shot.color = Gamma.GetShotColor(pmo.player);

	if (Gamma.HaveBlasterUpgrade(pmo.player)) // have blaster upgrade
		shot.flags2 = $1 | MF2_AMBUSH;
	end

	shot.tracer = targetData.mobj;
	shot.gammaComboID = comboID;

	if (shotNum == 0)
		shot.gammaComboPoints = targetData.points;
		S_StartSound(shot, sfx_gmshot);
	else
		shot.gammaComboPoints = 0;
	end

	shot.angle = hAngle;
	shot.movedir = vAngle;

	local speed = FixedMul(shot.info.speed, shot.scale);
	Gamma.ShotUpdateSpeed(shot, speed);

	shot.extravalue1 = 7;

	return shot;
end

Gamma.Shoot = function(pmo, targetList, comboID)
	if not (pmo and pmo.valid)
		return;
	end

	local targetData = nil;
	local index = nil;

	for i,t in pairs(targetList) do
		if (t == nil)
			continue;
		end

		if (Gamma.ShotTargetAlive(t.mobj) == false)
			// Your target died before you got to it?
			// That's OK, we'll still give you credit for locking onto it at least!
			Gamma.GiveTimeFromMobj(t.mobj, t.points, pmo.gammaVars.comboList[comboID], pmo.player);
			targetList[i] = nil; // don't fully remove yet, since we're still iterating
			continue;
		end

		index = i;
		targetData = t;
		break; // Found something to shoot!
	end

	if (targetData == nil)
		// Nothing to shoot.
		return;
	end

	local shotOffset = Gamma.GetJointCoordsOfPart(pmo.player, Gamma.AnimPartIDs.LOWERARM_R, 2);
	if (shotOffset == nil)
		// fallback?
		shotOffset = {
			x = pmo.x,
			y = pmo.y,
			z = pmo.z + FixedDiv(pmo.height / 2, pmo.scale),
			angle = pmo.gammaVars.shootAngle,
		};
	end

	shotOffset.x = $1 - pmo.x;
	shotOffset.y = $1 - pmo.y;
	shotOffset.z = $1 - pmo.z;

	local combo = pmo.gammaVars.comboList[comboID];
	local bullets = {};

	if (combo.powered == true)
	and (targetData.mobj.player and targetData.mobj.player.valid)
		// Battle Mod upgrade
		local NUM_POWER = 15;
		local ANGLE_OFFSET = FixedAngle((360*FRACUNIT) / NUM_POWER);

		for i = 0,NUM_POWER-1 do
			local hOffset = FixedAngle(45 * cos(i * ANGLE_OFFSET));
			local vOffset = FixedAngle(45 * sin(i * ANGLE_OFFSET));

			table.insert(
				bullets,
				Gamma.SpawnBullet(
					pmo, comboID, targetData, i,
					shotOffset, shotOffset.angle + hOffset, vOffset
				)
			);
		end

		combo.powered = false;
	else
		bullets[1] = Gamma.SpawnBullet(
			pmo, comboID, targetData, 0,
			shotOffset, shotOffset.angle, 0
		);
	end

	targetData.bullets = bullets;

	table.insert(
		pmo.gammaVars.comboList[comboID].shotList,
		targetData
	);

	if (index != nil)
		table.remove(targetList, index);
	end

	local blast = P_SpawnMobjFromMobj(pmo, shotOffset.x, shotOffset.y, shotOffset.z, MT_GAMMA_SHOTBLAST);
	blast.angle = shotOffset.angle + ANGLE_90;
end
