local FROM_NIGHTS_TIME = TICRATE >> 1;

Gamma.InNightsMode = function(mobj)
	if not (mobj and mobj.valid)
		return false;
	end

	if (mobj.player and mobj.player.valid)
		return (mobj.player.powers[pw_carry] == CR_NIGHTSMODE);
	end

	// detection for GHOSTS
	return (mobj.state >= S_PLAY_NIGHTS_TRANS1 and mobj.state <= S_PLAY_NIGHTS_ATTACK);
end

Gamma.InPain = function(mobj)
	if not (mobj and mobj.valid)
		return false;
	end

	if (mobj.player and mobj.player.valid)
		return P_PlayerInPain(mobj.player);
	end

	// detection for GHOSTS
	return (mobj.state == S_PLAY_PAIN or mobj.state == S_PLAY_STUN);
end

Gamma.InNightsRecoil = function(mobj)
	if not (mobj and mobj.valid)
		return false;
	end

	if (mobj.player and mobj.player.valid)
		return (mobj.player.powers[pw_carry] == CR_NIGHTSFALL);
	end

	// CANNOT DETECT FOR GHOSTS...
	return false;
end

local function CreateGammaPart(master, typeStr)
	local newPart = {};

	newPart.typeStr = typeStr;

	newPart.mobj = P_SpawnMobjFromMobj(master, 0, 0, 0, MT_GAMMA_PART);
	newPart.mobj.target = master;

	newPart.mobj.skin = "gamma";
	newPart.mobj.state = S_GAMMA_PART;

	newPart.mobj.sprite = SPR_PLAY;
	newPart.mobj.sprite2 = Gamma.AnimPartDefs[typeStr].sprite;

	newPart.mobj.frame = ($1 & ~FF_FRAMEMASK);
	newPart.mobj.flags2 = $1 | MF2_DONTDRAW;

	newPart.connect = nil;
	newPart.connectJoint = nil;

	newPart.angleOffset = 0;

	if (Gamma.AnimPartDefs[typeStr].splat == true)
		newPart.mobj.renderflags = $1 | RF_FLOORSPRITE | RF_NOSPLATBILLBOARD;
	end

	if (Gamma.AnimPartDefs[typeStr].bright != nil)
		newPart.bright = P_SpawnMobjFromMobj(newPart.mobj, 0, 0, 0, MT_OVERLAY);
		newPart.bright.target = newPart.mobj;

		newPart.bright.skin = "gamma";
		newPart.bright.state = S_GAMMA_PART_OVERLAY;

		newPart.bright.sprite = SPR_PLAY;
		newPart.bright.sprite2 = Gamma.AnimPartDefs[typeStr].bright;

		newPart.bright.frame = ($1 & ~FF_FRAMEMASK);
		newPart.bright.renderflags = $1 | (RF_FULLBRIGHT|RF_NOCOLORMAPS);
		newPart.bright.flags2 = $1 | MF2_DONTDRAW;
	end

	return newPart;
end

local function ConnectGammaPart(part, masterPartNum, joint)
	part.connect = masterPartNum
	part.connectJoint = joint
end

local function SetPartLinkDraw(partList, i, parent)
	if (partList[i] == nil)
		return;
	end

	local anchor = parent;
	Gamma.SetLinkDraw(partList[i].mobj, anchor);
	Gamma.SetLinkDraw(partList[i].bright, anchor);
end

local function InitGammaParts(mo, pmo)
	local partList = {}

	// First part is special; this is considered the "center point".
	partList[Gamma.AnimPartIDs.WAIST] = CreateGammaPart(mo, "waist")

	partList[Gamma.AnimPartIDs.PROPELLER] = CreateGammaPart(mo, "propeller")
	ConnectGammaPart(partList[Gamma.AnimPartIDs.PROPELLER], Gamma.AnimPartIDs.WAIST, 1)

	partList[Gamma.AnimPartIDs.TORSO] = CreateGammaPart(mo, "torso")
	ConnectGammaPart(partList[Gamma.AnimPartIDs.TORSO], Gamma.AnimPartIDs.PROPELLER, 2)

	partList[Gamma.AnimPartIDs.LASER] = CreateGammaPart(mo, "laser")
	ConnectGammaPart(partList[Gamma.AnimPartIDs.LASER], Gamma.AnimPartIDs.HEAD, 2)

	partList[Gamma.AnimPartIDs.HEAD] = CreateGammaPart(mo, "head")
	ConnectGammaPart(partList[Gamma.AnimPartIDs.HEAD], Gamma.AnimPartIDs.TORSO, 4)

	partList[Gamma.AnimPartIDs.UPPERLEG_L] = CreateGammaPart(mo, "upperleg")
	ConnectGammaPart(partList[Gamma.AnimPartIDs.UPPERLEG_L], Gamma.AnimPartIDs.WAIST, 2)

	partList[Gamma.AnimPartIDs.UPPERLEG_R] = CreateGammaPart(mo, "upperleg")
	ConnectGammaPart(partList[Gamma.AnimPartIDs.UPPERLEG_R], Gamma.AnimPartIDs.WAIST, 3)

	partList[Gamma.AnimPartIDs.MIDLEG_L] = CreateGammaPart(mo, "midleg")
	ConnectGammaPart(partList[Gamma.AnimPartIDs.MIDLEG_L], Gamma.AnimPartIDs.UPPERLEG_L, 2)

	partList[Gamma.AnimPartIDs.MIDLEG_R] = CreateGammaPart(mo, "midleg")
	ConnectGammaPart(partList[Gamma.AnimPartIDs.MIDLEG_R], Gamma.AnimPartIDs.UPPERLEG_R, 2)

	partList[Gamma.AnimPartIDs.LOWERLEG_L] = CreateGammaPart(mo, "lowerleg")
	ConnectGammaPart(partList[Gamma.AnimPartIDs.LOWERLEG_L], Gamma.AnimPartIDs.MIDLEG_L, 2)

	partList[Gamma.AnimPartIDs.LOWERLEG_R] = CreateGammaPart(mo, "lowerleg")
	ConnectGammaPart(partList[Gamma.AnimPartIDs.LOWERLEG_R], Gamma.AnimPartIDs.MIDLEG_R, 2)

	partList[Gamma.AnimPartIDs.FOOT_L] = CreateGammaPart(mo, "foot")
	ConnectGammaPart(partList[Gamma.AnimPartIDs.FOOT_L], Gamma.AnimPartIDs.LOWERLEG_L, 2)

	partList[Gamma.AnimPartIDs.FOOT_R] = CreateGammaPart(mo, "foot")
	ConnectGammaPart(partList[Gamma.AnimPartIDs.FOOT_R], Gamma.AnimPartIDs.LOWERLEG_R, 2)

	partList[Gamma.AnimPartIDs.UPPERARM_L] = CreateGammaPart(mo, "upperarm")
	ConnectGammaPart(partList[Gamma.AnimPartIDs.UPPERARM_L], Gamma.AnimPartIDs.TORSO, 2)

	partList[Gamma.AnimPartIDs.UPPERARM_R] = CreateGammaPart(mo, "upperarm")
	ConnectGammaPart(partList[Gamma.AnimPartIDs.UPPERARM_R], Gamma.AnimPartIDs.TORSO, 3)

	partList[Gamma.AnimPartIDs.MIDARM_L] = CreateGammaPart(mo, "midarm")
	ConnectGammaPart(partList[Gamma.AnimPartIDs.MIDARM_L], Gamma.AnimPartIDs.UPPERARM_L, 2)

	partList[Gamma.AnimPartIDs.MIDARM_R] = CreateGammaPart(mo, "midarm")
	ConnectGammaPart(partList[Gamma.AnimPartIDs.MIDARM_R], Gamma.AnimPartIDs.UPPERARM_R, 2)

	partList[Gamma.AnimPartIDs.LOWERARM_L] = CreateGammaPart(mo, "handarm")
	ConnectGammaPart(partList[Gamma.AnimPartIDs.LOWERARM_L], Gamma.AnimPartIDs.MIDARM_L, 2)

	partList[Gamma.AnimPartIDs.LOWERARM_R] = CreateGammaPart(mo, "gunarm")
	ConnectGammaPart(partList[Gamma.AnimPartIDs.LOWERARM_R], Gamma.AnimPartIDs.MIDARM_R, 2)

	partList[Gamma.AnimPartIDs.JET] = CreateGammaPart(mo, "jet")
	ConnectGammaPart(partList[Gamma.AnimPartIDs.JET], Gamma.AnimPartIDs.TORSO, 5)

	partList[Gamma.AnimPartIDs.SODA] = CreateGammaPart(mo, "soda")
	ConnectGammaPart(partList[Gamma.AnimPartIDs.SODA], Gamma.AnimPartIDs.LOWERARM_L, 2)

	SetPartLinkDraw(partList, Gamma.AnimPartIDs.WAIST, pmo)
	SetPartLinkDraw(partList, Gamma.AnimPartIDs.TORSO, pmo)
	//SetPartLinkDraw(partList, Gamma.AnimPartIDs.PROPELLER, pmo)
	SetPartLinkDraw(partList, Gamma.AnimPartIDs.HEAD, pmo)
	SetPartLinkDraw(partList, Gamma.AnimPartIDs.LASER, pmo)

	SetPartLinkDraw(partList, Gamma.AnimPartIDs.UPPERARM_L, pmo)
	SetPartLinkDraw(partList, Gamma.AnimPartIDs.MIDARM_L, pmo)
	SetPartLinkDraw(partList, Gamma.AnimPartIDs.LOWERARM_L, pmo)

	SetPartLinkDraw(partList, Gamma.AnimPartIDs.UPPERARM_R, pmo)
	SetPartLinkDraw(partList, Gamma.AnimPartIDs.MIDARM_R, pmo)
	SetPartLinkDraw(partList, Gamma.AnimPartIDs.LOWERARM_R, pmo)

	SetPartLinkDraw(partList, Gamma.AnimPartIDs.UPPERLEG_L, pmo)
	SetPartLinkDraw(partList, Gamma.AnimPartIDs.MIDLEG_L, pmo)
	SetPartLinkDraw(partList, Gamma.AnimPartIDs.LOWERLEG_L, pmo)
	SetPartLinkDraw(partList, Gamma.AnimPartIDs.FOOT_L, pmo)

	SetPartLinkDraw(partList, Gamma.AnimPartIDs.UPPERLEG_R, pmo)
	SetPartLinkDraw(partList, Gamma.AnimPartIDs.MIDLEG_R, pmo)
	SetPartLinkDraw(partList, Gamma.AnimPartIDs.LOWERLEG_R, pmo)
	SetPartLinkDraw(partList, Gamma.AnimPartIDs.FOOT_R, pmo)

	SetPartLinkDraw(partList, Gamma.AnimPartIDs.SODA, pmo)

	if (Gamma.IsCowboyMap(gamemap) == true)
		partList[Gamma.AnimPartIDs.HAT] = CreateGammaPart(mo, "hat")
		ConnectGammaPart(partList[Gamma.AnimPartIDs.HAT], Gamma.AnimPartIDs.HEAD, 3)
		SetPartLinkDraw(partList, Gamma.AnimPartIDs.HAT, pmo)
	end

	return partList
end

local function PartsDelete(master)
	if not (master and master.valid)
		return;
	end

	if (master.gammaAnim == nil)
		return
	end

	local length = #(master.gammaAnim.parts);

	if not (length > 0)
		// Already exploded.
		return;
	end

	local flip = P_MobjFlip(master);

	local part = nil;
	local mo = nil;
	local bright = nil;

	// Remove all parts.
	for i = 1,length
		part = master.gammaAnim.parts[i];
		mo = part.mobj;
		bright = part.bright;

		if (mo and mo.valid)
			P_RemoveMobj(mo);
		end

		if (bright and bright.valid)
			P_RemoveMobj(bright);
		end
	end

	// Empty parts list.
	master.gammaAnim.parts = {};
end

local function PartsExplosion(master, pmo)
	if not ((master and master.valid)
	and (pmo and pmo.valid))
		return;
	end

	if (master.gammaAnim == nil)
		return
	end

	local length = #(master.gammaAnim.parts);

	if not (length > 0)
		// Already exploded.
		return;
	end

	local flip = P_MobjFlip(master);

	local part = nil;
	local mo = nil;
	local bright = nil;

	// Explode parts!
	for i = 1,length
		part = master.gammaAnim.parts[i];
		mo = part.mobj;

		if (mo and mo.valid)
			local xydist = P_AproxDistance(mo.x - pmo.x, mo.y - pmo.y);
			local zdist = mo.z - pmo.z;

			local hAngle = R_PointToAngle2(pmo.x, pmo.y, mo.x, mo.y);
			local vAngle = R_PointToAngle2(0, 0, xydist, zdist);

			if (abs(vAngle) > ANGLE_45)
				// lets not use actual rng now that we have ghosts
				local fakeRng = leveltime - i + pmo.x - pmo.y;
				fakeRng = $1 << 7;
				fakeRng = $1 >> 9;
				fakeRng = $1 << 8;
				hAngle = (fakeRng % 360) * ANG1;
			end

			P_InstaThrust(mo, hAngle, 6*mo.scale);

			local momz = (8*FRACUNIT) + (((vAngle + ANGLE_90) * (8*FRACUNIT)) / ANGLE_90);
			P_SetObjectMomZ(mo, momz, false);

			mo.tics = TICRATE;
			mo.flags = $1 & ~MF_NOGRAVITY;
		end

		bright = part.bright;
		if (bright and bright.valid)
			P_RemoveMobj(bright);
		end
	end

	// Empty parts list.
	master.gammaAnim.parts = {};
end

local function InitGammaAnim(mo, pmo)
	// Remove existing parts when manually resetting an existing animator.
	PartsDelete(mo);

	mo.gammaAnim = {}
	mo.gammaAnim.parts = InitGammaParts(mo, pmo)
	mo.gammaAnim.frame = 0
	mo.gammaAnim.tics = 0
	mo.gammaAnim.holdFor = 0
	mo.gammaAnim.spr2 = SPR2_STND
	mo.gammaAnim.ready = false
end

local function CopyPropsFromMobj(mo, origin)
	mo.flags2 = ($1 & ~(MF2_DONTDRAW|MF2_SHADOW)) | (origin.flags2 & (MF2_DONTDRAW|MF2_SHADOW));
	mo.eflags = ($1 & ~(MFE_VERTICALFLIP)) | (origin.eflags & (MFE_VERTICALFLIP));

	mo.destscale = origin.destscale;
	P_SetScale(mo, origin.scale);

	mo.spritexscale = origin.spritexscale;
	mo.spriteyscale = origin.spriteyscale;

	/* causes glitches?
	mo.renderflags = ($1 & ~(RF_ABSOLUTEOFFSETS|RF_FLIPOFFSETS)) | (origin.eflags & (RF_ABSOLUTEOFFSETS|RF_FLIPOFFSETS));
	mo.spritexoffset = origin.spritexoffset;
	mo.spriteyoffset = origin.spriteyoffset;
	*/

	mo.color = origin.color;
	mo.colorized = origin.colorized;
	mo.blendmode = origin.blendmode;

	mo.frame = ($1 & ~FF_TRANSMASK) | (origin.frame & FF_TRANSMASK);
end

local function GetMobjFeet(mo)
	if (mo.eflags & MFE_VERTICALFLIP)
		return (mo.z + mo.height)
	end

	return mo.z
end

local function RotateCoords(x, y, ang)
	local r = {
		x = FixedMul(x, cos(ang)) - FixedMul(y, sin(ang)),
		y = FixedMul(x, sin(ang)) + FixedMul(y, cos(ang)),
	}

	return r
end

local function GetJointCoords(part, thisJointFrame, connectToPart, jointFrame, hscale, vscale, flip)
	local jointInfo = Gamma.AnimPartDefs[connectToPart.typeStr].frames[jointFrame].joints[part.connectJoint]

	-- jointFixOffset fixes any parts that don't have their connecting joint set as 0,0,0
	local jointFixOffset = Gamma.AnimPartDefs[part.typeStr].frames[thisJointFrame].joints[1]

	-- Rotate the joint coordinates around to match the object's angle
	local rotated = RotateCoords((jointInfo.x * hscale), (jointInfo.y * hscale), connectToPart.mobj.angle)

	local rotFixAngle = part.mobj.angle
	local rotFix = RotateCoords((jointFixOffset.x * hscale), (jointFixOffset.y * hscale), rotFixAngle)

	local jointCoords = {
		x = rotated.x - rotFix.x,
		y = rotated.y - rotFix.y,
		z = ((jointInfo.z * vscale) - (jointFixOffset.z * vscale)) * flip,
	}

	return jointCoords
end

local function GetFrameProperties(animation, frame, numFrames)
	local animFrame = (frame % numFrames) + 1
	return animation.props[animFrame]
end

local function GetSPR2FromState(state)
	local stateSPR2 = ((state.frame & FF_FRAMEMASK) & ~FF_SPR2SUPER)
	local checkSPR2 = (stateSPR2 & ~FF_SPR2SUPER)
	local recursion = 0

	assert((Gamma.AnimationDefs[SPR2_STND] != nil), "No default animation, UpdateGammaParts will not work properly")

	while (Gamma.AnimationDefs[checkSPR2] == nil)
		recursion = $1+1

		-- First do special cases that don't use spr2defaults
		if (checkSPR2 == SPR2_JUMP)
			checkSPR2 = SPR2_SPNG
		else
			-- Use spr2default
			checkSPR2 = (spr2defaults[checkSPR2] & ~FF_SPR2SUPER)
		end

		if (checkSPR2 == SPR2_STND)
			-- Reached default, we're good
			break
		end

		if (recursion >= 32)
			-- Infinite loop, return default
			return SPR2_STND
		end
	end

	return checkSPR2
end

local function CalcGunAimOffset(partDefs, partMobj, curOffset, destAngle)
	local fullAngle = partMobj.angle + curOffset
	local diff = destAngle - fullAngle
	local finalOffset = curOffset + diff

	if (finalOffset < -ANGLE_90)
		finalOffset = -ANGLE_90
	end

	if (finalOffset > ANGLE_45)
		finalOffset = ANGLE_45
	end

	return finalOffset
end

local function UpdateGammaParts(master, pmo)
	if not ((master and master.valid)
	and (pmo and pmo.valid))
		return;
	end

	if (master.gammaAnim == nil)
		return;
	end

	local length = #(master.gammaAnim.parts);

	local player = pmo.player
	local flip = P_MobjFlip(master);
	local part = nil;
	local mo = nil;

	if (Gamma.InNightsMode(pmo) == true)
		-- Remove ourselves if properly init,
		-- so the parts explode off of us.
		PartsExplosion(master, pmo);
		return;
	end

	if not (length > 0)
		-- try to fix this issue asap
		InitGammaAnim(master, pmo)
		length = #(master.gammaAnim.parts)
		return -- don't draw yet.
	else
		-- Check for all valid parts.
		for i = 1,length
			-- Update all parts per animation!
			part = master.gammaAnim.parts[i]
			mo = part.mobj

			if not (mo and mo.valid)
				InitGammaAnim(master, pmo)
				length = #(master.gammaAnim.parts)
				return -- don't draw yet
			end
		end
	end

	if (pmo.sprite != SPR_PLAY)
		-- drawing non-player, just hide all parts
		for i = 1,length do
			-- Update all parts per animation!
			part = master.gammaAnim.parts[i]
			mo = part.mobj

			if not (mo and mo.valid)
				-- bruh?!
				continue
			end

			mo.flags2 = $1|MF2_DONTDRAW;
		end

		return
	end

	local spr2 = SPR2_STND
	if (player and player.valid)
		-- Calculate current animation
		spr2 = GetSPR2FromState(states[pmo.state])
		if (Gamma.UsingBattleLaser(player) == true)
			spr2 = SPR2_FALL;
		end

		-- Write this for ghosts.
		-- We overwrite it to STND later so they don't show up
		pmo.sprite2 = spr2
	else
		if (pmo.ghostLastPos == nil)
			pmo.ghostLastPos = Vector3.New(
				pmo.x,
				pmo.y,
				pmo.z
			);
		end
		pmo.ghostMom = Vector3.New(
			pmo.x - pmo.ghostLastPos.x,
			pmo.y - pmo.ghostLastPos.y,
			pmo.z - pmo.ghostLastPos.z
		);
		pmo.ghostLastPos = Vector3.New(
			pmo.x,
			pmo.y,
			pmo.z
		);

		spr2 = pmo.sprite2
	end

	local animation = Gamma.AnimationDefs[spr2]

	local animNumFrames = #(animation.props)
	local animSpeed = animation.speed

	local playerSpeed = 0;
	if (player and player.valid)
		playerSpeed = FixedDiv(player.speed, FixedMul(pmo.scale, pmo.movefactor));
	else
		playerSpeed = FixedDiv(R_PointToDist2(0, 0, pmo.ghostMom.x, pmo.ghostMom.y), pmo.scale);
	end

	if (spr2 == SPR2_WALK or spr2 == SPR2_RUN)
		if (playerSpeed < 8*FRACUNIT)
			animSpeed = 4
		elseif (playerSpeed < 16*FRACUNIT)
			animSpeed = 3
		elseif (playerSpeed < 32*FRACUNIT)
			animSpeed = 2
		else
			animSpeed = 1
		end
	elseif (spr2 == SPR2_JUMP)
		if (playerSpeed < 16*FRACUNIT)
			animSpeed = 2
		else
			animSpeed = 1
		end
	end

	if (master.gammaAnim.tics > animSpeed)
		-- speeds things up a bit when switching animations
		master.gammaAnim.tics = animSpeed
	end

	local tempProps = GetFrameProperties(animation, master.gammaAnim.frame, animNumFrames);
	if (tempProps.hold == nil)
		-- Fix when switching animations
		master.gammaAnim.holdFor = 0;
	end

	master.gammaAnim.spr2 = spr2

	local drawAngle = pmo.angle;
	if (player and player.valid)
		drawAngle = player.drawangle;
	end

	local aimAngle = drawAngle; -- fallback
	if (pmo.gammaVars != nil)
		aimAngle = pmo.gammaVars.aimAngle
	end

	-- Head uses aiming angle
	part = master.gammaAnim.parts[Gamma.AnimPartIDs.HEAD]
	mo = part.mobj
	mo.angle = aimAngle

	-- If the animation allows for it, set everything to special angles!!
	if (Gamma.InDrinkState(pmo) == true)
		-- Everything uses aimAngle
		part = master.gammaAnim.parts[Gamma.AnimPartIDs.WAIST]
		mo = part.mobj
		mo.angle = aimAngle

		part = master.gammaAnim.parts[Gamma.AnimPartIDs.TORSO]
		mo = part.mobj
		mo.angle = aimAngle
	elseif (animation.noSwivel == true or master.gammaAnim.ready == false)
		-- Everything uses drawangle
		part = master.gammaAnim.parts[Gamma.AnimPartIDs.WAIST]
		mo = part.mobj
		mo.angle = drawAngle

		part = master.gammaAnim.parts[Gamma.AnimPartIDs.TORSO]
		mo = part.mobj
		mo.angle = drawAngle
	else
		-- Waist uses momentum angle
		part = master.gammaAnim.parts[Gamma.AnimPartIDs.WAIST];
		mo = part.mobj;

		if (player and player.valid)
			if (player.powers[pw_carry])
				mo.angle = drawAngle;
			elseif (Gamma.UsingBattleLaser(player) == true)
				mo.angle = pmo.gammaVars.aimAngle;
			elseif (player.rmomx or player.rmomy)
				mo.angle = $1 + ((R_PointToAngle2(0, 0, player.rmomx, player.rmomy) - $1) / 4);
			end
		else
			mo.angle = drawAngle;
		end

		-- Torso uses an average of head and waist
		part = master.gammaAnim.parts[Gamma.AnimPartIDs.TORSO];
		mo = part.mobj;

		--[[
		local anglediff = master.gammaAnim.parts[Gamma.AnimPartIDs.WAIST].mobj.angle - master.gammaAnim.parts[Gamma.AnimPartIDs.HEAD].mobj.angle;

		if (anglediff > ANGLE_90)
			anglediff = ANGLE_180 - $1;
		elseif (anglediff < -ANGLE_90)
			anglediff = -ANGLE_180 - $1;
		end

		mo.angle = master.gammaAnim.parts[Gamma.AnimPartIDs.WAIST].mobj.angle - anglediff;
		--]]
		mo.angle = master.gammaAnim.parts[Gamma.AnimPartIDs.WAIST].mobj.angle;
	end

	-- Hat uses head angle
	part = master.gammaAnim.parts[Gamma.AnimPartIDs.HAT];
	if (part != nil)
		mo = part.mobj;
		mo.angle = master.gammaAnim.parts[Gamma.AnimPartIDs.HEAD].mobj.angle;
	end

	-- First loop: Update sprites.
	for i = 1,length do
		-- Update all parts per animation!
		part = master.gammaAnim.parts[i];
		mo = part.mobj;

		if not (mo and mo.valid)
			-- bruh?!
			continue;
		end

		if (part.typeStr == "gunarm")
		or (part.typeStr == "supergunarm")
			local newType = "gunarm";
			if (Gamma.HaveBlasterUpgrade(player) == true)
				newType = "supergunarm";
			end

			if (part.typeStr != newType)
				-- Switch gun types.
				part.typeStr = newType;
				mo.sprite2 = Gamma.AnimPartDefs[newType].sprite;
			end
		end

		-- Set animation frames
		local destAngOff = 0;
		local destSpr = (mo.frame & FF_FRAMEMASK);
		local changeSprite = false;

		local frameProps = GetFrameProperties(animation, master.gammaAnim.frame, animNumFrames);

		if (frameProps[i] != nil)
			destAngOff = frameProps[i].ang;
			destSpr = frameProps[i].spr;
		end

		-- Default angle to their "parent" part
		if (part.typeStr == "upperleg")
		or (part.typeStr == "midleg")
		or (part.typeStr == "lowerleg")
		or (part.typeStr == "foot")
			mo.angle = master.gammaAnim.parts[Gamma.AnimPartIDs.WAIST].mobj.angle;
		elseif (part.typeStr == "upperarm")
		or (part.typeStr == "midarm")
		or (part.typeStr == "handarm")
		or (part.typeStr == "gunarm")
		or (part.typeStr == "supergunarm")
			mo.angle = master.gammaAnim.parts[Gamma.AnimPartIDs.TORSO].mobj.angle;
		elseif (part.typeStr == "laser")
			mo.angle = master.gammaAnim.parts[Gamma.AnimPartIDs.HEAD].mobj.angle;
			destSpr = (master.gammaAnim.parts[Gamma.AnimPartIDs.HEAD].mobj.frame & FF_FRAMEMASK);
			changeSprite = true;
		end

		if (Gamma.InPain(pmo) == true)
			if (i == Gamma.AnimPartIDs.HEAD)
			or (i == Gamma.AnimPartIDs.TORSO)
			or (i == Gamma.AnimPartIDs.WAIST)
				-- Tilt body back when hurt
				destSpr = $1 - 1;
			end
		end

		if (player and player.valid)
			if (pmo.gammaVars.aiming == true)
			or (player.weapondelay)
				if (i == Gamma.AnimPartIDs.UPPERARM_R)
					destSpr = $1 + 1;
					if (player.weapondelay > 4)
						destSpr = $1 - 1;
					end

					destAngOff = $1 + ANGLE_45;

					changeSprite = true;
				elseif (i == Gamma.AnimPartIDs.MIDARM_R)
					local subtract = 1
					if (player.weapondelay > 4)
						subtract = $1+2
					end
					destSpr = $1 - subtract

					destAngOff = $1 - ANGLE_22h
					destAngOff = CalcGunAimOffset(part, mo, $1, pmo.gammaVars.shootAngle)

					changeSprite = true
				elseif (i == Gamma.AnimPartIDs.LOWERARM_R)
					destSpr = 4
					local subtract = 0
					if (player.weapondelay > 4)
						subtract = $1+3
					end
					destSpr = $1 - subtract

					destAngOff = $1 - ANGLE_45
					destAngOff = CalcGunAimOffset(part, mo, $1, pmo.gammaVars.shootAngle)

					changeSprite = true
				end
			end

			if (pmo.gammaVars.aiming == true)
			and (i == Gamma.AnimPartIDs.LASER)
				if (destSpr == 1)
					// Set aiming pose if in neutral state
					destSpr = 4
					changeSprite = true
				end
			end

			if (pmo.gammaVars.landedTics > 0)
				local bend = 1

				if (pmo.gammaVars.landedTics >= (TICRATE/4))
				and (pmo.gammaVars.landedTics <= (3*TICRATE/4))
					bend = 2
				end

				if (part.typeStr == "upperleg")
				or (part.typeStr == "midleg")
				or (part.typeStr == "upperarm")
				or (part.typeStr == "midarm")
					destSpr = $1 + bend
					changeSprite = true
				elseif (part.typeStr == "lowerleg")
				or (part.typeStr == "handarm")
				or (part.typeStr == "gunarm")
				or (part.typeStr == "supergunarm")
					destSpr = $1 - bend
					changeSprite = true
				end
			end
		end

		local maxSpr = #(Gamma.AnimPartDefs[part.typeStr].frames)

		if (destSpr < 0) then destSpr = 0 end
		if (destSpr > maxSpr) then destSpr = maxSpr end

		if (part.typeStr == "propeller")
			if ((mo.frame & FF_FRAMEMASK) == 0)
				// even an empty frame gets dirty splat pixels on it...
				mo.flags2 = $1 | MF2_DONTDRAW;
			else
				mo.flags2 = $1 & ~MF2_DONTDRAW;

				// Propeller should just forcefully rotate.
				mo.angle = (leveltime % 8) * ANGLE_45;
			end
		else
			local diff = (destAngOff - part.angleOffset);
			local spd = ANGLE_22h / max(1, animSpeed);

			if (abs(diff) <= spd)
				part.angleOffset = destAngOff;
			else
				if (diff < 0)
					part.angleOffset = $1 - spd;
				else
					part.angleOffset = $1 + spd;
				end
			end

			mo.angle = $1 + part.angleOffset;
		end

		if (animation.setInstantly == true)
		or (master.gammaAnim.ready != true)
		or (part.typeStr == "laser")
			mo.frame = ($1 & ~FF_FRAMEMASK) | destSpr;

			if (part.bright and part.bright.valid)
				part.bright.frame = ($1 & ~FF_FRAMEMASK) | destSpr;
			end
		else
			if (master.gammaAnim.tics <= 0)
				changeSprite = true
			end

			if (changeSprite == true)
				local newSpr = (mo.frame & FF_FRAMEMASK);

				if (newSpr < destSpr)
					newSpr = $1 + 1;
				elseif (newSpr > destSpr)
					newSpr = $1 - 1;
				end

				mo.frame = ($1 & ~FF_FRAMEMASK) | newSpr;

				if (part.bright and part.bright.valid)
					part.bright.frame = ($1 & ~FF_FRAMEMASK) | newSpr;
				end
			end
		end
	end

	local lowestZ = GetMobjFeet(master)

	-- Second loop: Now that all of the parts' sprites are updated,
	-- we can update the positioning to make them make sense.
	for i = 1,length
		part = master.gammaAnim.parts[i]
		mo = part.mobj

		if not (mo and mo.valid)
			-- bruh?!
			continue
		end

		CopyPropsFromMobj(mo, pmo)

		if (part.bright and part.bright.valid)
			CopyPropsFromMobj(part.bright, pmo)
		end

		local coords = {
			x = 0,
			y = 0,
			z = 0,
		}

		local connectToPart = nil

		if (part.connect == nil or part.connectJoint == nil)
			coords.x = master.x
			coords.y = master.y
			coords.z = GetMobjFeet(master)
		else
			connectToPart = master.gammaAnim.parts[part.connect]

			if not (connectToPart.mobj and connectToPart.mobj.valid)
				-- bruh?!
				continue
			end

			local thisJointFrame = (part.mobj.frame & FF_FRAMEMASK)
			local jointFrame = (connectToPart.mobj.frame & FF_FRAMEMASK)
			local jointCoords = GetJointCoords(
				part, thisJointFrame,
				connectToPart, jointFrame,
				FixedMul(master.spritexscale, master.scale), FixedMul(master.spriteyscale, master.scale),
				flip
			)

			--[[
			-- Interpolate object positions for smoother slow animations :)
			local interpolate = {
				x = 0,
				y = 0,
				z = 0,
			}

			local interpolateMul = (animSpeed - master.gammaAnim.tics)

			if (animation.setInstantly == false and interpolateMul > 0)
				local thisJointNextFrame = (part.mobj.frame & FF_FRAMEMASK)
				local jointNextFrame = (connectToPart.mobj.frame & FF_FRAMEMASK)

				local nextProps = GetFrameProperties(animation, master.gammaAnim.frame + 1, animNumFrames)

				if (nextProps[i] != nil)
					if (nextProps[i].spr > thisJointNextFrame)
						thisJointNextFrame = $1+1
					elseif (nextProps[i].spr < thisJointNextFrame)
						thisJointNextFrame = $1-1
					end
				end

				if (nextProps[part.connect] != nil)
					if (nextProps[part.connect].spr > jointNextFrame)
						jointNextFrame = $1+1
					elseif (nextProps[part.connect].spr < jointNextFrame)
						jointNextFrame = $1-1
					end
				end

				local nextJointCoords = GetJointCoords(
					part, thisJointNextFrame,
					connectToPart, jointNextFrame,
					FixedMul(master.spritexscale, master.scale), FixedMul(master.spriteyscale, master.scale),
					flip
				)

				print(
					string.format("current frame: %f %f %f", jointCoords.x, jointCoords.y, jointCoords.z),
					string.format("next frame: %f %f %f", nextJointCoords.x, nextJointCoords.y, nextJointCoords.z),
					"animation speed: " + animSpeed,
					"multiplier: " + interpolateMul
				)

				interpolate.x = (nextJointCoords.x - jointCoords.x) / (animSpeed+1)
				interpolate.y = (nextJointCoords.y - jointCoords.y) / (animSpeed+1)
				interpolate.z = (nextJointCoords.z - jointCoords.z) / (animSpeed+1)
			end
			--]]

			coords.x = connectToPart.mobj.x + jointCoords.x -- - (interpolate.x * interpolateMul)
			coords.y = connectToPart.mobj.y + jointCoords.y -- - (interpolate.y * interpolateMul)
			coords.z = connectToPart.mobj.z + jointCoords.z -- - (interpolate.z * interpolateMul)
		end

		if (part.typeStr == "jet")
			-- Jet booster bobs up and down!
			local amp = 8
			local speed = 4*TICRATE
			local pi = 22*FRACUNIT/7
			local sine = amp * sin((2*pi*speed) * leveltime)

			coords.z = $1 + FixedMul(master.spriteyscale, sine)
		end

		if (pmo.colorized == false and part.typeStr == "hat")
			mo.color = Gamma.CowboyHatColor(pmo);
		end

		P_MoveOrigin(mo, coords.x, coords.y, coords.z)

		if (part.bright and part.bright.valid)
			// Fixes level load / ghost shenanigans...
			P_MoveOrigin(part.bright, coords.x, coords.y, coords.z)
		end

		if not ((part.typeStr == "upperarm")
		or (part.typeStr == "handarm")
		or (part.typeStr == "gunarm")
		or (part.typeStr == "supergunarm")
		or (part.typeStr == "jet")
		or (part.typeStr == "propeller"))
			if (flip == -1)
				if (coords.z > lowestZ)
					lowestZ = coords.z
				end
			else
				if (coords.z < lowestZ)
					lowestZ = coords.z
				end
			end
		end
	end

	-- Get the model out of the ground
	local zOffset = 0;
	if (master.eflags & MFE_VERTICALFLIP)
		lowestZ = $1 + FixedMul(mobjinfo[MT_GAMMA_PART].height, FixedMul(master.spriteyscale, master.scale));
	end
	zOffset = (GetMobjFeet(master) - lowestZ) + (animation.zOffset * FixedMul(master.spriteyscale, master.scale) * flip);

	for i = 1,length
		part = master.gammaAnim.parts[i]
		mo = part.mobj

		mo.z = $1 + zOffset

		if ((master.eflags & MFE_VERTICALFLIP) and (mo.renderflags & RF_FLOORSPRITE))
			// not sure why we need this, but we need it
			mo.z = $1 + FixedMul(mobjinfo[MT_GAMMA_PART].height, FixedMul(master.spriteyscale, master.scale));
		end

		if (part.bright and part.bright.valid)
			// Fixes level load / ghost shenanigans...
			part.bright.z = mo.z;
		end

		--[[
		if (part.typeStr == "foot" and P_IsObjectOnGround(pmo))
			-- Reset feet instantly on flat ground, or tilt them acoording to slopes
			local footOnGround = false
			local groundZ = GetMobjFeet(pmo)
			local tiltSpr = 2

			if (pmo.standingslope != nil)
				groundZ = P_GetZAt(pmo.standingslope, mo.x, mo.y)
			end

			if (flip == -1)
				if (mo.z >= groundZ)
					footOnGround = true
				end
			else
				if (mo.z <= groundZ)
					footOnGround = true
				end
			end

			if (footOnGround == true)
				if (pmo.standingslope != nil)
					local zangle = pmo.standingslope.zangle
					local xydir = pmo.standingslope.xydirection

					if (flip == -1)
						zangle = InvAngle(zangle)
					end

					local roll = FixedMul(zangle, cos(mo.angle - xydir));

					if (roll > ANG30)
						tiltSpr = 0;
					elseif (roll > ANG10)
						tiltSpr = 1;
					elseif (roll < -ANG10)
						tiltSpr = 3;
					elseif (roll < -ANG30)
						tiltSpr = 4;
					end
				end

				mo.frame = ($1 & ~FF_FRAMEMASK) | tiltSpr;
			end
		end
		--]]
	end

	-- Update soda easter egg
	part = master.gammaAnim.parts[Gamma.AnimPartIDs.SODA]
	mo = part.mobj
	mo.color = SKINCOLOR_NONE
	mo.angle = master.gammaAnim.parts[Gamma.AnimPartIDs.LOWERARM_L].mobj.angle
	mo.frame = (master.gammaAnim.parts[Gamma.AnimPartIDs.LOWERARM_L].mobj.frame & FF_FRAMEMASK)
	if (Gamma.InDrinkState(pmo) == true)
		mo.flags2 = $1 & ~MF2_DONTDRAW
	else
		mo.flags2 = $1 | MF2_DONTDRAW
	end

	-- NiGHTS mode thing
	if (master.gammaFromNights != nil and master.gammaFromNights > 0)
		local frac = (FRACUNIT * master.gammaFromNights) / FROM_NIGHTS_TIME;
		local dist = ease.insine(frac, 0, 256 * master.scale);

		local centerOffset = 20 * pmo.scale;
		local centerPoint = Vector3.New(pmo.x, pmo.y, pmo.z + (pmo.height / 2) + (flip * centerOffset));

		local showMeTheBird = ((master.gammaFromNights & 1) != 0);

		if not (master.hurtFlicky and master.hurtFlicky.valid)
			master.hurtFlicky = P_SpawnMobjFromMobj(master, 0, 0, 0, MT_THOK);
			master.hurtFlicky.state = S_GAMMA_FLICKY_HURT;
			master.hurtFlicky.angle = pmo.angle;
		end

		CopyPropsFromMobj(master.hurtFlicky, pmo);
		P_MoveOrigin(
			master.hurtFlicky,
			centerPoint.x,
			centerPoint.y,
			centerPoint.z
		);
		master.hurtFlicky.angle = $1 + ANGLE_22h;
		master.hurtFlicky.color = Gamma.BirdColor(pmo);

		if (showMeTheBird)
			master.hurtFlicky.flags2 = $1 & ~MF2_DONTDRAW;
		else
			master.hurtFlicky.flags2 = $1 | MF2_DONTDRAW;
		end

		for i = 1,length do
			if (i == Gamma.AnimPartIDs.SODA)
				continue;
			end

			part = master.gammaAnim.parts[i]
			mo = part.mobj

			local partPoint = Vector3.New(mo.x, mo.y, mo.z);
			local dir = Vector3.DirectionTo(centerPoint, partPoint);

			local newPoint = Vector3.New(
				partPoint.x + FixedMul(dir.x, dist),
				partPoint.y + FixedMul(dir.y, dist),
				partPoint.z + FixedMul(dir.z, dist)
			);
			P_MoveOrigin(mo, newPoint.x, newPoint.y, newPoint.z);

			if (part.bright and part.bright.valid)
				// Fixes level load / ghost shenanigans...
				P_MoveOrigin(part.bright, newPoint.x, newPoint.y, newPoint.z);
			end

			if (showMeTheBird)
				mo.flags2 = $1 | MF2_DONTDRAW;
				if (part.bright and part.bright.valid)
					part.bright.flags2 = $1 | MF2_DONTDRAW;
				end
			else
				mo.flags2 = $1 & ~MF2_DONTDRAW;
				if (part.bright and part.bright.valid)
					part.bright.flags2 = $1 & ~MF2_DONTDRAW;
				end
			end
		end

		if (Gamma.InNightsRecoil(pmo) == false)
			master.gammaFromNights = $1 - 1;
		end
	elseif (master.hurtFlicky and master.hurtFlicky.valid)
		P_RemoveMobj(master.hurtFlicky);

		for i = 1,length do
			if (i == Gamma.AnimPartIDs.SODA)
				continue;
			end

			part = master.gammaAnim.parts[i]
			mo = part.mobj

			mo.flags2 = $1 & ~MF2_DONTDRAW;
			if (part.bright and part.bright.valid)
				part.bright.flags2 = $1 & ~MF2_DONTDRAW;
			end
		end
	end

	-- Update animation properties
	if (master.gammaAnim.tics <= 0)
		master.gammaAnim.tics = animSpeed;

		if (master.gammaAnim.holdFor != 0)
			if (master.gammaAnim.holdFor > 0)
				master.gammaAnim.holdFor = $1 - 1;
			end
		else
			master.gammaAnim.frame = ($1 % animNumFrames) + 1;

			local frameProps = GetFrameProperties(animation, master.gammaAnim.frame, animNumFrames);
			if (frameProps.hold != nil)
				master.gammaAnim.holdFor = frameProps.hold;
			else
				master.gammaAnim.holdFor = 0;
			end

			if (frameProps.sfx != nil)
				S_StartSound(pmo, frameProps.sfx);
			end
		end
	else
		master.gammaAnim.tics = $1 - 1;
	end

	master.gammaAnim.ready = true;
end

Gamma.FollowMobj = function(mobj, master)
	CopyPropsFromMobj(master, mobj)

	master.radius = mobj.radius
	master.height = mobj.height
	P_MoveOrigin(master, mobj.x, mobj.y, mobj.z)

	local nights = Gamma.InNightsMode(mobj);

	if (nights == true)
		if (master.gammaFromNights != FROM_NIGHTS_TIME)
			master.gammaOldColor = mobj.color;
		end
		master.gammaFromNights = FROM_NIGHTS_TIME;
		mobj.color = Gamma.BirdColor(mobj);
	else
		if (master.gammaAnim == nil)
			InitGammaAnim(master, mobj)
		end

		if (master.gammaOldColor != nil)
			mobj.color = master.gammaOldColor;
			master.gammaOldColor = nil;
		end
	end

	UpdateGammaParts(master, mobj);
	return true;
end

addHook("FollowMobj", function(player, mo)
	return Gamma.FollowMobj(player.mo, mo);
end, MT_GAMMA_ANIMATIONMASTER)

addHook("MobjRemoved", function(master)
	if not (master and master.valid)
		-- specifically only happens when server closes
		return;
	end

	if (master.hurtFlicky and master.hurtFlicky.valid)
		P_RemoveMobj(master.hurtFlicky);
	end

	if (master.gammaAnim != nil)
		local pmo = master.tracer;

		if (pmo and pmo.valid)
			if (pmo.health <= 0)
				PartsExplosion(master, pmo);
				// Spawn in the cute pink flicky
				local center = P_SpawnMobjFromMobj(pmo, 0, 0, 32*FRACUNIT, MT_GAMMA_FLICKY_CENTER);
				if (center and center.valid)
				and (center.tracer and center.tracer.valid)
					center.tracer.color = Gamma.BirdColor(pmo);
				end
			else
				// Otherwise, it's a skin change -- just remove the parts
				PartsDelete(master);
			end
		else
			PartsDelete(master);
		end
	end
end, MT_GAMMA_ANIMATIONMASTER)

local function PartThinker(mo)
	if (mo.tics != -1)
		mo.flags2 = $1 ^^ MF2_DONTDRAW;
	end
end

addHook("MobjThinker", PartThinker, MT_GAMMA_PART);

Gamma.GetPartMobj = function(player, partID)
	if not ((player and player.valid)
	and (player.followmobj and player.followmobj.valid))
		return nil
	end

	local master = player.followmobj

	if (master.gammaAnim == nil)
	or (master.gammaAnim.parts[partID] == nil)
		return nil
	end

	return master.gammaAnim.parts[partID].mobj
end

Gamma.GetJointCoordsOfPart = function(player, partID, jointID)
	if not ((player and player.valid)
	and (player.followmobj and player.followmobj.valid))
		return nil
	end

	local master = player.followmobj

	if (master.gammaAnim == nil)
	or (master.gammaAnim.parts[partID] == nil)
		return nil
	end

	local part = master.gammaAnim.parts[partID]
	local mo = part.mobj

	if not (mo and mo.valid)
		return nil
	end

	local flip = P_MobjFlip(mo)
	local hscale = FixedMul(mo.spritexscale, mo.scale)
	local vscale = FixedMul(mo.spriteyscale, mo.scale)

	local jointFrame = (mo.frame & FF_FRAMEMASK)
	local jointInfo = Gamma.AnimPartDefs[part.typeStr].frames[jointFrame].joints[jointID]

	local rotated = RotateCoords((jointInfo.x * hscale), (jointInfo.y * hscale), mo.angle)

	local jointCoords = {
		x = mo.x + rotated.x,
		y = mo.y + rotated.y,
		z = mo.z + (jointInfo.z * vscale * flip),
		angle = mo.angle,
	}

	return jointCoords
end

Gamma.PlaySoundOnPart = function(player, partID, soundID, volume, repeating)
	if not ((player and player.valid)
	and (player.followmobj and player.followmobj.valid))
		return;
	end

	local master = player.followmobj;

	if (master.gammaAnim == nil)
	or (master.gammaAnim.parts[partID] == nil)
		return;
	end

	local part = master.gammaAnim.parts[partID];
	local mo = part.mobj;

	if not (mo and mo.valid)
		return;
	end

	if (soundID == nil)
		S_StopSound(mo);
		return;
	end

	if (repeating == true) and (S_SoundPlaying(mo, soundID))
		return;
	end

	if (volume != nil)
		S_StartSoundAtVolume(mo, soundID, volume);
	else
		S_StartSound(mo, soundID);
	end
end
