//
// Gamma battle support
//

freeslot(
	"MT_GAMMA_BIGLASER",
	"S_GAMMA_BIGLASER_A",
	"S_GAMMA_BIGLASER_B",
	"SPR_GBLZ",

	-- Battle laser colors
	"SKINCOLOR_GAMMABIGLASER1",
	"SKINCOLOR_GAMMABIGLASER2",
	"SKINCOLOR_GAMMABIGLASER3",
	"SKINCOLOR_GAMMABIGLASER4",
	"SKINCOLOR_GAMMABIGLASER5"
);

mobjinfo[MT_GAMMA_BIGLASER] = {
	spawnstate = S_GAMMA_BIGLASER_A,
	meleestate = S_GAMMA_BIGLASER_B,
	spawnhealth = 1000,
	radius = 32*FRACUNIT,
	height = 32*FRACUNIT,
	flags = MF_NOBLOCKMAP|MF_NOCLIP|MF_NOCLIPHEIGHT|MF_NOCLIPTHING|MF_NOGRAVITY|MF_SCENERY
}

states[S_GAMMA_BIGLASER_A] = {SPR_GBLZ, FF_ANIMATE|FF_RANDOMANIM|FF_ADD|FF_FULLBRIGHT|A, 2, nil, 7, 1, S_NULL}
states[S_GAMMA_BIGLASER_B] = {SPR_GBLZ, FF_ANIMATE|FF_RANDOMANIM|FF_ADD|FF_FULLBRIGHT|I, 2, nil, 7, 1, S_NULL}

skincolors[SKINCOLOR_GAMMABIGLASER1] = {
	name = "Gamma Big Laser 1",
	ramp = {73,73,73,55,36,255,255,255,135,149,0,0,0,0,0,0},
	accessible = false
}

skincolors[SKINCOLOR_GAMMABIGLASER2] = {
	name = "Gamma Big Laser 2",
	ramp = {73,73,55,36,38,255,255,135,149,152,0,0,0,0,0,0},
	accessible = false
}

skincolors[SKINCOLOR_GAMMABIGLASER3] = {
	name = "Gamma Big Laser 3",
	ramp = {73,55,36,38,41,255,135,149,152,154,0,0,0,0,0,0},
	accessible = false
}

skincolors[SKINCOLOR_GAMMABIGLASER4] = {
	name = "Gamma Big Laser 4",
	ramp = {55,36,38,41,41,135,149,152,154,154,0,0,0,0,0,0},
	accessible = false
}

skincolors[SKINCOLOR_GAMMABIGLASER5] = {
	name = "Gamma Big Laser 5",
	ramp = {36,38,41,41,41,135,149,152,154,154,0,0,0,0,0,0},
	accessible = false
}

local BattleHooked = false;

Gamma.BattleAdded = function()
	if (CBW_Battle != nil)
		return BattleHooked;
	end

	return false;
end

local ACTION_NONE = 0;
local ACTION_PAUSE = 1;
local ACTION_STARTUP = 2;
local ACTION_BLAST = 3;
local ACTION_COOLDOWN = 4;
local ACTION_DONE = 5;
local ACTION_POWERCOMBO = 6;

local DOACTION_UNABLE = -1;
local DOACTION_ABLE = 0;
local DOACTION_PRESSED = 1;
local DOACTION_HELD = 2;

local LASER_TIME = TICRATE*2/3;
local MAX_LASER_PARTS = 256;

local COOLDOWN_LASER = 5*TICRATE;
local COOLDOWN_POWER = 5*TICRATE;

local EXHAUST_INIT = FRACUNIT/20;
local EXHAUST_TIME = 5*TICRATE;
local EXHAUST_REDUCE = FRACUNIT / EXHAUST_TIME;

Gamma.UsingBattleLaser = function(player)
	if (Gamma.BattleAdded() == false)
		return false;
	end

	if not (player and player.valid)
		return false;
	end

	return ((player.actionstate > ACTION_NONE)
		and (player.actionstate <= ACTION_DONE));
end

Gamma.ClosetPointOnLine = function(x, y, aX, aY, bX, bY)
	// next breaks this function, YAY
	local dX = bX - aX;
	local dY = bY - aY;
	local d = R_PointToDist2(0, 0, dX, dY);

	local cX = x - aX;
	local cY = y - aY;

	local vX = dX;
	local vY = dY;

	local magnitude = R_PointToDist2(bX, bY, aX, aY);
	vX = FixedDiv($1, magnitude);
	vY = FixedDiv($1, magnitude);

	local t = FixedMul(vX, cX) + FixedMul(vY, cY);

	if (t <= 0)
		return aX, aY;
	elseif (t >= d)
		return bX, bY;
	end

	vX = FixedMul($1, t);
	vY = FixedMul($1, t);

	return aX + vX, aY + vY;
end

Gamma.TryBattleDamage = function(victim, source, inflictor, drain)
	CBW_Battle.PlayerCreditPusher(victim.player, source);

	if (victim.player.powers[pw_shield])
		// Use regular damage.
		P_DamageMobj(victim, inflictor, source, 1);
		return;
	end

	if (victim.player.rings == 0)
		// No rings? Die.
		P_DamageMobj(victim, inflictor, source, DMG_INSTAKILL);
		return;
	end

	// Doesn't work since they get them back instantly...
	//P_PlayerRingBurst(victim.player, 10);

	if (drain == true)
	and (source.player and source.player.valid)
		source.player.rings = $1 + min(victim.player.rings, 10);
	end

	// Knockback stun
	P_DoPlayerPain(victim.player, source, inflictor);

	// Remove some rings.
	victim.player.rings = max(0, $1 - 10);
	P_PlayRinglossSound(victim);

	// Set flashing tics.
	victim.player.powers[pw_flashing] = flashingtics;
end

Gamma.TryLaserAttack = function(laser, aX, aY, bX, bY)
	local owner = laser.target;
	if not (owner and owner.valid)
		// ignore
		return;
	end

	local function DestroyEverything(_, victim)
		if (victim == owner)
			return nil;
		end

		if (Gamma.ShotTargetAlive(victim) == false)
			return nil;
		end

		local vX, vY = Gamma.ClosetPointOnLine(victim.x, victim.y, aX, aY, bX, bY);

		local hHeight = victim.height / 2;
		local vZ = (victim.z + hHeight);

		local xydist = R_PointToDist2(victim.x, victim.y, vX, vY);
		local zdist = (vZ - laser.z);
		local fullDist = R_PointToDist2(0, 0, xydist, zdist);

		local heightCheck = abs(zdist) - (laser.height + hHeight);
		if (heightCheck > 0)
			return nil;
		end

		local radiusCheck = xydist - (laser.radius + victim.radius);
		if (radiusCheck > 0)
			return nil;
		end

		if (victim.player and victim.player.valid)
			// Special behavior against players.
			// Rapidly remove rings.
			if (victim.player.powers[pw_flashing] > 3)
				victim.player.powers[pw_flashing] = 3;
			end

			if (victim.player.powers[pw_flashing] > 0)
				return nil;
			end

			Gamma.TryBattleDamage(victim, owner, laser, true);
			return nil;
		end

		// Reuse damage function from bullets.
		local hookVal = Gamma.GetDamageValue(victim, laser, owner);
		if (hookVal == nil)
			if (victim.flags & MF_PUSHABLE)
			or ((mobjinfo[victim.type].flags & MF_PUSHABLE) and victim.fuse)
				Gamma.ShotPush(victim, laser);
			end

			if (victim.flags & MF_SHOOTABLE)
				P_DamageMobj(victim, laser, owner);
			elseif (victim.flags & MF_MISSILE)
				P_ExplodeMissile(victim);
			end
		end

		return nil;
	end

	local searchDist = laser.radius + (64*FRACUNIT);
	searchBlockmap(
		"objects", DestroyEverything, laser,
		min(aX, bX) - searchDist, max(aX, bX) + searchDist,
		min(aY, bY) - searchDist, max(aY, bY) + searchDist
	);
end

Gamma.InitBattle = function()
	// battlemod is a spammy unfocused disaster that
	// I'm happy to be done supporting in my mods now
	// that all of the official authors have fled due
	// to its toxic as fuck community LMAOOO.
	// (still better than ringslinger, but that's not a high bar)
	if (true)
		return;
	end

	if (CBW_Battle == nil)
		return;
	end

	if (BattleHooked == true)
		return;
	end

	local GammaActions = function(mo, doAction)
		local player = mo.player;

		if not (player and player.valid)
			return;
		end

		if (mo.gammaVars == nil)
			return;
		end

		local tryLaser = (mo.gammaVars.aiming == false);
		if (tryLaser == true)
			player.actiontext = "Power Laser";
			player.actionrings = 40;
		else
			player.actiontext = "Laser Missile";
			player.actionrings = 10;
		end

		if not (CBW_Battle.CanDoAction(player))
			player.actionstate = ACTION_NONE;
			return;
		end

		if (player.actionstate == ACTION_NONE)
			if (tryLaser == true)
				// Massive laser when pressing the button.
				if (doAction == DOACTION_PRESSED)
					player.actionstate = $1 + 1;
					player.actiontime = 0;
					CBW_Battle.PayRings(player);
				end
			else
				// Power up your current combo when pressing the button.
				local combo = nil;
				if (mo.gammaVars.currentCombo != nil)
					combo = mo.gammaVars.comboList[mo.gammaVars.currentCombo];
				end

				if (combo != nil)
					if (combo.powered == false)
					and (combo.fired == false)
					and (doAction == DOACTION_PRESSED)
						combo.powered = true;
						combo.fired = true;
						player.mo.gammaVars.currentCombo = nil;

						player.actionstate = ACTION_POWERCOMBO;
						player.actiontime = 0;
						CBW_Battle.PayRings(player);
					end
				end
			end
		elseif (player.actionstate != ACTION_POWERCOMBO)
			// In laser using state.
			player.actionsuper = true;
			player.weapondelay = 3;
			mo.momx = 0;
			mo.momy = 0;
			mo.momz = 0;
		end

		if (player.actionstate == ACTION_PAUSE)
			player.actiontime = $1 + 1;

			if (player.actiontime >= TICRATE/2)
				player.actionstate = $1 + 1;
				player.actiontime = 0;
			end
		end

		if (player.actionstate == ACTION_STARTUP)
		or (player.actionstate == ACTION_BLAST)
		or (player.actionstate == ACTION_COOLDOWN)
			local laserScale = FRACUNIT;
			local laserTic = LASER_TIME;

			player.actiontime = $1 + 1;

			if (player.actionstate == ACTION_STARTUP)
				laserTic = player.actiontime;
			elseif (player.actionstate == ACTION_COOLDOWN)
				laserTic = (LASER_TIME - player.actiontime) + 1;
			end

			if (laserTic < LASER_TIME)
				laserScale = (laserTic * FRACUNIT) / LASER_TIME;
			end

			laserScale = FixedMul(mo.scale, $1 * 3);

			local point = {
				x = mo.x,
				y = mo.y,
				z = mo.z + (mo.height / 2),
				angle = mo.gammaVars.aimAngle,
			};

			point.c = cos(point.angle);
			point.s = sin(point.angle);

			local r = player.mo.radius + FixedMul(mobjinfo[MT_GAMMA_BIGLASER].radius, laserScale);
			point.x = $1 + FixedMul(r, point.c);
			point.y = $1 + FixedMul(r, point.s);

			local numColors = 5;
			local mod = ((numColors * 2) - 1);
			local offset = numColors - 1;

			local startPoint = {
				x = point.x,
				y = point.y,
			};

			for i = 1,MAX_LASER_PARTS do
				local laser = P_SpawnMobj(point.x, point.y, point.z, MT_GAMMA_BIGLASER);
				laser.target = mo;
				laser.scale = laserScale;
				laser.angle = point.angle;
				laser.color = SKINCOLOR_GAMMABIGLASER1 + abs( ( abs(leveltime - i) % mod ) - offset );

				local overlay = P_SpawnMobjFromMobj(laser, 0, 0, 0, MT_PARTICLE);
				overlay.flags2 = $1 | MF2_LINKDRAW;
				overlay.tracer = laser;
				overlay.state = laser.info.meleestate;
				overlay.scale = ($1 * 6) / 5;
				overlay.color = laser.color;

				point.x = $1 + FixedMul(laser.radius, point.c);
				point.y = $1 + FixedMul(laser.radius, point.s);

				if (Gamma.LaserTestPos(laser, point) == false)
					if (laser and laser.valid)
						Gamma.TryLaserAttack(
							laser,
							startPoint.x, startPoint.y,
							point.x, point.y
						);
					end

					break;
				end
			end
		end

		if ((player.actionstate == ACTION_STARTUP)
		or (player.actionstate == ACTION_COOLDOWN))
		and (player.actiontime > LASER_TIME)
			player.actionstate = $1 + 1;
			player.actiontime = 0;
		end

		if (player.actionstate == ACTION_BLAST)
		and (player.actiontime > LASER_TIME*2)
			player.actionstate = $1 + 1;
			player.actiontime = 0;
		end

		if (player.actionstate == ACTION_DONE)
			CBW_Battle.ApplyCooldown(player, COOLDOWN_LASER);
			player.actionstate = ACTION_NONE;
			player.actiontime = 0;
		end

		if (player.actionstate == ACTION_POWERCOMBO)
			CBW_Battle.ApplyCooldown(player, COOLDOWN_POWER);
			player.actionstate = ACTION_NONE;
			player.actiontime = 0;
		end
	end

	local JetBoosterExhaustion = function(player)
		if (player.mo and player.mo.valid)
		and (player.mo.gammaVars != nil)
		and (player.mo.gammaVars.jetHovering == true
		or player.mo.gammaVars.propellerMode == true)
			if not (player.prevGammaFloat)
				player.exhaustmeter = max(0, $1 - EXHAUST_INIT);
			end
			player.prevGammaFloat = true;

			player.exhaustmeter = max(0, $1 - EXHAUST_REDUCE);

			if (player.exhaustmeter <= 0)
				Gamma.StopHover(player);
				Gamma.StopPropeller(player);
			end
		else
			player.prevGammaFloat = false;
		end

		return false;
	end

	CBW_Battle.SkinVars["gamma"] = {
		flags = SKINVARS_NOSPINSHIELD, //SKINVARS_GUARD
		weight = 150,
		special = GammaActions,
		func_exhaust = JetBoosterExhaustion,
	};
	BattleHooked = true;
end
