assert((FRACBITS == 16), "Update Gamma.tau to match new FRACUNIT.");
Gamma.tau = 411769;
Gamma.pi = Gamma.tau >> 1;

Gamma.BlasterExplosionSides = 8;
Gamma.BlasterExplosionSlices = 6;

Gamma.RegularPolygonInradius = function(length, n)
	return FixedDiv(length, 2 * tan(ANGLE_180 / n));
end

Gamma.RegularPolygonCircumradius = function(length, n)
	return FixedDiv(length, 2 * abs(sin(ANGLE_180 / n)));
end

Gamma.RegularPolygon = function(parts, angle, scale, sideLength, center, zOffset, yScale)
	local numSides = #parts;
	if (numSides == 0)
		return;
	end

	local length = FixedMul(sideLength, scale);
	local inradius = Gamma.RegularPolygonInradius(length, numSides);

	local delta = FixedAngle(360*FRACUNIT / numSides);
	local a = angle;
	for i = 1,numSides do
		local part = parts[i];

		if (part and part.valid)
			local offset = {
				x = center.x,
				y = center.y,
				z = center.z,
			};

			offset.x = $1 + FixedMul(inradius, cos(a));
			offset.y = $1 + FixedMul(inradius, sin(a));
			offset.z = $1 + zOffset;

			P_MoveOrigin(part, offset.x, offset.y, offset.z);
			part.angle = a - ANGLE_90;
			part.scale = scale;
			part.spriteyscale = yScale;
		end

		a = $1 + delta;
	end
end

Gamma.UpdateBlasterExplosionPolygons = function(parent)
	if (parent.gammaExplodePolygons == nil)
		return;
	end

	local numPolygons = #(parent.gammaExplodePolygons);
	if (numPolygons == 0)
		return;
	end

	local sideLength = mobjinfo[MT_GAMMA_BLASTEREXPLOSION].radius << 1;

	parent.angle = $1 + ANG1;

	local center = {
		x = parent.x,
		y = parent.y,
		z = parent.z
	};

	local yScale = FRACUNIT;
	local timeLeft = parent.fuse;
	local shrinkTime = TICRATE/2;
	if (timeLeft < shrinkTime and timeLeft > 0)
		yScale = FixedDiv(max(1, timeLeft), shrinkTime);
	end

	local a = parent.angle;
	for i = 1,numPolygons do
		local data = parent.gammaExplodePolygons[i];
		Gamma.RegularPolygon(
			data.list,
			a,
			FixedMul(parent.scale, data.scale),
			sideLength,
			center,
			FixedMul(data.z, parent.scale),
			yScale
		);

		a = $1 + ANGLE_202h;
	end
end

Gamma.CreateBlasterExplosionPolygons = function(parent)
	local sideLength = mobjinfo[MT_GAMMA_BLASTEREXPLOSION].radius << 1;
	local sideHeight = mobjinfo[MT_GAMMA_BLASTEREXPLOSION].height << 1;

	local data = {};
	local zOffs = 0;
	local sc = FRACUNIT;
	local inrad = Gamma.RegularPolygonInradius(sideLength, Gamma.BlasterExplosionSides);

	for i = 1,Gamma.BlasterExplosionSlices do
		local sliceCnt = 2;

		if (zOffs == 0)
			sliceCnt = 1;
		end

		for j = 1,sliceCnt do
			local newData = {
				list = {},
				z = zOffs,
				scale = sc,
			};

			if not (j & 1)
				newData.z = -newData.z;
			end

			for k = 1,Gamma.BlasterExplosionSides do
				local newPart = P_SpawnMobjFromMobj(
					parent,
					0, 0, 0,
					MT_GAMMA_BLASTEREXPLOSION
				);
				table.insert(newData.list, newPart);
			end

			table.insert(data, newData);
		end

		zOffs = $1 + FixedMul(sideHeight, sc);
		sc = abs(cos(R_PointToAngle2(0, 0, FixedMul(inrad, sc), zOffs)));
		sc = FixedMul(sc, sc);
	end

	parent.gammaExplodePolygons = data;
	Gamma.UpdateBlasterExplosionPolygons(parent);
end

Gamma.DeleteBlasterExplosionPolygons = function(parent)
	if (parent.gammaExplodePolygons == nil)
		return;
	end

	local numPolygons = #(parent.gammaExplodePolygons);
	if (numPolygons == 0)
		return;
	end

	for i = 1,numPolygons do
		local data = parent.gammaExplodePolygons[i];
		local numParts = #(data.list);

		if (numParts == 0)
			continue;
		end

		for j = 1,numParts do
			local part = data.list[j];

			if (part and part.valid)
				P_RemoveMobj(part);
			end
		end
	end
end
