Gamma.JetSmokeThink = function(mo)
	mo.momz = FixedMul($1, mo.friction);
end

Gamma.SpawnJetSmoke = function(player, jetMobj, momentum)
	if not (jetMobj and jetMobj.valid)
		return;
	end

	local h = FixedMul(mobjinfo[MT_GAMMA_JETSMOKE].height, player.mo.scale) / 2;
	local smoke = P_SpawnMobjFromMobj(jetMobj, 0, 0, -h, MT_GAMMA_JETSMOKE);

	if (jetMobj.eflags & MFE_VERTICALFLIP)
		smoke.z = $1 + (8 * jetMobj.scale);
	end

	smoke.destscale = player.mo.destscale;
	P_SetScale(smoke, player.mo.scale);

	if (player.mo.eflags & MFE_VERTICALFLIP)
		smoke.flags2 = $1|MF2_OBJECTFLIP;
	end

	smoke.momx = player.mo.momx;
	smoke.momy = player.mo.momy;
	smoke.momz = player.mo.momz - (momentum * 8 * P_MobjFlip(jetMobj));

	local ang = P_RandomRange(0, 359) * ANG1;
	P_Thrust(smoke, ang, P_RandomRange(1, (momentum * 2 / smoke.scale)) * smoke.scale);
end

Gamma.UpdateJetOverlay = function(player)
	if not (player.mo and player.mo.valid)
		return;
	end

	if (player.mo.gammaVars == nil)
		return;
	end

	local jetOverlay = player.mo.gammaVars.jetOverlay;
	if not (jetOverlay and jetOverlay.valid)
		return;
	end

	if (player.mo.skin != "gamma")
		// we shouldn't have this
		P_RemoveMobj(jetOverlay);
		return;
	end

	local jetMobj = Gamma.GetPartMobj(player, Gamma.AnimPartIDs.JET);
	if not (jetMobj and jetMobj.valid)
		// no jet booster?!
		P_RemoveMobj(jetOverlay);
		return;
	end

	local z = jetMobj.z;
	if (jetMobj.eflags & MFE_VERTICALFLIP)
		jetOverlay.eflags = $1 | MFE_VERTICALFLIP;
		jetOverlay.flags2 = $1 | MF2_OBJECTFLIP;
		z = $1 + (8 * jetMobj.scale);
	else
		jetOverlay.eflags = $1 & ~MFE_VERTICALFLIP;
		jetOverlay.flags2 = $1 & ~MF2_OBJECTFLIP;
	end

	P_MoveOrigin(jetOverlay, jetMobj.x, jetMobj.y, z);

	P_SetScale(jetOverlay, jetMobj.scale);
	jetOverlay.destscale = jetMobj.destscale;
	jetOverlay.scalespeed = jetMobj.scalespeed;

	Gamma.SetLinkDraw(jetOverlay, jetMobj);
end

Gamma.PropellerBubbles = function(player)
	if (leveltime & 1)
		return;
	end

	local propellerMobj = Gamma.GetPartMobj(player, Gamma.AnimPartIDs.PROPELLER);

	local angleOffset = ((leveltime/2) % 8) * ANGLE_45;
	local radius = 32;

	local xOffset = radius * cos(angleOffset);
	local yOffset = radius * sin(angleOffset);
	local zOffset = FixedDiv(propellerMobj.z - player.mo.z, player.mo.scale);

	for i = 1,2 do
		P_SpawnMobjFromMobj(player.mo, xOffset, yOffset, zOffset, MT_SMALLBUBBLE);

		xOffset = -$1;
		yOffset = -$1;
	end
end
