Gamma.GetLaserColor = function(player, time)
	local expireStart = Gamma.ComboExpireWarning;
	local maxColors = 16;

	local colorOffset = expireStart - (Gamma.ComboExpire - time);

	if (colorOffset < 0)
		colorOffset = 0;
	elseif (colorOffset > maxColors)
		colorOffset = maxColors;
	end

	return (SKINCOLOR_GAMMAEXPIRE00 + colorOffset);
end

Gamma.LaserThink = function(mo)
	local transparency = (mo.frame & FF_TRANSMASK) >> FF_TRANSSHIFT;

	transparency = $1 + max(1, (NUMTRANSMAPS - $1) / 2);

	if (transparency >= NUMTRANSMAPS)
		P_RemoveMobj(mo);
		return;
	end

	mo.frame = ($1 & ~FF_TRANSMASK) | (transparency << FF_TRANSSHIFT);

	Gamma.DoPaperRoll(mo);
end

Gamma.LaserTestPos = function(laser, pos)
	local testFlags = (laser.info.flags & (MF_NOCLIP|MF_NOCLIPHEIGHT));
	local oldPos = {
		x = laser.x,
		y = laser.y,
		z = laser.z
	};

	laser.flags = ($1 & ~testFlags) | MF_NOCLIPTHING;

	// P_CheckPosition on its own isn't working for whatever reason.
	local result = (P_TryMove(laser, pos.x, pos.y, true)
		and P_CheckPosition(laser, pos.x, pos.y, pos.z));

	laser.flags = (($1 | testFlags) & ~MF_NOCLIPTHING) | (laser.info.flags & MF_NOCLIPTHING);
	P_MoveOrigin(laser, oldPos.x, oldPos.y, oldPos.z);

	return result;
end

Gamma.DrawLaser = function(player, combo, pos, angle, pitch)
	pitch = Gamma.CappedPitch($1);

	if (player.mo.eflags & MFE_VERTICALFLIP)
		pos.z = $1 + (8 * player.mo.scale);
	end

	// Draw a dot at the start point.
	local dot = P_SpawnMobj(pos.x, pos.y, pos.z, MT_GAMMA_LASER);
	if not (dot and dot.valid) then
		return;
	end

	dot.tracer = player.mo;

	dot.momx = player.mo.momx;
	dot.momy = player.mo.momy;
	dot.momz = player.mo.momz + player.mo.pmomz;

	P_SetScale(dot, player.mo.scale);
	dot.destscale = player.mo.destscale;
	dot.scalespeed = player.mo.scalespeed;

	dot.color = Gamma.GetLaserColor(player, combo.expireTime);

	dot.frame = ($1 & ~FF_PAPERSPRITE);
	if (leveltime & 1)
		dot.frame = $1 + 1;
	else
		dot.frame = $1 + 2 + ((leveltime / 2) & 1);
	end

	if (combo.expireTime >= Gamma.ComboExpire - Gamma.ComboExpireBlink)
	and (leveltime & 1)
		// Flicker the laser itself.
		return;
	end

	local laserDia = FixedMul(player.mo.scale, Gamma.LASERDIAMETER);
	local laserRad = FixedMul(player.mo.scale, Gamma.LASERRADIUS);

	// Draw the laser
	pos.x = $1 + FixedMul(FixedMul(cos(angle), cos(pitch)), laserRad);
	pos.y = $1 + FixedMul(FixedMul(sin(angle), cos(pitch)), laserRad);
	pos.z = $1 + FixedMul(sin(pitch), laserRad);

	for i = 1,Gamma.LASERCOUNT do
		local laser = P_SpawnMobj(pos.x, pos.y, pos.z, MT_GAMMA_LASER);

		if not (laser and laser.valid) then
			break;
		end

		laser.angle = angle;
		laser.gammaTrailAng = pitch;
		laser.tracer = player.mo;

		laser.momx = player.mo.momx;
		laser.momy = player.mo.momy;
		laser.momz = player.mo.momz + player.mo.pmomz;

		P_SetScale(laser, player.mo.scale);
		laser.destscale = player.mo.destscale;
		laser.scalespeed = player.mo.scalespeed;

		laser.color = Gamma.GetLaserColor(player, combo.expireTime);

		local halfway = Gamma.LASERCOUNT / 2;
		local transFlag = (NUMTRANSMAPS * (i - 1 - halfway)) / halfway;
		if (transFlag > tr_trans90)
			transFlag = tr_trans90;
		end
		if (transFlag < 0)
			transFlag = 0;
		end
		transFlag = $1 << FF_TRANSSHIFT;

		laser.frame = ($1 & ~FF_TRANSMASK) | transFlag;

		pos.x = $1 + FixedMul(FixedMul(cos(angle), cos(pitch)), laserDia);
		pos.y = $1 + FixedMul(FixedMul(sin(angle), cos(pitch)), laserDia);
		pos.z = $1 + FixedMul(sin(pitch), laserDia);

		if (Gamma.LaserTestPos(laser, pos) == false)
			break;
		end
	end
end
