// ========
// Gamma hook system
// Allows for custom enemies and bosses to define custom interactions for Gamma
// ========

//
// Damage hook returns
//

Gamma.SHOT_ABSORB = 0;
Gamma.SHOT_EXPLODE = 1;
Gamma.SHOT_REFLECT = 2;

//
// Hook setup
//
local GammaHooks_Target = {};
local GammaHooks_Damage = {};
local GammaHooks_Points = {};

Gamma.SetTargetFunc = function(mobjType, func)
	local maxMobj = #mobjinfo - 1;
	if (mobjType <= 0) or (mobjType > maxMobj)
		error("mobj type "..mobj.type.." out of range (1 - "..maxMobj..")");
		return;
	end

	if (type(func) != "function")
		error("argument 2 not given (expected function)");
		return;
	end

	GammaHooks_Target[mobjType] = func;
end

Gamma.SetDamageFunc = function(mobjType, func)
	local maxMobj = #mobjinfo - 1;
	if (mobjType <= 0) or (mobjType > maxMobj)
		error("mobj type "..mobj.type.." out of range (1 - "..maxMobj..")");
		return;
	end

	if (type(func) != "function")
		error("argument 2 not given (expected function)");
		return;
	end

	GammaHooks_Damage[mobjType] = func;
end

Gamma.SetPointsFunc = function(mobjType, func)
	local maxMobj = #mobjinfo - 1;
	if (mobjType <= 0) or (mobjType > maxMobj)
		error("mobj type "..mobj.type.." out of range (1 - "..maxMobj..")");
		return;
	end

	if (type(func) != "function")
		error("argument 2 not given (expected function)");
		return;
	end

	GammaHooks_Points[mobjType] = func;
end

Gamma.GetTargetValue = function(mobj, aimList, owner)
	if not (mobj and mobj.valid)
		return nil;
	end

	local hookVal = nil;
	local hookFunc = GammaHooks_Target[mobj.type];

	if (hookFunc != nil)
		hookVal = hookFunc(mobj, aimList, owner);
	end

	if (hookVal != nil)
		if (type(hookVal) != "boolean")
			error("Recieved an invalid return value for mobj type "..mobj.type.." (recieved '"..tostring(hookVal).."', expected 'true', 'false', or 'nil'.)");
			hookVal = nil;
		end
	end

	return hookVal;
end

Gamma.GetDamageValue = function(mobj, inflictor, source)
	if not (mobj and mobj.valid)
		return nil;
	end

	local hookVal = nil;
	local hookFunc = GammaHooks_Damage[mobj.type];

	if (hookFunc != nil)
		hookVal = hookFunc(mobj, inflictor, source);
	end

	if (hookVal != nil)
		if (type(hookVal) != "number")
		or (hookVal < Gamma.SHOT_ABSORB)
		or (hookVal > Gamma.SHOT_REFLECT)
			error("Recieved an invalid return value for mobj type "..mobj.type.." (recieved '"..tostring(hookVal).."', expected a number between 0 and 2.)");
			hookVal = nil;
		end
	end

	return hookVal;
end

Gamma.GetPointsValue = function(mobj)
	if not (mobj and mobj.valid)
		return nil;
	end

	local hookVal = nil;
	local hookFunc = GammaHooks_Points[mobj.type];

	if (hookFunc != nil)
		hookVal = hookFunc(mobj);
	end

	if (hookVal != nil)
		if (type(hookVal) != "number") or (hookVal < 0)
			error("Recieved an invalid return value for mobj type "..mobj.type.." (recieved '"..tostring(hookVal).."', expected a number >= 0.)");
			hookVal = nil;
		end
	end

	return hookVal;
end
