local hudIdentifier = "gamma";

customhud.SetupFont("GMSFT", -1, 5);

local function GammaCountdown(v, player)
	if (maptol & TOL_NIGHTS)
	or (G_IsSpecialStage())
		return;
	end

	if not (player.mo and player.mo.valid)
		return;
	end

	local displayTime = player.realtime;
	local pause = 0;
	local countdown = false;

	local levelTimeLeft = nil;

	if (mapheaderinfo[gamemap].countdown > 0)
		// do your best to barely support this instance
		levelTimeLeft = (mapheaderinfo[gamemap].countdown * TICRATE);
		displayTime = max(0, levelTimeLeft - leveltime);
		countdown = true;
	end

	if (player.gammaTimerRan != nil)
		if (player.gammaTime != nil)
			if (countdown == true)
				displayTime = min($1, player.gammaTime);
			else
				displayTime = player.gammaTime;
				pause = player.gammaTimePause;
				countdown = true;
			end
		elseif (countdown == false)
			displayTime = player.gammaTimeAttack;
			pause = player.gammaTimeAttackPause;
		end
	end

	local numbers = {};
	for i = 1, 10, 1 do
		local patchnum = i-1;
		numbers[i] = v.cachePatch("STTNUM\$patchnum\");
	end

	local timePatch = nil;
	if (displayTime < 30*TICRATE) and (countdown == true)
	and (leveltime/5 & 1) then
		timePatch = v.cachePatch("STTRTIME");
	else
		timePatch = v.cachePatch("STTTIME");
	end

	local colonPatch = v.cachePatch("STTCOLON");
	local dotPatch = v.cachePatch("STTPERIO");
	local baseFlags = V_PERPLAYER|V_HUDTRANS;

	v.draw(hudinfo[HUD_TIME].x, hudinfo[HUD_TIME].y, timePatch, hudinfo[HUD_TIME].f | baseFlags);

	if (pause != 0)
	and (pause / 10 & 1)
		// Draw attention to the time limit by flashing it
		return;
	end

	v.draw(hudinfo[HUD_TIMECOLON].x, hudinfo[HUD_TIMECOLON].y, colonPatch, hudinfo[HUD_TIMECOLON].f | baseFlags);
	v.draw(hudinfo[HUD_TIMETICCOLON].x, hudinfo[HUD_TIMETICCOLON].y, dotPatch, hudinfo[HUD_TIMETICCOLON].f | baseFlags);

	local times = {
		G_TicsToMinutes(displayTime),
		G_TicsToSeconds(displayTime),
		G_TicsToCentiseconds(displayTime)
	};

	local numWidth = 8;
	for x = 1,table.maxn(times),1 do
		if (times[x] > 99) then
			times[x] = "99";
		else
			if (times[x] < 10) and (x > 1) then
				times[x] = "0\$times[x]\";
			else
				times[x] = tostring(times[x]);
			end
		end

		local offset = (-2 + (2 - string.len(times[x]))) * numWidth;

		local hudItem = HUD_MINUTES;
		if (x == 2)
			hudItem = HUD_SECONDS;
		elseif (x == 3)
			hudItem = HUD_TICS;
		end

		for i = 1,string.len(times[x]),1 do
			local numPos = (i-1);
			local numPatch = numbers[1 + string.sub(times[x], i, i)];
			v.draw(
				hudinfo[hudItem].x + offset + (numWidth * numPos),
				hudinfo[hudItem].y,
				numPatch,
				hudinfo[hudItem].f | baseFlags
			);
		end
	end
end

customhud.SetupItem("time", hudIdentifier, GammaCountdown);

local function GammaDrawCombo(v, player, combo)
	if (combo == nil)
		return
	end

	if (combo.addTime <= 0)
		return
	end

	local circle = v.cachePatch("GMCIRCL")

	local numbers = {}
	for i = 0, 9, 1 do
		numbers[i] = v.cachePatch("GMNUM\$i\")
	end

	local STARTX = 286 << FRACBITS + (combo.offset << FRACBITS)
	local STARTY = 28 << FRACBITS

	local ENDX = 92 << FRACBITS
	local ENDY = 32 << FRACBITS

	local XDIFF = STARTX - ENDX
	local YDIFF = STARTY - ENDY

	local interp = (combo.cashIn * FRACUNIT) / Gamma.CASHININTERP
	if (interp > FRACUNIT)
		interp = FRACUNIT
	end

	local x = ENDX + FixedMul(XDIFF, interp)
	local y = ENDY + FixedMul(YDIFF, interp)

	local WAUGHMPSCALE = FRACUNIT/3
	local waughmpVal = combo.waughmp

	if (waughmpVal > Gamma.WAUGHMPTIME)
		waughmpVal = Gamma.WAUGHMPTIME - abs(Gamma.WAUGHMPTIME - $1)
	end

	if (waughmpVal < 0)
		waughmpVal = 0
	end

	local waughmp = FRACUNIT + (((waughmpVal * waughmpVal) * WAUGHMPSCALE) / (Gamma.WAUGHMPTIME * Gamma.WAUGHMPTIME))
	local scale = FixedMul(interp, waughmp)

	local flags = V_SNAPTORIGHT|V_SNAPTOTOP|V_PERPLAYER

	v.drawScaled(
		x, y, scale,
		circle,
		flags,
		nil)

	local timeY = y + (3 * scale)

	local timeVal = combo.addTime
	local timeDigits = 0

	while (timeVal > 0)
		timeDigits = $1 + 1
		timeVal = $1 / 10
	end

	local timeX = x + ((timeDigits-1) * 4 * scale)

	timeVal = combo.addTime
	while (timeVal > 0)
		v.drawScaled(
			timeX, timeY, scale,
			numbers[timeVal % 10],
			flags,
			nil)

		timeX = $1 - (9 * scale)
		timeVal = $1 / 10
	end
end

local function GammaDrawCoopCombo(v, player, coopCombo)
	if (coopCombo == nil)
		return
	end

	if (coopCombo.addTime <= 0)
		return
	end

	local circle = v.cachePatch("GMCOOPC")

	local numbers = {}
	for i = 0, 9, 1 do
		numbers[i] = v.cachePatch("GMSNUM\$i\")
	end

	local plus = v.cachePatch("GMPLUS")

	local STARTX = 292 << FRACBITS
	local STARTY = 66 << FRACBITS + (coopCombo.offset << FRACBITS)

	local ENDX = 92 << FRACBITS
	local ENDY = 32 << FRACBITS

	local XDIFF = STARTX - ENDX
	local YDIFF = STARTY - ENDY

	local interp = (coopCombo.cashIn * FRACUNIT) / Gamma.CASHININTERP
	if (interp > FRACUNIT)
		interp = FRACUNIT
	end

	local x = ENDX + FixedMul(XDIFF, interp)
	local y = ENDY + FixedMul(YDIFF, interp)

	local WAUGHMPSCALE = FRACUNIT/3
	local waughmpVal = coopCombo.waughmp

	if (waughmpVal > Gamma.WAUGHMPTIME)
		waughmpVal = Gamma.WAUGHMPTIME - abs(Gamma.WAUGHMPTIME - $1)
	end

	if (waughmpVal < 0)
		waughmpVal = 0
	end

	local waughmp = FRACUNIT + (((waughmpVal * waughmpVal) * WAUGHMPSCALE) / (Gamma.WAUGHMPTIME * Gamma.WAUGHMPTIME))
	local scale = FixedMul(interp, waughmp)

	local flags = V_SNAPTORIGHT|V_SNAPTOTOP|V_PERPLAYER

	v.drawScaled(
		x, y, scale,
		circle,
		flags,
		nil)

	local playerIcon = nil
	local playerColormap = nil
	local playerHires = FRACUNIT
	if (coopCombo.player and coopCombo.player.valid)
	and (coopCombo.player.realmo and coopCombo.player.realmo.valid)
		playerIcon = v.getSprite2Patch(coopCombo.player.realmo.skin, SPR2_LIFE)
		playerColormap = v.getColormap(coopCombo.player.realmo.skin, coopCombo.player.skincolor)
		playerHires = skins[coopCombo.player.realmo.skin].highresscale
	end

	if (playerIcon == nil)
		// Can happen if the player leaves...
		playerIcon = v.cachePatch("GMNOPLYR")
	end

	v.drawScaled(
		x, y, FixedMul(playerHires, scale),
		playerIcon,
		flags,
		playerColormap)

	local timeY = y + (11 * scale)

	local timeVal = coopCombo.addTime
	local timeDigits = 0

	while (timeVal > 0)
		timeDigits = $1 + 1
		timeVal = $1 / 10
	end

	local timeX = x + ((timeDigits-1) * 4 * scale)

	timeVal = coopCombo.addTime
	while (timeVal > 0)
		v.drawScaled(
			timeX, timeY, scale,
			numbers[timeVal % 10],
			flags,
			nil)

		timeX = $1 - (9 * scale)
		timeVal = $1 / 10
	end
end

local function GammaDrawAllCombos(v, player)
	if (maptol & TOL_NIGHTS)
	or (G_IsSpecialStage())
		return;
	end

	if not (player.mo and player.mo.valid)
		return;
	end

	if (player.gammaTimerRan == true)
		local tetherTeam = nil;
		local comboList = {};
		local coopCombos = {};
		local playersAdded = {};

		if (Gamma.TetherAdded() == true)
		and (player.tetherPersist != nil)
			local teamID = player.tetherPersist.teamID;
			local team = Tether.Teams[teamID];

			if (Tether.TeamIsValid(team))
				tetherTeam = team;
			end
		end

		-- Start with your own lists.
		if (player.mo.gammaVars != nil)
			for _,v in pairs(player.mo.gammaVars.comboList) do
				table.insert(comboList, v);
			end

			playersAdded[#player] = true;

			if (tetherTeam == nil) -- This will be done below for Tether.
				if (player.mo.gammaVars != nil)
					for _,v in pairs(player.mo.gammaVars.coopCombos) do
						local ccPlayer = v.player;

						if (ccPlayer and ccPlayer.valid)
						and (playersAdded[#ccPlayer] != true)
							table.insert(coopCombos, v);
							playersAdded[#ccPlayer] = true;
						end
					end
				end
			end
		end

		if (tetherTeam != nil)
			-- We're on a Tether team, so display everyone's combos for everyone involved!
			local teamPlayers = tetherTeam.playerList;
			local numPlayers = #teamPlayers;

			for i = 1,numPlayers do
				local otherPlayer = teamPlayers[i];
				if not (otherPlayer and otherPlayer.valid)
					continue;
				end

				if (Tether.PlayerReady(otherPlayer) == false)
					continue;
				end

				if (otherPlayer == player)
					continue;
				end

				if (otherPlayer.mo.gammaVars != nil)
				and (otherPlayer.mo.gammaVars.comboList != nil)
				and (playersAdded[#otherPlayer] != true)
					for _,v in pairs(otherPlayer.mo.gammaVars.comboList) do
						table.insert(comboList, v);
					end

					playersAdded[#otherPlayer] = true;
				end
			end

			-- Now collect the coop combos.
			if (player.mo.gammaVars != nil)
				-- This is done above for not-Tether
				for _,v in pairs(player.mo.gammaVars.coopCombos) do
					local ccPlayer = v.player;

					if (ccPlayer and ccPlayer.valid)
					and (playersAdded[#ccPlayer] != true)
						table.insert(coopCombos, v);
						playersAdded[#ccPlayer] = true;
					end
				end
			end

			for i = 1,numPlayers do
				local otherPlayer = teamPlayers[i];
				if not (otherPlayer and otherPlayer.valid)
					continue;
				end

				if (Tether.PlayerReady(otherPlayer) == false)
					continue;
				end

				if (otherPlayer == player)
					continue;
				end

				if (otherPlayer.mo.gammaVars != nil)
				and (otherPlayer.mo.gammaVars.coopCombos != nil)
					for _,v in pairs(otherPlayer.mo.gammaVars.coopCombos) do
						local ccPlayer = v.player;

						if (ccPlayer and ccPlayer.valid)
						and (playersAdded[#ccPlayer] != true)
							table.insert(coopCombos, v);
							playersAdded[#ccPlayer] = true;
						end
					end
				end
			end

			-- Fun bug with this:
			-- Having multiple Gammas do combos at the same time on a team will cause the time indicators to overlap.
			-- I didn't mind this enough to fix since it will increase the code complexity even further,
			-- and technically I already went the extra mile fixing it for SP, since SA1 didn't even bother :p
		end

		for _,combo in pairs(comboList) do
			GammaDrawCombo(v, player, combo);
		end

		for _,coopCombo in pairs(coopCombos) do
			GammaDrawCoopCombo(v, player, coopCombo);
		end
	end
end

customhud.SetupItem("gammaCombo", hudIdentifier, GammaDrawAllCombos, "game", 1);

local function GammaObjectTracking(viewpoint, point)
	local result = {
		pos = Vector2.New(0, 0),
		scale = FRACUNIT,
		onScreen = false,
	};

	if not (viewpoint)
		return result;
	end

	local viewZ = viewpoint.pos.z - point.z;

	-- Calculate screen size adjustments.
	-- TODO: Properly support non-green resolutions
	local screenSize = Vector2.New(320, 200);

	if (splitscreen)
		screenSize.y = $1 / 2;
	end

	local screenHalf = (screenSize / 2) * FRACUNIT;

	-- Calculate FOV adjustments.
	local fovCvar = CV_FindVar("fov");

	local baseFov = 90*FRACUNIT;
	local fovDiff = fovCvar.value - baseFov;
	local fov = ((baseFov - fovDiff) / 2) - (viewpoint.fovAdd / 2);
	local fovTangent = tan(FixedAngle(fov));

	if (splitscreen)
		-- Splitscreen FOV is adjusted to maintain expected vertical view
		fovTangent = 10*fovTangent/17;
	end

	local fg = (screenSize.x >> 1) * fovTangent;

	-- Determine viewpoint factors.
	local h = R_PointToDist2(point.x, point.y, viewpoint.pos.x, viewpoint.pos.y);
	local da = viewpoint.angle - R_PointToAngle2(point.x, point.y, viewpoint.pos.x, viewpoint.pos.y);
	local dp = FixedDiv(viewZ, 1 + FixedMul(cos(da), h));

	-- Set results!
	result.pos = Vector2.New(
		screenHalf.x + FixedMul(tan(da), fg),
		screenHalf.y + FixedMul(tan(viewpoint.aiming) - dp, fg)
	);

	result.scale = FixedDiv(screenHalf.x, h + 1);

	result.onScreen = (abs(da) > ANG60 or abs(viewpoint.aiming - R_PointToAngle2(0, 0, h, viewZ)) > ANGLE_45);

	if (viewpoint.flip == -1)
		-- Flip vertically.
		result.pos.y = (screenSize.y << FRACBITS) - $1;
	end

	-- Cheap dirty hacks for some split-screen related cases
	if (result.pos.x < 0) or (result.pos.x > (screenSize.x << FRACBITS))
		result.onScreen = false;
	end

	if (result.pos.y < 0) or (result.pos.y > (screenSize.y << FRACBITS))
		result.onScreen = false;
	end

	return result;
end

local function GammaDrawTarget(v, viewpoint, viewpointPlayer, target, order, highlight)
	local theirFlip = P_MobjFlip(target.mobj);
	local combineFlip = theirFlip * viewpoint.flip;

	local feetZ = target.mobj.z;
	if (combineFlip < 0)
		feetZ = $1 + target.mobj.height;
	end

	local point = Vector3.New(target.mobj.x, target.mobj.y, feetZ);
	local result = GammaObjectTracking(viewpoint, point);

	if (result.onScreen == false)
		return;
	end

	-- Since we aren't using V_PERPLAYER,
	-- we need to do this adjustment manually.
	local splitTwo = false;
	if (splitscreen)
	and (secondarydisplayplayer and secondarydisplayplayer.valid)
	and (viewpointPlayer == secondarydisplayplayer)
		result.pos.y = $1 + 100*FRACUNIT;
		splitTwo = true;
	end

	local xPatch = v.cachePatch("GMWNUMX");

	local numbers = {};
	for i = 0, 9, 1 do
		if (highlight)
			numbers[i] = v.cachePatch("GMRNUM\$i\");
		else
			numbers[i] = v.cachePatch("GMWNUM\$i\");
		end
	end

	result.pos.x = $1 + 8*FRACUNIT;
	result.pos.y = $1 + 8*FRACUNIT;

	v.drawScaled(
		result.pos.x, result.pos.y,
		FRACUNIT,
		xPatch,
		0, nil
	);

	local val = order;
	local digits = 0;

	while (val > 0)
		digits = $1 + 1;
		val = $1 / 10;
	end

	result.pos.x = $1 + (digits * 8 * FRACUNIT)

	val = order;
	while (val > 0)
		v.drawScaled(
			result.pos.x, result.pos.y,
			FRACUNIT,
			numbers[val % 10],
			0, nil
		);

		result.pos.x = $1 - (8 * FRACUNIT);
		val = $1 / 10;
	end
end

local function GammaTargetTracking(v, player, cam, combo)
	if (combo.fired == false)
	and (combo.expireTime >= Gamma.ComboExpire - Gamma.ComboExpireBlink)
	and (leveltime & 1)
		-- Blink to make it obvious your combo is about to expire!
		return;
	end

	local sortedTargets = {};
	local bestOrder = 0;
	local lowestOrder = 0;

	for _,x in pairs(combo.aimList) do
		if not (x)
			continue;
		end

		if (x.order <= 0)
			continue;
		end

		if not (x.mobj and x.mobj.valid)
			continue;
		end

		if (x.order > bestOrder)
			bestOrder = x.order;
		end

		if (x.order < lowestOrder)
		or (lowestOrder == 0)
			lowestOrder = x.order;
		end

		table.insert(sortedTargets, x);
	end

	for _,x in pairs(combo.shotList) do
		if not (x)
			continue;
		end

		if (x.order <= 0)
			continue;
		end

		if not (x.mobj and x.mobj.valid)
			continue;
		end

		if (x.order > bestOrder)
			bestOrder = x.order;
		end

		if (x.order < lowestOrder)
		or (lowestOrder == 0)
			lowestOrder = x.order;
		end

		table.insert(sortedTargets, x);
	end

	local viewpoint = {
		pos = Vector3.New(0, 0, 0),
		angle = 0,
		aiming = 0,
		roll = 0,
		fovAdd = 0,
		flip = 1,
	};

	if (player.awayviewmobj and player.awayviewmobj.valid and player.awayviewtics > 0)
		viewpoint.pos = Vector3.New(player.awayviewmobj.x, player.awayviewmobj.y, player.awayviewmobj.z);
		viewpoint.angle = player.awayviewmobj.angle;
		viewpoint.aiming = player.awayviewaiming;

		-- Source code seems to indicate this situation uses the player's flip,
		-- not the awayviewmobj's flip. Couldn't find an example to confirm this,
		-- and I kinda don't care enough to make my own LOL
		if (player.pflags & PF_FLIPCAM)
			viewpoint.flip = P_MobjFlip(player.mo);
		else
			viewpoint.flip = 1;
		end
	elseif (cam.chase == true) and (player.spectator == false)
		-- Use the camera's properties.
		viewpoint.pos = Vector3.New(cam.x, cam.y, cam.z);
		viewpoint.angle = cam.angle;
		viewpoint.aiming = cam.aiming;

		if (player.pflags & PF_FLIPCAM)
			viewpoint.flip = P_MobjFlip(player.mo);
		else
			viewpoint.flip = 1;
		end
	else
		-- Use the player's properties.
		viewpoint.pos = Vector3.New(player.mo.x, player.mo.y, player.mo.z);
		viewpoint.angle = player.mo.angle;
		viewpoint.aiming = player.aiming;
		viewpoint.flip = P_MobjFlip(player.mo);
	end

	viewpoint.roll = player.viewrollangle; -- TODO: Properly support camera tilt
	viewpoint.fovAdd = player.fovadd;

	table.sort(sortedTargets, function(a, b)
		local aDist = P_AproxDistance(P_AproxDistance(viewpoint.pos.x - a.mobj.x, viewpoint.pos.y - a.mobj.y), viewpoint.pos.z - a.mobj.z);
		local bDist = P_AproxDistance(P_AproxDistance(viewpoint.pos.x - b.mobj.x, viewpoint.pos.y - b.mobj.y), viewpoint.pos.z - b.mobj.z);

		return (aDist < bDist);
	end);

	for i = 1,#sortedTargets do
		local target = sortedTargets[i];
		local highlight = (target.order == bestOrder);
		local order = (target.order - lowestOrder) + 1;
		GammaDrawTarget(v, viewpoint, player, target, order, highlight);
	end
end

local function GammaDrawAllTracking(v, player, camera)
	if (maptol & TOL_NIGHTS)
	or (G_IsSpecialStage())
		return;
	end

	if not (player.mo and player.mo.valid)
		return;
	end

	if (player.mo.gammaVars == nil)
		return;
	end

	local sortedCombos = {};

	for _,c in pairs(player.mo.gammaVars.comboList) do
		table.insert(sortedCombos, c);
	end

	table.sort(sortedCombos, function(a, b)
		return (a.startTime > b.startTime);
	end);

	for _,c in pairs(sortedCombos) do
		GammaTargetTracking(v, player, camera, c);
	end
end

customhud.SetupItem("gammaTracking", hudIdentifier, GammaDrawAllTracking, "game", INT32_MIN);

local function GammaDrawBonus(v, player, bonus)
	if (bonus == nil)
		return;
	end

	if (bonus.time < TICRATE/4)
	and (bonus.time & 1)
		return;
	end

	local x = 92 << FRACBITS;
	local y = 11 << FRACBITS + (bonus.offset << FRACBITS);
	local flags = V_SNAPTOLEFT|V_SNAPTOTOP|V_PERPLAYER;

	local perfectColors = {
		SKINCOLOR_GAMMABONUS1,
		SKINCOLOR_GAMMABONUS3, // these being flipped is the
		SKINCOLOR_GAMMABONUS2, // only reason this is a table
		SKINCOLOR_GAMMABONUS4,
		SKINCOLOR_GAMMABONUS5,
		SKINCOLOR_GAMMABONUS6,
		SKINCOLOR_GAMMABONUS7,
		SKINCOLOR_GAMMABONUS8,
		SKINCOLOR_GAMMABONUS9,
		SKINCOLOR_GAMMABONUS10,
		SKINCOLOR_GAMMABONUS11
	};

	local isPerfect = false;
	local c = SKINCOLOR_NONE;
	if (bonus.i >= 10)
		c = (leveltime % 11);
		isPerfect = true;
	else
		c = SKINCOLOR_GAMMABONUS1 + (bonus.i - 1);
	end

	local str = "";
	if (bonus.time < TICRATE)
		str = tostring(bonus.amt * 100);
	else
		local bonusLabels = {
			"Good!",
			"Nice!",
			"Great!",
			"Jammin'!",
			"Cool!",
			"Radical!",
			"Tight!",
			"Awesome!",
			"Extreme!",
			"Perfect!"
		};
		str = bonusLabels[ bonus.i ];
	end

	local strwidth = customhud.CustomFontStringWidth(v, str, "GMSFT", FRACUNIT);
	local nextx = x - (strwidth / 2);

	for i = 1,str:len() do
		local nextbyte = str:byte(i,i);
		local clr = c;

		if (isPerfect == true)
			clr = perfectColors[ c + 1 ];
			c = ($1 + 1) % 11;
		end

		local delta = abs( i - bonus.charHop );
		local hop = 0;
		if (delta == 1)
			hop = 2;
		elseif (delta == 0)
			hop = 3;
		end

		local result = customhud.CustomFontChar(
			v,
			nextx, y - (hop << FRACBITS), nextbyte,
			"GMSFT",
			flags, FRACUNIT, clr
		);

		nextx = result;
	end
end

local function GammaDrawAllBonus(v, player)
	if (maptol & TOL_NIGHTS)
	or (G_IsSpecialStage())
		return;
	end

	if not (player.mo and player.mo.valid)
		return;
	end

	if (player.mo.gammaVars != nil)
		// Display the higher score values on top.
		local bonusSorted = {};

		for _,bonus in pairs(player.mo.gammaVars.bonusHud) do
			table.insert(bonusSorted, bonus);
		end
		table.sort(bonusSorted, function(a, b)
			return a.amt < b.amt;
		end);

		for _,bonus in ipairs(bonusSorted) do
			GammaDrawBonus(v, player, bonus);
		end
	end
end

customhud.SetupItem("gammaBonus", hudIdentifier, GammaDrawAllBonus, "game", 2);
