Gamma.AllowAbilities = function(player, stasis, allowInCarry)
	if (player.exiting) or (player.pflags & PF_GAMETYPEOVER)
		// No abilities when you're done
		return false;
	end

	if ((gametyperules & GTR_RACE) and (leveltime < 4*TICRATE))
		// Not during start countdown
		return false;
	end

	if (player.gotflag)
		// No jet abilities with the CTF flag.
		return false;
	end

	if (Gamma.BattleAdded())
		if (player.iseggrobo)
		or (player.eggrobo_transforming)
		or (player.isjettysyn)
			// No abilities in morphed states.
			return false;
		end

		if (player.gotcrystal)
		or (player.gotflagdebuff)
			// No abilities with the CTF flag.
			return false;
		end
	end

	if (stasis != nil)
		if (player.pflags & stasis)
			// Don't allow in this stasis type
			return false;
		end

		if (player.powers[pw_nocontrol])
			if ((stasis & PF_JUMPSTASIS) == 0)
			or (player.powers[pw_nocontrol] & (1<<15))
				return false;
			end
		end

		if (stasis & PF_JUMPSTASIS)
			if (Gamma.TetherAdded())
				if (player.mo and player.mo.valid)
				and (player.mo.tether != nil)
				and (player.mo.tether.wallKickDir != nil)
					// Can do a Tether wall kick.
					// Don't activate jet abilities during these.
					return false;
				end
			end

			if (Gamma.BattleAdded())
				if (player.exhuastmeter != nil)
				and (player.exhuastmeter <= 0)
					// Exhausted. Don't allow jet abilities.
					return false;
				end
			end
		end
	end

	if (player.mo.state >= S_PLAY_SUPER_TRANS1 and player.mo.state <= S_PLAY_SUPER_TRANS6)
		// No abilities during transform
		return false;
	end

	if ((player.powers[pw_carry] or player.mo.gammaVars.prevCarry) and (allowInCarry == false))
		// No abilities when in a special carry state
		return false;
	end

	if (P_PlayerInPain(player))
		// Don't allow this ability in pain state
		return false;
	end

	return true;
end

Gamma.Landing = function(player)
	player.mo.gammaVars.landedTics = TICRATE/2;
	S_StartSound(player.mo, sfx_gmland);
	Gamma.LandDust(player.mo);

	local checkBust = nil;
	local flip = P_MobjFlip(player.mo);
	if (flip == -1)
		checkBust = player.mo.ceilingrover;
	else
		checkBust = player.mo.floorrover;
	end

	if (checkBust and checkBust.valid)
	and (checkBust.flags & FF_EXISTS)
	and (checkBust.flags & FF_BUSTUP)
		EV_CrumbleChain(nil, checkBust);

		if (checkBust.master.flags & ML_EFFECT5)
			P_LinedefExecute(
				P_AproxDistance(checkBust.master.dx, checkBust.master.dy) >> FRACBITS,
				player.mo,
				player.mo.subsector.sector
			);
		end
	end
end

Gamma.GameOverMusic = function(player)
	if (player.lives <= 0)
	and (player.deadtimer == 1) then
		P_PlayJingleMusic(player, "GMAOVR", 0, false, JT_GOVER)
	end
end

Gamma.Drink = function(player)
	if (Gamma.CanDrink(player) == true) and (player.cmd.buttons & BT_CUSTOM1)
		if (Gamma.InDrinkState(player.mo) == false)
			player.mo.state = S_PLAY_GAMMA_DRINK_START;
		end
	elseif (player.mo.state == S_PLAY_GAMMA_DRINK_LOOP)
		player.mo.state = S_PLAY_GAMMA_DRINK_END;
	end

	if (Gamma.InDrinkState(player.mo) == true)
		player.pflags = $1 | PF_FULLSTASIS;
	end
end

Gamma.DefaultGammaVars = function(player)
	if (player.gammaSwitchSound)
		S_StartSound(player.mo, sfx_gmsksw);
		player.gammaSwitchSound = false;
	end

	player.mo.gammaVars = {
		driveMode = false,
		driveTimer = 0,
		driving = false,
		driveSound = false,
		keepDrive = false,

		aimAngle = player.drawangle,
		shootAngle = player.drawangle,

		focusTime = 0,
		focusTarget = nil,

		laserAngle = player.drawangle,
		prevLaserAngle = player.drawangle,
		laserPitch = 0,
		prevLaserPitch = 0,

		jetDJTime = 0,
		jetHovering = false,
		jetStart = 0,
		jetOverlay = nil,

		propellerMode = false,
		propellerDelay = 0,

		aiming = false,
		comboGoing = false,

		currentCombo = nil,
		comboList = {},

		numCoopCombos = 0,
		coopCombos = {},

		landedTics = 0,

		addSeconds = 0,
		addBonus = 0,
		addingPause = 0,

		prevOnGround = P_IsObjectOnGround(player.mo),
		prevJump = false,
		prevSpin = false,
		prevCarry = player.powers[pw_carry],

		minecartSideMove = 0,
		ropeSpin = 0,

		bonusHud = {},
		didHurry = false,
	};

	if (player.gammaTime == nil)
		Gamma.InitPlayerTime(player);
	end

	player.gammaTimeAttack = player.realtime;
	player.gammaTimeAttackPause = 0;
end

Gamma.NewCombo = function()
	return {
		aimList = {},
		shotList = {},
		expireTime = 0,
		fired = false,
		addTime = 0,
		numGotten = 0,
		cashIn = Gamma.CASHINTIME,
		waughmp = Gamma.WAUGHMPTIME * 2,
		offset = 0,
		startTime = 0,
		numWithPoints = 0,
		powered = false,
	};
end

Gamma.Main = function(player)
	player.mo.sprite2 = SPR2_STND; // Prevent silly afterimages.

	// Fix empty spots in these tables before using them!
	for _,combo in pairs(player.mo.gammaVars.comboList)
		combo.aimList = Gamma.RemoveBlankSpotsFromTable(combo.aimList);
		combo.shotList = Gamma.RemoveBlankSpotsFromTable(combo.shotList);
	end

	local onGround = P_IsObjectOnGround(player.mo);

	Gamma.Drink(player);

	player.mo.gammaVars.aiming = false;
	Gamma.UpdateAimAngle(player);
	Gamma.UpdateLaserAngle(player);

	Gamma.DriveModeLogic(player);
	Gamma.AbilityHandler(player);

	Gamma.DoSounds(player);

	local spinBT = (player.cmd.buttons & BT_SPIN)
	if (player.powers[pw_carry] == CR_ROPEHANG)
		spinBT = player.mo.gammaVars.ropeSpin;
	end

	if (spinBT) and (Gamma.AllowAbilities(player, PF_STASIS, true))
		if not (player.mo.gammaVars.prevSpin) and (player.mo.gammaVars.currentCombo == nil)
			local i = 1;
			while (player.mo.gammaVars.comboList[i] != nil)
				i = $1+1;
			end

			player.mo.gammaVars.currentCombo = i;
			player.mo.gammaVars.comboList[i] = Gamma.NewCombo();
		end

		if (player.mo.gammaVars.currentCombo != nil)
			local droppedCombo = Gamma.BeamAim(
				player,
				player.mo.gammaVars.currentCombo
			);

			if (droppedCombo == true)
				player.mo.gammaVars.comboList[player.mo.gammaVars.currentCombo] = nil;
				player.mo.gammaVars.currentCombo = nil;
				Gamma.PlaySoundOnPart(player, Gamma.AnimPartIDs.LASER, nil);
			end
		end
	else
		if (player.mo.gammaVars.currentCombo != nil)
			player.mo.gammaVars.comboList[player.mo.gammaVars.currentCombo].fired = true;
			player.mo.gammaVars.currentCombo = nil;
		end

		Gamma.PlaySoundOnPart(player, Gamma.AnimPartIDs.LASER, nil);
	end

	if not (player.mo.gammaVars.aiming)
		-- Reset focus when not aiming.
		player.mo.gammaVars.focusTime = 0;
		player.mo.gammaVars.focusTarget = nil;
	end

	local timeWillBeAdded = false;
	local didShot = false;
	local aimGunAt = nil;

	player.mo.gammaVars.comboGoing = false;

	for k,combo in pairs(player.mo.gammaVars.comboList)
		if (combo.waughmp > 0)
			// HUD animation
			combo.waughmp = $1 - 1;
		end

		local destOffset = Gamma.ComboOffset(player.mo.gammaVars.comboList, combo.startTime);
		local diff = abs(combo.offset - destOffset);

		if (diff <= Gamma.OFFSETSPEED)
			combo.offset = destOffset;
		else
			if (combo.offset > destOffset)
				combo.offset = $1 - Gamma.OFFSETSPEED;
			else
				combo.offset = $1 + Gamma.OFFSETSPEED;
			end
		end

		if (combo.fired != true)
			continue;
		end

		if (combo.addTime > 0)
			// Prevents us from killing the player in close calls!
			timeWillBeAdded = true;
		end

		if (P_PlayerInPain(player))
			local aimLen = #(combo.aimList);
			if (aimLen > 0)
				Gamma.ComboDropSound(player);
			end

			// End whatever combos are going in pain state.
			combo.aimList = {};
		end

		if (#combo.aimList > 0)
			if (didShot == false)
				for _,t in pairs(combo.aimList) do
					if (t and t.mobj and t.mobj.valid)
						aimGunAt = t.mobj;
						break;
					end
				end

				if (player.weapondelay <= 0)
					Gamma.Shoot(player.mo, combo.aimList, k);
					player.weapondelay = 7;
				end
			end

			player.mo.gammaVars.comboGoing = true;
			didShot = true;
		else
			if (combo.cashIn > 0)
				combo.cashIn = $1 - 1;
				continue;
			end

			// End our combo
			if (combo.addTime > 0)
				player.mo.gammaVars.addSeconds = $1 + combo.addTime;
			end

			if (combo.numGotten > 1)
				// Sonic Adventure 2 style
				// extra score bonus.
				local bonus = 0;
				for i = 2,combo.numGotten do
					if (i >= 10)
						// Add 500 points at a time.
						// Unlike SA2, this is uncapped!!
						bonus = $1 + 5;
					elseif (i >= 8)
						// Add 200 points at a time.
						bonus = $1 + 2;
					else
						bonus = $1 + 1;
					end
				end

				P_AddPlayerScore(player, bonus * 100);

				local newBonus = {
					amt = bonus,
					i = min(combo.numGotten - 1, 10),
					offset = 16,
					charHop = -1,
					time = 2*TICRATE,
				};

				table.insert(player.mo.gammaVars.bonusHud, newBonus);
			end

			player.mo.gammaVars.comboList[k] = nil;
		end
	end

	Gamma.UpdateShootAngle(player, aimGunAt);

	player.mo.gammaVars.numCoopCombos = #(player.mo.gammaVars.coopCombos);

	for k,coopCombo in pairs(player.mo.gammaVars.coopCombos)
		if (coopCombo.waughmp > 0)
			// HUD animation
			coopCombo.waughmp = $1 - 1;
		end

		local destOffset = Gamma.CoopComboOffset(player.mo.gammaVars.coopCombos, k);
		local diff = abs(coopCombo.offset - destOffset);

		if (diff <= Gamma.OFFSETSPEED)
			coopCombo.offset = destOffset;
		else
			if (coopCombo.offset > destOffset)
				coopCombo.offset = $1 - Gamma.OFFSETSPEED;
			else
				coopCombo.offset = $1 + Gamma.OFFSETSPEED;
			end
		end

		if (coopCombo.addTime > 0)
			// Prevents us from killing the player in close calls!
			timeWillBeAdded = true;
		end

		local keepAlive = false;

		if (coopCombo.player and coopCombo.player.valid)
			keepAlive = (coopCombo.player.scoreadd > 0);

			if (coopCombo.player.mo and coopCombo.player.mo.valid)
			and (coopCombo.player.mo.gammaVars != nil)
			and (coopCombo.player.mo.gammaVars.comboGoing == true)
				keepAlive = true;
			end
		end

		if (keepAlive == true)
			coopCombo.cashIn = Gamma.CASHINTIME;
			continue;
		elseif (coopCombo.cashIn > 0)
			coopCombo.cashIn = $1 - 1;
			continue;
		end

		// End this combo
		if (coopCombo.addTime > 0)
			player.mo.gammaVars.addSeconds = $1 + coopCombo.addTime;
		end

		player.mo.gammaVars.coopCombos[k] = nil;
	end

	for k,bonus in pairs(player.mo.gammaVars.bonusHud) do
		if (bonus.offset > 0)
			bonus.offset = $1 / 2;
		else
			bonus.charHop = $1 + 1;
		end

		if (bonus.time > 0)
			bonus.time = $1 - 1;
		end

		if (bonus.time == 0)
			player.mo.gammaVars.bonusHud[k] = nil;
		end
	end

	if (player.mo.gammaVars.addSeconds > 0)
		timeWillBeAdded = true;
	end

	if (Gamma.MapTimeLimit == nil)
		// mid-game adding fix
		Gamma.MapTimeLimit = Gamma.DefaultTime(gamemap);
	end

	if (Gamma.TetherAdded() != true) -- We do this somewhere else for tether.
		Gamma.RunTimer(player, timeWillBeAdded);
	end

	if (player.mo.gammaVars.landedTics > 0)
		player.mo.gammaVars.landedTics = $1 - 1;
	end

	if (onGround == true and player.mo.gammaVars.prevOnGround == false)
		Gamma.Landing(player);
	end

	player.mo.gammaVars.prevOnGround = onGround;

	if (player.cmd.buttons & BT_JUMP) or (player.powers[pw_carry])
		player.mo.gammaVars.prevJump = true;
	else
		player.mo.gammaVars.prevJump = false;
	end

	if (spinBT)
		player.mo.gammaVars.prevSpin = true;
	else
		player.mo.gammaVars.prevSpin = false;
	end

	player.mo.gammaVars.prevCarry = player.powers[pw_carry];
end

addHook("MobjThinker", Gamma.ShotThink, MT_GAMMA_SHOT);
addHook("MobjDeath", Gamma.ShotRemove, MT_GAMMA_SHOT);
addHook("MobjRemoved", Gamma.ShotRemove, MT_GAMMA_SHOT);
addHook("MobjFuse", Gamma.ShotFuse, MT_GAMMA_SHOT);

addHook("MobjThinker", Gamma.ShotTrailThink, MT_GAMMA_SHOTTRAIL);

addHook("MobjThinker", Gamma.TargetThink, MT_GAMMA_TARGET);

addHook("MobjThinker", Gamma.LaserThink, MT_GAMMA_LASER);
addHook("MobjThinker", Gamma.JetSmokeThink, MT_GAMMA_JETSMOKE);

addHook("MobjThinker", Gamma.LandDustSpawner, MT_GAMMA_LANDDUST_SPAWNER);

addHook("MobjThinker", Gamma.BlasterExplosionThink, MT_GAMMA_BLASTEREXPLOSION_CENTER);
addHook("MobjRemoved", Gamma.BlasterExplosionRemove, MT_GAMMA_BLASTEREXPLOSION_CENTER);

addHook("MobjThinker", Gamma.CoopJumpGlowThink, MT_GAMMA_COOPJUMPGLOW);

addHook("PlayerThink", function(player)
	if not (player.mo and player.mo.valid)
		return
	end

	local dead = (player.mo.health <= 0)
	local onGround = P_IsObjectOnGround(player.mo)

	if (dead)
		if (player.mo.gammaCoopJump != nil)
			if (player.mo.gammaCoopJump.glow and player.mo.gammaCoopJump.glow.valid)
				P_RemoveMobj(player.mo.gammaCoopJump.glow);
			end
		end

		if (player.gammaJumpSet)
			player.jumpfactor = skins[player.mo.skin].jumpfactor;
			player.gammaJumpSet = nil;
		end
	else
		Gamma.CoopJumpLogic(player);
	end

	if not (player.mo.skin == "gamma")
		player.gammaSwitchSound = true;

		if (player.mo.gammaVars != nil)
		and (player.mo.gammaVars.driveMode == true)
			-- Undo drive mode
			player.mo.state = S_PLAY_STND;
		end

		if (Gamma.InDrinkState(player.mo) == true)
			-- Undo drink state
			player.mo.state = S_PLAY_STND;
		end
	end

	if (dead) and (player.mo.skin == "gamma")
		-- Gamma game over handling
		player.mo.momz = -P_GetMobjGravity(player.mo);
		//player.mo.flags = $1|MF_NOGRAVITY;
		Gamma.GameOverMusic(player);
	end

	if (dead) or not (player.mo.skin == "gamma")
		-- Resetting when you're not gamma!
		if (player.mo.gammaVars != nil)
			if (player.mo.gammaVars.jetOverlay and player.mo.gammaVars.jetOverlay.valid)
				P_RemoveMobj(player.mo.gammaVars.jetOverlay);
			end
		end

		player.mo.gammaVars = nil;
		return;
	end

	if (player.mo.gammaVars == nil)
	or (player.powers[pw_carry] == CR_NIGHTSMODE)
		Gamma.DefaultGammaVars(player);
		player.mo.gammaVars.prevCarry = CR_NIGHTSMODE;
	end

	if (player.powers[pw_carry] == CR_NIGHTSMODE)
		return;
	end

	Gamma.Main(player);
end)

addHook("PlayerSpawn", function(player)
	if (player.mo and player.mo.valid)
	and (player.gammaJumpSet)
		player.jumpfactor = skins[player.mo.skin].jumpfactor;
		player.gammaJumpSet = nil;
	end

	Gamma.InitPlayerTime(player);
end)

local function AntiRail(player)
	if (player.mo and player.mo.valid)
	and (player.mo.skin == "gamma")
	and (player.mo.gammaVars != nil)
		if (player.powers[pw_carry] == CR_MINECART)
			player.mo.gammaVars.minecartSideMove = player.cmd.sidemove;

			if (player.cmd.buttons & BT_SPIN)
				// No jumping between rails if you're trying to aim at things
				player.cmd.sidemove = 0;
			end
		elseif (player.powers[pw_carry] == CR_ROPEHANG)
			if (player.cmd.buttons & BT_SPIN)
				// Don't jump off of ropes
				player.mo.gammaVars.ropeSpin = BT_SPIN;
				player.cmd.buttons = $1 & ~BT_SPIN;
			else
				player.mo.gammaVars.ropeSpin = 0;
			end
		end
	end
end

addHook("PreThinkFrame", function()
	Gamma.InitBattle();

	for player in players.iterate do
		AntiRail(player);
		player.gammaTimerRan = nil;
	end
end)

local function CalcHeight(player, pviewheight)
	// Doesn't work super great but w/e
	local mo = player.mo;
	local nudge = mo.scale;

	player.viewheight = pviewheight;

	player.bob = (FixedMul(player.rmomx, player.rmomx)
		+ FixedMul(player.rmomy, player.rmomy)) >> 2;

	local MAXBOB = FixedMul(0x10 << FRACBITS, mo.scale);
	if (player.bob > MAXBOB)
		player.bob = MAXBOB;
	end

	local onGround = P_IsObjectOnGround(mo);
	if not (onGround)
		if (mo.eflags & MFE_VERTICALFLIP)
			player.viewz = mo.z + mo.height - player.viewheight;

			if (player.viewz < mo.floorz + nudge)
				player.viewz = mo.floorz + nudge;
			end
		else
			player.viewz = mo.z + player.viewheight;

			if (player.viewz > mo.ceilingz - nudge)
				player.viewz = mo.ceilingz - nudge;
			end
		end

		return;
	end

	local angle = ANGLE_MAX / 20 * leveltime;
	local bob = FixedMul(player.bob / 2, sin(angle));

	if (player.playerstate == PST_LIVE)
		player.viewheight = $1 + player.deltaviewheight;

		if (player.viewheight > pviewheight)
			player.viewheight = pviewheight;
			player.deltaviewheight = 0;
		end

		if (player.viewheight < pviewheight/2)
			player.viewheight = pviewheight/2;

			if (player.deltaviewheight <= 0)
				player.deltaviewheight = 1;
			end
		end

		if (player.deltaviewheight)
			player.deltaviewheight = $1 + (mo.scale / 4);

			if (player.deltaviewheight == 0)
				player.deltaviewheight = 1;
			end
		end
	end

	if (mo.eflags & MFE_VERTICALFLIP)
		player.viewz = mo.z + mo.height - player.viewheight - bob;
	else
		player.viewz = mo.z + player.viewheight + bob;
	end

	nudge = 4 * mo.scale;

	if (player.viewz > mo.ceilingz - nudge)
		player.viewz = mo.ceilingz - nudge;
	end

	if (player.viewz < mo.floorz + nudge)
		player.viewz = mo.floorz + nudge;
	end
end

addHook("PostThinkFrame", function()
	for player in players.iterate do
		Gamma.UpdateJetOverlay(player);

		if (player.mo and player.mo.valid)
		and (player.mo.skin == "gamma")
			CalcHeight(player, player.mo.height * 64 / 48);

			if not (Gamma.InNightsMode(player.mo))
				player.mo.sprite2 = SPR2_STND -- Hides stuff like MP character select animation
			end
		end
	end

	Gamma.TetherTimeSharing();
	Gamma.RunRecordAttackGhosts();
end)

local function MapChanged(map)
	Gamma.MapTimeLimit = 0;
	Gamma.CoopTimeTotals = {};
	Gamma.BombList = {};
	Gamma.GhostList = {};
end

addHook("MapChange", MapChanged);

local function MapLoaded(map)
	Gamma.MapTimeLimit = Gamma.DefaultTime(map);
	Gamma.CoopTimeTotals = {};

	for player in players.iterate do
		if not (player.mo and player.mo.valid)
			continue;
		end

		Gamma.InitPlayerTime(player);
	end
end

addHook("MapLoad", MapLoaded);

local function ObjectDeath(mo, inf, src)
	if not ((src and src.valid) and (src.player and src.player.valid))
		return;
	end

	if (inf and inf.valid and inf.type == MT_GAMMA_SHOT)
		-- We give Gamma shot coop time somewhere else.
		mo.gaveCoopTime = true;
		return;
	end

	if (mo.gaveCoopTime == true)
		-- Just in case.
		return;
	end

	local combo = src.player.scoreadd + 1;
	local sec = Gamma.ComboToSeconds(combo) * Gamma.ComboPoints(mo) * 2;
	sec = Gamma.TetherTimeScale($1, src.player);

	Gamma.GiveCoopTime(sec, src.player);
	mo.gaveCoopTime = true;
end

addHook("MobjDeath", ObjectDeath);

addHook("NetVars", function(net)
	Gamma.MapTimeLimit = net(Gamma.MapTimeLimit);
	Gamma.CoopTimeTotals = net(Gamma.CoopTimeTotals);
	Gamma.BombList = net(Gamma.BombList);
end)

Gamma.AddToUnidusList = function(owner, ball)
	if (owner.unidusList == nil)
		owner.unidusList = {};
	end

	if (Gamma.TableContains(owner.unidusList, ball) == true)
		return;
	end

	table.insert(owner.unidusList, ball);
end

function A_UnidusBall(mo, v1, v2)
	super(mo, v1, v2);

	if not (mo and mo.valid)
		return;
	end

	local owner = mo.target;
	if not (owner and owner.valid and owner.health > 0)
		return;
	end

	if (mo.movecount)
		// was thrown, remove from list
		Gamma.RemoveFromTable(owner.unidusList, mo);
		return;
	end

	Gamma.AddToUnidusList(owner, mo);
end

function A_CheckBuddy(mo, v1, v2)
	super(mo, v1, v2);

	if not (mo and mo.valid)
		return;
	end

	local owner = mo.target;
	if (v1)
		owner = mo.tracer;
	end

	if not (owner and owner.valid and owner.health > 0)
		return;
	end

	Gamma.AddToUnidusList(owner, mo);
end
