//
// Base edits (enemies, projectiles, other gameplay objects)
//

// This function is used for the other files.
Gamma.FindAttachedObjects = function(sourceObject, searchType, exRadius, exHeight)
	-- Most multi-object contrapions in SRB2 don't set pointers,
	-- or if they do, it's one-way,
	-- so we just have to assume by searching blockmap for any parts.

	if (exRadius == nil)
		exRadius = 0;
	end

	if (exHeight == nil)
		exHeight = 0;
	end

	local function MatchType(mo)
		if (type(searchType) == "table")
			-- Multi-type matching
			for _,t in pairs(searchType) do
				if (mo.type == t)
					return true;
				end
			end
		else
			-- Only a single match
			if (mo.type == searchType)
				return true;
			end
		end

		return false;
	end

	local matches = {};
	local function FindObjs(ref, mo)
		if (mo.z + mo.height + exHeight < ref.z - exHeight)
		or (mo.z - exHeight > ref.z + ref.height + exHeight)
			return nil;
		end

		local xdist = abs(mo.x - ref.x);
		local ydist = abs(mo.y - ref.y);

		if (xdist > ref.radius + mo.radius + exRadius)
		or (ydist > ref.radius + mo.radius + exRadius)
			return nil;
		end

		if (MatchType(mo) == true)
			table.insert(matches, mo);
		end

		return nil;
	end

	searchBlockmap("objects", FindObjs, sourceObject);
	return matches;
end

// Edit existing projectiles to make them targetable.
mobjinfo[MT_FBOMB].flags = $1 & ~MF_NOBLOCKMAP;
mobjinfo[MT_CYBRAKDEMON_MISSILE].flags = $1 & ~MF_NOBLOCKMAP;
mobjinfo[MT_CYBRAKDEMON_NAPALM_BOMB_LARGE].flags = $1 & ~MF_NOBLOCKMAP;
mobjinfo[MT_CYBRAKDEMON_NAPALM_BOMB_SMALL].flags = $1 & ~MF_NOBLOCKMAP;
mobjinfo[MT_ROCKET].flags = $1 & ~MF_NOBLOCKMAP;
mobjinfo[MT_TORPEDO].flags = $1 & ~MF_NOBLOCKMAP;
mobjinfo[MT_TORPEDO2].flags = $1 & ~MF_NOBLOCKMAP;
mobjinfo[MT_ENERGYBALL].flags = $1 & ~MF_NOBLOCKMAP;
mobjinfo[MT_MINE].flags = $1 & ~MF_NOBLOCKMAP;
mobjinfo[MT_JETTBULLET].flags = $1 & ~MF_NOBLOCKMAP;
mobjinfo[MT_CANNONBALL].flags = $1 & ~MF_NOBLOCKMAP;
mobjinfo[MT_ARROW].flags = $1 & ~MF_NOBLOCKMAP;
mobjinfo[MT_CORK].flags = $1 & ~MF_NOBLOCKMAP;
mobjinfo[MT_REDRING].flags = $1 & ~MF_NOBLOCKMAP;
mobjinfo[MT_THROWNBOUNCE].flags = $1 & ~MF_NOBLOCKMAP;
mobjinfo[MT_THROWNINFINITY].flags = $1 & ~MF_NOBLOCKMAP;
mobjinfo[MT_THROWNAUTOMATIC].flags = $1 & ~MF_NOBLOCKMAP;
mobjinfo[MT_THROWNSCATTER].flags = $1 & ~MF_NOBLOCKMAP;
mobjinfo[MT_THROWNEXPLOSION].flags = $1 & ~MF_NOBLOCKMAP;
mobjinfo[MT_THROWNGRENADE].flags = $1 & ~MF_NOBLOCKMAP;
mobjinfo[MT_KOOPAFLAME].flags = $1 & ~MF_NOBLOCKMAP;
mobjinfo[MT_POPSHOT].flags = $1 & ~MF_NOBLOCKMAP;
mobjinfo[MT_CACOFIRE].flags = $1 & ~MF_NOBLOCKMAP;
mobjinfo[MT_FBOMB].flags = $1 & ~MF_NOBLOCKMAP;
mobjinfo[MT_TURRET].flags = $1 & ~MF_NOBLOCKMAP;
mobjinfo[MT_BLACKEGGMAN_GOOPFIRE].flags = $1 & ~MF_NOBLOCKMAP;
mobjinfo[MT_BLACKEGGMAN_MISSILE].flags = $1 & ~MF_NOBLOCKMAP;

// These flags fix a bunch of bugs with this object in general :V
mobjinfo[MT_BOSSJUNK].flags = $1|MF_NOCLIP|MF_NOCLIPTHING;

//
// Default target hooks
//

Gamma.TargetFuncAlways = function(mo, aimList, owner)
	// Always able to target.
	return true;
end

Gamma.SetTargetFunc(MT_TURRET, Gamma.TargetFuncAlways);
Gamma.SetTargetFunc(MT_CRUSHCLAW, Gamma.TargetFuncAlways);
Gamma.SetTargetFunc(MT_BANPSPRING, Gamma.TargetFuncAlways);
Gamma.SetTargetFunc(MT_MINUS, Gamma.TargetFuncAlways);
Gamma.SetTargetFunc(MT_DRAGONTAIL, Gamma.TargetFuncAlways);
Gamma.SetTargetFunc(MT_EGGROBO1, Gamma.TargetFuncAlways);
Gamma.SetTargetFunc(MT_EGGMOBILE4, Gamma.TargetFuncAlways);
Gamma.SetTargetFunc(MT_UNIBALL, Gamma.TargetFuncAlways);
Gamma.SetTargetFunc(MT_EGGMOBILE_BALL, Gamma.TargetFuncAlways);
Gamma.SetTargetFunc(MT_POINTYBALL, Gamma.TargetFuncAlways);
Gamma.SetTargetFunc(MT_METALSONIC_RACE, Gamma.TargetFuncAlways);
Gamma.SetTargetFunc(MT_SPINBOBERT_FIRE1, Gamma.TargetFuncAlways);
Gamma.SetTargetFunc(MT_SPINBOBERT_FIRE2, Gamma.TargetFuncAlways);
Gamma.SetTargetFunc(MT_BALLOON, Gamma.TargetFuncAlways);
Gamma.SetTargetFunc(MT_CANNONBALLDECOR, Gamma.TargetFuncAlways);
Gamma.SetTargetFunc(MT_TARGET, Gamma.TargetFuncAlways);
Gamma.SetTargetFunc(MT_BOMBSPHERE, Gamma.TargetFuncAlways);
Gamma.SetTargetFunc(MT_TNTBARREL, Gamma.TargetFuncAlways);

local function GammaHook_TargetPlayer(mo, aimList, owner)
	if not (mo.player and mo.player.valid)
		// Invalid player?
		return false;
	end

	if (mo.spectator == true)
		// No spectators please
		return false;
	end

	if not (mo.player.mo and mo.player.mo.valid and mo == mo.player.mo)
		// Invalid mobj?
		return false;
	end

	// Always target
	return true;
end

Gamma.SetTargetFunc(MT_PLAYER, GammaHook_TargetPlayer);

local function GammaHook_TargetGoop(mo, aimList, owner)
	if (mo.state != mo.info.meleestate)
		// Target goop when it's being sent out as a projectile.
		return true;
	end

	// Use default logic in other states.
	return nil;
end

Gamma.SetTargetFunc(MT_GOOP, GammaHook_TargetGoop);

local function GammaHook_TargetBrakAttack(mo, aimList, owner)
	if not (mo.flags & MF_NOCLIPTHING)
		// Target goop when it's being sent out as a projectile.
		return true;
	end

	// Otherwise, never target.
	return false;
end

Gamma.SetTargetFunc(MT_BLACKEGGMAN_GOOPFIRE, GammaHook_TargetBrakAttack);
Gamma.SetTargetFunc(MT_BLACKEGGMAN_MISSILE, GammaHook_TargetBrakAttack);

local function GammaHook_TargetCrushstacean(mo, aimList, owner)
	if (mo.tracer and mo.tracer.valid and mo.tracer.health > 0)
		// Only aim at the crab itself when the claw is targeted first.
		if (Gamma.AimListContains(aimList, mo.tracer) == false)
			return false;
		end
	end

	// Use default logic if the claw is targeted.
	return nil;
end

Gamma.SetTargetFunc(MT_CRUSHSTACEAN, GammaHook_TargetCrushstacean);
Gamma.SetTargetFunc(MT_BANPYURA, GammaHook_TargetCrushstacean);

local function GammaHook_TargetEggGuard(mo, aimList, owner)
	if (mo.tracer and mo.tracer.valid and mo.tracer.health > 0)
		// Never target while shield exists.
		return false;
	end

	// Use default logic when shield is dead.
	return nil;
end

Gamma.SetTargetFunc(MT_EGGGUARD, GammaHook_TargetEggGuard);

local function GammaHook_TargetDragonbomber(mo, aimList, owner)
	// Only aim at the head when the segments are targeted first.
	if (mo.tracer and mo.tracer.valid)
		local checkTail = mo.tracer;

		while (checkTail and checkTail.valid)
			if (Gamma.AimListContains(aimList, checkTail) == false)
				return false;
			end

			checkTail = checkTail.tracer;
		end
	end

	// Use default logic when all segments are targeted.
	return nil;
end

Gamma.SetTargetFunc(MT_DRAGONBOMBER, GammaHook_TargetDragonbomber);

local function GammaHook_TargetRolloutRock(mo, aimList, owner)
	if (owner and owner.valid)
	and (mo.tracer and mo.tracer.valid)
	and (mo.tracer == owner)
		// Do not target our own ball.
		return false;
	end

	// Always target.
	return true;
end

Gamma.SetTargetFunc(MT_ROLLOUTROCK, GammaHook_TargetRolloutRock);

local function GammaHook_TargetMetalBoss(mo, aimList, owner)
	if (mo.state == mo.info.spawnstate)
		// Always able to target while roaming.
		return true;
	end

	// Use default logic in other states.
	return nil;
end

Gamma.SetTargetFunc(MT_METALSONIC_BATTLE, GammaHook_TargetMetalBoss);

local function GammaHook_TargetSpinbobert(mo, aimList, owner)
	// Only aim at the ghost when the flames are targeted.
	if (mo.hnext and mo.hnext.valid)
	and (Gamma.AimListContains(aimList, mo.hnext) == false)
		return false;
	end

	if (mo.hprev and mo.hprev.valid)
	and (Gamma.AimListContains(aimList, mo.hprev) == false)
		return false;
	end

	// Use default logic when all segments are targeted.
	return nil;
end

Gamma.SetTargetFunc(MT_SPINBOBERT, GammaHook_TargetSpinbobert);

local function GammaHook_TargetOtherShot(mo, aimList, owner)
	if (mo.target == owner)
		// Don't target your own shots :V
		return false;
	end

	// Target opposing player's shots
	return true;
end

Gamma.SetTargetFunc(MT_GAMMA_SHOT, GammaHook_TargetOtherShot);

//
// Default damage hooks
//

Gamma.DamageFuncAlways = function(target, shot, source)
	// Always destroy
	P_KillMobj(target, shot, source);
	return Gamma.SHOT_EXPLODE;
end

Gamma.SetDamageFunc(MT_UNIBALL, Gamma.DamageFuncAlways);
Gamma.SetDamageFunc(MT_EGGMOBILE_BALL, Gamma.DamageFuncAlways);
Gamma.SetDamageFunc(MT_POINTYBALL, Gamma.DamageFuncAlways);
Gamma.SetDamageFunc(MT_SPINBOBERT_FIRE1, Gamma.DamageFuncAlways);
Gamma.SetDamageFunc(MT_SPINBOBERT_FIRE2, Gamma.DamageFuncAlways);
Gamma.SetDamageFunc(MT_CANNONBALLDECOR, Gamma.DamageFuncAlways);
Gamma.SetDamageFunc(MT_BOMBSPHERE, Gamma.DamageFuncAlways);
Gamma.SetDamageFunc(MT_GOOP, Gamma.DamageFuncAlways);

local function GammaHook_DamagePlayer(target, shot, source)
	if not (target.player and target.player.valid)
		// This should not be possible.
		return nil;
	end

	if not (target.player.mo and target.player.mo.valid and target == target.player.mo)
		// This should also not be possible.
		return nil;
	end

	if (Gamma.PlayerIsFriendly(target.player, source.player) == true)
		if (target.gammaCoopJump == nil)
			Gamma.DefaultGammaCoopJumpVars(target.player);
		end

		// Enable Coop jump ability.
		target.gammaCoopJump.active = true;
		target.gammaCoopJump.origin = source.player;

		return Gamma.SHOT_EXPLODE;
	end

	if not (target.flags & MF_SHOOTABLE)
		// Not shootable, just reflect.
		return Gamma.SHOT_REFLECT;
	end

	// Ignore flashing tics for this.
	local prevFlashing = target.player.powers[pw_flashing];
	target.player.powers[pw_flashing] = 0;

	if (Gamma.BattleAdded() == true)
		// Battle logic.
		if not (CBW_Battle.PlayerCanBeDamaged(target.player))
			target.player.powers[pw_flashing] = prevFlashing;
			return Gamma.SHOT_REFLECT;
		end

		Gamma.TryBattleDamage(victim, source, shot, false);
		return Gamma.SHOT_EXPLODE;
	end

	if (gametyperules & GTR_RACE)
		// Stun, don't damage.
		P_DoPlayerPain(target.player, source, shot);
		return Gamma.SHOT_EXPLODE;
	else
		// Cause actual damage.
		P_DamageMobj(target, shot, source, 1);
		return Gamma.SHOT_EXPLODE;
	end
end

Gamma.SetDamageFunc(MT_PLAYER, GammaHook_DamagePlayer);

local function GammaHook_DamageBalloon(target, shot, source)
	if (target.state < target.info.raisestate)
		// Popped.
		target.state = target.info.raisestate;
		return Gamma.SHOT_EXPLODE;
	end

	// Didn't pop.
	return Gamma.SHOT_ABSORB;
end

Gamma.SetDamageFunc(MT_BALLOON, GammaHook_DamageBalloon);

local function GammaHook_DamageCrushClaw(target, shot, source)
	// Clean up the chain.
	local chain = target.target;
	while (chain and chain.valid)
		local chainNext = chain.target;
		P_RemoveMobj(chain);
		chain = chainNext;
	end

	// Always destroy.
	P_KillMobj(target, shot, source);
	return Gamma.SHOT_EXPLODE;
end

Gamma.SetDamageFunc(MT_CRUSHCLAW, GammaHook_DamageCrushClaw);
Gamma.SetDamageFunc(MT_BANPSPRING, GammaHook_DamageCrushClaw);

local function GammaHook_DamageBigMine(target, shot, source)
	if not (target.flags & MF_SHOOTABLE)
		// Use default logic if the object is not shootable.
		return nil;
	end

	// Fix some broken values on the DSZ mines that make them go apeshit...
	local a = R_PointToAngle2(0, 0, shot.momx, shot.momy);
	local dist = shot.radius + target.radius + FRACUNIT;

	P_MoveOrigin(
		shot,
		shot.x + FixedMul(dist, cos(a)),
		shot.y + FixedMul(dist, sin(a)),
		shot.z
	);

	target.momx = shot.momx / 3;
	target.momy = shot.momy / 3;
	target.momz = shot.momz / 3;

	// Consider us damaged.
	P_DamageMobj(target, shot, source);
	return Gamma.SHOT_EXPLODE;
end

Gamma.SetDamageFunc(MT_BIGMINE, GammaHook_DamageBigMine);

local function GammaHook_DamageEggShield(target, shot, source)
	// Always destroy
	P_KillMobj(target, shot, source);

	local eggGuard = target.target;
	if (eggGuard and eggGuard.valid)
		// Ensures that they are shootable afterwards

		// Set the guard's target to you, if they didn't have one.
		if not (eggGuard.target and eggGuard.target.valid)
			eggGuard.target = source;
		end

		eggGuard.state = eggGuard.info.seestate;
	end

	return Gamma.SHOT_EXPLODE;
end

Gamma.SetDamageFunc(MT_EGGSHIELD, GammaHook_DamageEggShield);

local function GammaHook_DamageMinus(target, shot, source)
	if not (target.flags & MF_SHOOTABLE)
		// Pop this guy outta the ground
		target.state = S_MINUS_POPUP;
		return Gamma.SHOT_EXPLODE;
	end

	// Use default logic when above ground
	return nil;
end

Gamma.SetDamageFunc(MT_MINUS, GammaHook_DamageMinus);

local function GammaHook_DamageDragonTail(target, shot, source)
	// Fix the segment list
	local prevSeg = target.target;
	local nextSeg = target.tracer;

	if (prevSeg and prevSeg.valid)
		if (nextSeg and nextSeg.valid)
			prevSeg.tracer = nextSeg;
		else
			prevSeg.tracer = nil;
		end
	end

	if (nextSeg and nextSeg.valid)
		if (prevSeg and prevSeg.valid)
			nextSeg.target = prevSeg;
		else
			nextSeg.target = nil;
		end
	end

	// Always destroy
	S_StartSound(target, mobjinfo[MT_DRAGONBOMBER].deathsound);
	P_KillMobj(target, shot, source);
	return Gamma.SHOT_EXPLODE;
end

Gamma.SetDamageFunc(MT_DRAGONTAIL, GammaHook_DamageDragonTail);

local function GammaHook_DamageRolloutRock(target, shot, source)
	if (target.tracer and target.tracer.valid and target.tracer == source)
		// Do not push our own ball.
		return Gamma.SHOT_REFLECT;
	end

	// Use default logic otherwise.
	return nil;
end

Gamma.SetDamageFunc(MT_ROLLOUTROCK, GammaHook_DamageRolloutRock);

local function GammaHook_DamageEggRobo(target, shot, source)
	if (target.state == target.info.deathstate)
		return Gamma.SHOT_REFLECT;
	end

	target.state = target.info.meleestate;
	target.angle = R_PointToAngle2(
		target.x - target.momx, target.y - target.momy,
		shot.x - shot.momx, shot.y - shot.momy
	);
	target.momx = 0;
	target.momy = 0;

	S_StartSound(shot, target.info.attacksound);
	return Gamma.SHOT_REFLECT;
end

Gamma.SetDamageFunc(MT_EGGROBO1, GammaHook_DamageEggRobo);

local function GammaHook_DamageCEZ3Boss(target, shot, source)
	if (target.movefactor == 0 and target.extravalue1 == 0)
		// Prevent damage while the cage is down.
		return Gamma.SHOT_REFLECT;
	end

	// Use default logic in other states.
	return nil;
end

Gamma.SetDamageFunc(MT_EGGMOBILE4, GammaHook_DamageCEZ3Boss);

local function GammaHook_DamageMetalBoss(target, shot, source)
	if (target.state == target.info.spawnstate)
		// Cause the special damage required for Metal Sonic's roaming phase.
		target.flags2 = $1|MF2_CLASSICPUSH;
		return Gamma.SHOT_EXPLODE;
	end

	// Use default logic in other states.
	return nil;
end

Gamma.SetDamageFunc(MT_METALSONIC_BATTLE, GammaHook_DamageMetalBoss);

local function GammaHook_DamageOldbrak(target, shot, source)
	if (target.state >= S_BLACKEGG_WALK1
	and target.state <= S_BLACKEGG_WALK6) // Fuck this boss. Let's kill him
		// This boss is SO FUCKING WEIRD and
		// does damage by just going to pain state
		// instead of calling actual damage functions
		target.state = S_BLACKEGG_PAIN1;
		return Gamma.SHOT_EXPLODE;
	end

	return nil;
end

Gamma.SetDamageFunc(MT_BLACKEGGMAN, GammaHook_DamageOldbrak);

local function GammaHook_DamageATZTarget(target, shot, source)
	if (target.flags & MF_SHOOTABLE)
		// Call the action of the first state
		A_PlaySound(target, sfx_ding, 1);
		target.state = S_TARGET_HIT2;
		target.tics = $1 * 3; // Last longer
		return Gamma.SHOT_EXPLODE;
	end

	// Use default logic in other states.
	return nil;
end

Gamma.SetDamageFunc(MT_TARGET, GammaHook_DamageATZTarget);

local function GammaHook_DamageOtherShot(target, shot, source)
	// Do shot explosion.
	Gamma.ShotExplode(target);
	return Gamma.SHOT_EXPLODE;
end

Gamma.SetDamageFunc(MT_GAMMA_SHOT, GammaHook_DamageOtherShot);

//
// Default combo points hooks
//

local function GammaHook_AlwaysGiveTime(mo)
	// Force this object to count towards combo.
	return 1;
end

Gamma.SetPointsFunc(MT_EGGSHIELD, GammaHook_AlwaysGiveTime);
Gamma.SetPointsFunc(MT_CRUSHCLAW, GammaHook_AlwaysGiveTime);
Gamma.SetPointsFunc(MT_BANPSPRING, GammaHook_AlwaysGiveTime);
Gamma.SetPointsFunc(MT_DRAGONTAIL, GammaHook_AlwaysGiveTime);
Gamma.SetPointsFunc(MT_TNTBARREL, GammaHook_AlwaysGiveTime);

local function GammaHook_NeverGiveTime(mo)
	// This object never counts toward the combo.
	return 0;
end

Gamma.SetPointsFunc(MT_PLAYER, GammaHook_NeverGiveTime);

//
// Never interact with these
//

local function GammaHook_TargetNever(mo, aimList, owner)
	// Never able to target.
	return false;
end

local function GammaHook_DamageNever(target, shot, source)
	// Never able to damage.
	return Gamma.SHOT_ABSORB;
end

Gamma.SetNoInteract = function(t)
	Gamma.SetTargetFunc(t, GammaHook_TargetNever);
	Gamma.SetDamageFunc(t, GammaHook_DamageNever);
end

Gamma.SetNoInteract(MT_MINECART);
Gamma.SetNoInteract(MT_CYBRAKDEMON_ELECTRIC_BARRIER);
Gamma.SetNoInteract(MT_SHOCKWAVE);
Gamma.SetNoInteract(MT_FLAMEJETFLAME);
Gamma.SetNoInteract(MT_FLAMEJETFLAMEB);
