//
// Deep Sea decorations
//

freeslot(
	"S_GARGOYLESHARD",
	"S_BIGGARGOYLESHARD",
	"SPR_GRGD",
	"SPR_GRBD",

	"S_DSZSTALAGMITE_SHARD",

	"S_DSZSTALAGMITE_DEAD",
	"SPR_DSAD",

	"S_DSZ2STALAGMITE_DEAD",
	"SPR_DSBD"
);

states[S_GARGOYLESHARD] = {SPR_GRGD, A, 35, nil, 0, 0, S_NULL}
states[S_BIGGARGOYLESHARD] = {SPR_GRBD, A, 35, nil, 0, 0, S_NULL}

states[S_DSZSTALAGMITE_SHARD] = {SPR_DSAD, B, -1, nil, 0, 0, S_NULL}

states[S_DSZSTALAGMITE_DEAD] = {SPR_DSAD, A, -1, nil, 0, 0, S_NULL}
states[S_DSZ2STALAGMITE_DEAD] = {SPR_DSBD, A, -1, nil, 0, 0, S_NULL}

mobjinfo[MT_WATERDRIP].flags = $1 & ~MF_NOBLOCKMAP;

mobjinfo[MT_DSZSTALAGMITE].deathstate = S_DSZSTALAGMITE_DEAD;
mobjinfo[MT_DSZSTALAGMITE].deathsound = sfx_crumbl;
mobjinfo[MT_DSZSTALAGMITE].gammaTargetSize = mobjinfo[MT_DSZSTALAGMITE].radius * 3;

mobjinfo[MT_DSZ2STALAGMITE].deathstate = S_DSZ2STALAGMITE_DEAD;
mobjinfo[MT_DSZ2STALAGMITE].deathsound = sfx_crumbl;
mobjinfo[MT_DSZ2STALAGMITE].gammaTargetSize = mobjinfo[MT_DSZ2STALAGMITE].radius * 3;

Gamma.SetTargetFunc(MT_GARGOYLE, Gamma.TargetFuncAlways);
Gamma.SetTargetFunc(MT_BIGGARGOYLE, Gamma.TargetFuncAlways);

Gamma.SetTargetFunc(MT_DSZSTALAGMITE, Gamma.TargetFuncAlways);
Gamma.SetTargetFunc(MT_DSZ2STALAGMITE, Gamma.TargetFuncAlways);

local function DamageGargoyle(target, shot, source)
	if (target.flags & MF_PUSHABLE)
		-- Can't blow it up, it's for progression.
		Gamma.ShotPush(target, shot);
		return Gamma.SHOT_EXPLODE;
	end

	local offsetScale = FRACUNIT;
	local st = S_GARGOYLESHARD;
	if (target.type == MT_BIGGARGOYLE)
		offsetScale = $1 << 1;
		st = S_BIGGARGOYLESHARD;
	end

	local shard = nil;
	local tossAngle = nil;
	local sign = nil;

	-- Head
	shard = P_SpawnMobjFromMobj(target, 12*offsetScale, 0, 24*offsetScale, MT_BOSSJUNK);
	shard.state = st;
	shard.angle = target.angle;

	P_SetObjectMomZ(shard, 6*FRACUNIT, false);
	S_StartSound(shard, sfx_s3k96);

	-- Body
	tossAngle = target.angle + ANGLE_180;

	shard = P_SpawnMobjFromMobj(target, 8*offsetScale, 0, 8*offsetScale, MT_BOSSJUNK);
	shard.state = st;
	shard.frame = $1 + 1;
	shard.angle = tossAngle;

	P_InstaThrust(shard, tossAngle, shard.scale);
	P_SetObjectMomZ(shard, 4*FRACUNIT, false);

	-- Wings
	tossAngle = target.angle + ANGLE_135;
	sign = 1;
	for i = 0,1
		shard = P_SpawnMobjFromMobj(target, -8*offsetScale, 12*offsetScale * sign, 8*offsetScale, MT_BOSSJUNK);
		shard.state = st;
		shard.frame = $1 + 2 + i;
		shard.angle = tossAngle;

		P_InstaThrust(shard, tossAngle, 4*shard.scale);
		P_SetObjectMomZ(shard, 4*FRACUNIT, false);

		tossAngle = $1 + ANGLE_90;
		sign = -$1;
	end

	-- Front legs
	tossAngle = target.angle + ANGLE_45;
	sign = 1;
	for i = 0,1
		shard = P_SpawnMobjFromMobj(target, 10*offsetScale, 6*offsetScale * sign, 0, MT_BOSSJUNK);
		shard.state = st;
		shard.frame = $1 + 4 + i;
		shard.angle = tossAngle;

		P_InstaThrust(shard, tossAngle, 4*shard.scale);
		P_SetObjectMomZ(shard, 4*FRACUNIT, false);

		tossAngle = $1 - ANGLE_90;
		sign = -$1;
	end

	-- Back legs
	tossAngle = target.angle + ANGLE_90;
	sign = 1;
	for i = 0,1
		shard = P_SpawnMobjFromMobj(target, 6*offsetScale, 10*offsetScale * sign, 0, MT_BOSSJUNK);
		shard.state = st;
		shard.frame = $1 + 6 + i;
		shard.angle = tossAngle;

		P_InstaThrust(shard, tossAngle, 4*shard.scale);
		P_SetObjectMomZ(shard, 4*FRACUNIT, false);

		tossAngle = $1 + ANGLE_180;
		sign = -$1;
	end

	-- Tail (?)

	P_RemoveMobj(target);
	return Gamma.SHOT_EXPLODE;
end

Gamma.SetDamageFunc(MT_GARGOYLE, DamageGargoyle);
Gamma.SetDamageFunc(MT_BIGGARGOYLE, DamageGargoyle);

local function DamageDSStalagmite(target, shot, source)
	local shardList = {
		{spr = 0, mom = 2, ang = ANGLE_22h},
		{spr = 1, mom = 2, ang = -ANGLE_22h},
		{spr = 2, mom = 3, ang = 0},
		{spr = 3, mom = 4, ang = ANGLE_11hh},
		{spr = 4, mom = 4, ang = -ANGLE_11hh},
		{spr = 5, mom = 5, ang = -ANG1},
		{spr = 6, mom = 6, ang = ANG1},
	};

	local origTossAngle = R_PointToAngle2(shot.x, shot.y, target.x, target.y);

	for i = 1,#shardList
		local shardInfo = shardList[i];
		local tossAngle = origTossAngle + shardInfo.ang;

		local shard = P_SpawnMobjFromMobj(target, 0, 0, 0, MT_BOSSJUNK);
		shard.flags = $1|MF_NOCLIPHEIGHT|MF_NOCLIP|MF_NOCLIPTHING;
		shard.state = S_DSZSTALAGMITE_SHARD;
		shard.frame = $1 + shardInfo.spr;
		shard.angle = tossAngle;

		if (target.type == MT_DSZ2STALAGMITE)
			shard.sprite = SPR_DSBD;
		end

		P_InstaThrust(shard, tossAngle, 4 * shard.scale);
		P_SetObjectMomZ(shard, shardInfo.mom * FRACUNIT, false);

		if (shard.eflags & MFE_VERTICALFLIP)
			-- Just flip the sprite, not the gravity.
			shard.eflags = $1 & ~MFE_VERTICALFLIP;
			shard.flags2 = $1 & ~MF2_OBJECTFLIP;
			shard.renderflags = $1 | RF_VERTICALFLIP;
		end
	end

	A_Scream(target, 0, 0);
	P_KillMobj(target, shot, source);

	-- Find attached water droplets & destroy.
	local drip = Gamma.FindAttachedObjects(target, MT_WATERDRIP);
	for _,v in pairs(drip) do
		if (v and v.valid)
			P_RemoveMobj(v);
		end
	end

	return Gamma.SHOT_EXPLODE;
end

Gamma.SetDamageFunc(MT_DSZSTALAGMITE, DamageDSStalagmite);
Gamma.SetDamageFunc(MT_DSZ2STALAGMITE, DamageDSStalagmite);
