//
// Haunted Heights decor
//

mobjinfo[MT_JACKO1].flags = $1 & ~MF_NOBLOCKMAP;
mobjinfo[MT_JACKO2].flags = $1 & ~MF_NOBLOCKMAP;
mobjinfo[MT_JACKO3].flags = $1 & ~MF_NOBLOCKMAP;

mobjinfo[MT_HHZSTALAGMITE_TALL].flags = $1 & ~MF_NOBLOCKMAP;
mobjinfo[MT_HHZSTALAGMITE_SHORT].flags = $1 & ~MF_NOBLOCKMAP;

freeslot(
	"MT_JACKODEAD",
	"S_JACKODEAD",
	"S_JACKOSHARD",
	"SPR_PUMD",

	"S_HHZSTALAGMITE_TALL_DEAD",
	"S_HHZSTALAGMITE_TALL_SHARD",
	"S_HHZSTALAGMITE_SHORT_DEAD",
	"S_HHZSTALAGMITE_SHORT_SHARD",
	"SPR_HHSD"
);

mobjinfo[MT_JACKODEAD] = {
	spawnstate = S_JACKODEAD,
	spawnhealth = 1000,
	radius = 24*FRACUNIT,
	height = 24*FRACUNIT,
	flags = MF_SCENERY|MF_NOBLOCKMAP
};

states[S_JACKODEAD] = {SPR_PUMD, A, -1, nil, 0, 0, S_NULL};
states[S_JACKOSHARD] = {SPR_PUMD, D, 35, nil, 0, 0, S_NULL};

mobjinfo[MT_JACKO1].gammaTargetSize = 24*FRACUNIT;
mobjinfo[MT_JACKO2].gammaTargetSize = 60*FRACUNIT;
mobjinfo[MT_JACKO3].gammaTargetSize = 36*FRACUNIT;

states[S_HHZSTALAGMITE_TALL_DEAD] = {SPR_HHSD, A, -1, nil, 0, 0, S_NULL};
states[S_HHZSTALAGMITE_TALL_SHARD] = {SPR_HHSD, B, 35, nil, 0, 0, S_NULL};

mobjinfo[MT_HHZSTALAGMITE_TALL].deathstate = S_HHZSTALAGMITE_TALL_DEAD;
mobjinfo[MT_HHZSTALAGMITE_TALL].deathsound = sfx_crumbl;
mobjinfo[MT_HHZSTALAGMITE_TALL].gammaTargetSize = 20*FRACUNIT;

states[S_HHZSTALAGMITE_SHORT_DEAD] = {SPR_HHSD, E, -1, nil, 0, 0, S_NULL};
states[S_HHZSTALAGMITE_SHORT_SHARD] = {SPR_HHSD, F, 35, nil, 0, 0, S_NULL};

mobjinfo[MT_HHZSTALAGMITE_SHORT].deathstate = S_HHZSTALAGMITE_SHORT_DEAD;
mobjinfo[MT_HHZSTALAGMITE_SHORT].deathsound = sfx_crumbl;
mobjinfo[MT_HHZSTALAGMITE_SHORT].gammaTargetSize = 10*FRACUNIT;

Gamma.SetTargetFunc(MT_JACKO1, Gamma.TargetFuncAlways);
Gamma.SetTargetFunc(MT_JACKO2, Gamma.TargetFuncAlways);
Gamma.SetTargetFunc(MT_JACKO3, Gamma.TargetFuncAlways);

Gamma.SetTargetFunc(MT_HHZSTALAGMITE_TALL, Gamma.TargetFuncAlways);
Gamma.SetTargetFunc(MT_HHZSTALAGMITE_SHORT, Gamma.TargetFuncAlways);

local function DamagePumpkin(target, shot, source)
	// Do pumpkin explosion!
	local numShards = 7;
	local angRang = (360 / numShards);
	local numSprites = 5;
	for i = 1,7
		local shardAngle = angRang * (i-1) * ANG1;
		shardAngle = $1 + (P_RandomRange(0, angRang-1) * ANG1);

		local shard = P_SpawnMobjFromMobj(target, 0, 0, 0, MT_BOSSJUNK);
		shard.state = S_JACKOSHARD;
		shard.frame = $1 + P_RandomKey(numSprites);

		P_InstaThrust(shard, shardAngle, P_RandomRange(4, 6) * shard.scale);
		P_SetObjectMomZ(shard, P_RandomRange(6, 10) * FRACUNIT, false);
	end

	// We have to remove it to get rid of the overlay,
	// but we want a death state, so extra dummy object it is.
	local corpse = P_SpawnMobjFromMobj(target, 0, 0, 0, MT_JACKODEAD);
	corpse.frame = $1 + (target.type - MT_JACKO1);
	A_PlaySound(corpse, sfx_pumpkn, 1);

	P_RemoveMobj(target);
	return Gamma.SHOT_EXPLODE;
end

Gamma.SetDamageFunc(MT_JACKO1, DamagePumpkin);
Gamma.SetDamageFunc(MT_JACKO2, DamagePumpkin);
Gamma.SetDamageFunc(MT_JACKO3, DamagePumpkin);

local function DamageHHZStalagmite(target, shot, source)
	local shardList = {
		{spr = 0, mom = 3, ang = ANGLE_22h},
		{spr = 1, mom = 3, ang = -ANGLE_22h},
		{spr = 2, mom = 5, ang = 0},
	};

	local origTossAngle = R_PointToAngle2(shot.x, shot.y, target.x, target.y);
	local st = S_HHZSTALAGMITE_TALL_SHARD;
	if (target.type == MT_HHZSTALAGMITE_SHORT)
		st = S_HHZSTALAGMITE_SHORT_SHARD;
	end

	for i = 1,#shardList
		local shardInfo = shardList[i];
		local tossAngle = origTossAngle + shardInfo.ang;

		local shard = P_SpawnMobjFromMobj(target, 0, 0, 0, MT_BOSSJUNK);
		shard.state = st;
		shard.frame = $1 + shardInfo.spr;
		shard.angle = tossAngle;

		P_InstaThrust(shard, tossAngle, 4 * shard.scale);
		P_SetObjectMomZ(shard, shardInfo.mom * FRACUNIT, false);

		if (shard.eflags & MFE_VERTICALFLIP)
			-- Just flip the sprite, not the gravity.
			shard.eflags = $1 & ~MFE_VERTICALFLIP;
			shard.flags2 = $1 & ~MF2_OBJECTFLIP;
			shard.renderflags = $1 | RF_VERTICALFLIP;
		end
	end

	A_Scream(target, 0, 0);
	P_KillMobj(target, shot, source);

	return Gamma.SHOT_EXPLODE;
end

Gamma.SetDamageFunc(MT_HHZSTALAGMITE_TALL, DamageHHZStalagmite);
Gamma.SetDamageFunc(MT_HHZSTALAGMITE_SHORT, DamageHHZStalagmite);
