//
// Gamma tether support
//

Gamma.TetherAdded = function()
	if (Tether != nil)
		return true;
	end

	return false;
end

Gamma.TetherTimeSharing = function()
	if (Gamma.TetherAdded() != true)
		return;
	end

	local teamList = Tether.Teams;
	local numTeams = #teamList;

	for i = 1,numTeams
		local team = teamList[i];

		if (Tether.TeamIsValid(team) == true)
			local teamPlayers = team.playerList;
			local numPlayers = #teamPlayers;

			local anyGammas = false;
			local timeWillBeAdded = false;
			local highestTime = INT32_MIN;
			local lowestPause = INT32_MAX;

			for j = 1,numPlayers
				local player = teamPlayers[j];
				if not ((player and player.valid)
				and (player.mo and player.mo.valid))
					continue;
				end

				if (Tether.PlayerReady(player) == true)
					if (player.gammaTime != nil)
						highestTime = max($1, player.gammaTime);
					end

					if (player.gammaTimePause != nil)
						lowestPause = min($1, player.gammaTimePause);
					end
				end

				if (player.mo.skin == "gamma")
					anyGammas = true;

					if (player.mo.gammaVars != nil)
					and (timeWillBeAdded == false)
						-- Check if we're adding time to anyone before deciding to kill them.
						if (player.mo.gammaVars.addSeconds > 0)
							timeWillBeAdded = true;
						end

						for _,combo in pairs(player.mo.gammaVars.comboList)
							if (combo.addTime > 0)
								timeWillBeAdded = true;
							end
						end

						for _,coopCombo in pairs(player.mo.gammaVars.coopCombos)
							if (coopCombo.addTime > 0)
								timeWillBeAdded = true;
							end
						end
					end
				end
			end

			if (anyGammas == false)
				-- No Gamma on this team, don't do anything.
				continue;
			end

			if (highestTime < 0)
				-- No valid timer.
				continue;
			end

			-- Run Gamma's timer logic for all of these players.
			for j = 1,numPlayers
				local player = teamPlayers[j];
				if not (player and player.valid)
				or not (player.mo and player.mo.valid)
					continue;
				end

				player.gammaTime = highestTime;
				if (player.gammaTimePause == nil)
					player.gammaTimePause = lowestPause;
				end

				Gamma.RunTimer(player, timeWillBeAdded);
			end
		end
	end
end

Gamma.TetherTimeScale = function(secs, player)
	if (Gamma.TetherAdded() != true)
		// No Tether, normal time add
		return secs;
	end

	if not (player and player.valid)
		// No player, normal time add
		return secs;
	end

	if (player.mo and player.mo.valid and player.mo.skin == "gamma")
		// Decided not to do this for
		// Gamma players after all
		return secs;
	end

	local team = nil;
	if (Gamma.TetherAdded() == true)
	and (player.tetherPersist != nil)
		local teamID = player.tetherPersist.teamID;
		team = Tether.Teams[teamID];
	end

	if (Tether.TeamIsValid(team) != true)
		// No team, normal time add
		return secs;
	end

	// Scale time bonuses with the number
	// of players on your team.
	local teamPlayers = team.playerList;
	local numPlayers = #teamPlayers;

	if (numPlayers <= 1)
		// Empty team, normal time add
		return secs;
	end

	// Scale time bonuses up with number of players,
	// to make time for Tether's hijinks
	local start = FRACUNIT >> 14;
	local log = Gamma.Log2((numPlayers * FRACUNIT) + start);

	if (log <= FRACUNIT)
		// This shouldn't happen.
		return secs;
	end

	local newSecs = FixedMul(secs * FRACUNIT, log);

	if (newSecs > 8*FRACUNIT + FRACUNIT/2)
		// Round to nearest 5sec after a certain
		// threshold, for nicer looking numbers.
		newSecs = FixedDiv($1, 5*FRACUNIT) * 5;
	end

	return Gamma.RoundedInt(newSecs);
end

Gamma.TetherDiscardMobj = function(mobj, originPlayer)
	if (Gamma.TetherAdded() != true)
		// No Tether, so everyone is fending for themselves.
		// We can ignore this.
		return;
	end

	if not (mobj and mobj.valid)
		// The object we're discarded is, well,
		// already discarded!
		return;
	end

	if not (originPlayer and originPlayer.valid)
		// No origin player, so there's no team to discard from.
		return;
	end

	local team = nil;
	if (Gamma.TetherAdded() == true)
	and (originPlayer.tetherPersist != nil)
		local teamID = originPlayer.tetherPersist.teamID;
		team = Tether.Teams[teamID];
	end

	if (Tether.TeamIsValid(team) != true)
		// No team to discard from.
		return;
	end

	local teamPlayers = team.playerList;
	local numPlayers = #teamPlayers;

	if (numPlayers <= 1)
		// Empty team, no need to waste our time.
		return;
	end

	// If there's any Gamma players who are trying to
	// also destroy this object, remove it from their combo,
	// so you can't double-up on time add from the leniency
	// mechanic that was only intended for SP & "selfish" Coop.
	for playerID = 1,numPlayers do
		local player = teamPlayers[playerID];

		if not ((player and player.valid)
		and (player.mo and player.mo.valid))
			// Invalid right now.
			continue;
		end

		if (player == originPlayer)
			// Don't remove from our own stuff
			// (if it needs to happen, it'll happen later)
			continue;
		end

		if (player.mo.skin != "gamma")
			// There aren't any combos to delete off a non-Gamma :P
			continue;
		end

		if (player.mo.gammaVars == nil)
			// This Gamma's not ready yet
			continue;
		end

		for k,combo in pairs(player.mo.gammaVars.comboList) do
			// Gather aim list objects to remove
			local aimRemove = {};
			for k,t in pairs(combo.aimList) do
				if (t and t.mobj and t.mobj.valid and mobj == t.mobj)
					table.insert(aimRemove, k);
					if (t.points > 0)
						combo.numWithPoints = $1 - 1;
					end
				end
			end

			// Fix our order list to account for what we're removing
			for _,t in pairs(combo.aimList) do
				for i = 1,#aimRemove do
					local removedOrder = combo.aimList[ aimRemove[i] ].order;
					if (removedOrder > 0 and t.order >= removedOrder)
						t.order = $1 - 1;
					end
				end
			end

			// Now we can remove
			for i = 1,#aimRemove do
				table.remove(combo.aimList, aimRemove[i]);
			end

			// Gather shot list objects to remove
			local shotRemove = {};
			for k,t in pairs(combo.shotList) do
				if (t and t.mobj and t.mobj.valid and mobj == t.mobj)
					if (t.bullets != nil)
						for _,bullet in pairs(t.bullets) do
							if (bullet and bullet.valid)
								bullet.tracer = nil; // Don't keep the reference to this object
								bullet.gammaComboPoints = 0; // Removed from combo
								Gamma.ShotFizzle(bullet); // Fizzle away immediately
							end
						end
					end

					table.insert(shotRemove, k);
				end
			end

			// Now we can remove
			for i = 1,#shotRemove do
				table.remove(combo.shotList, shotRemove[i]);
			end
		end
	end
end
