//
// Combo stuff
//

Gamma.ComboToSeconds = function(numGotten)
	return min(max(numGotten, 1), 5);
end

Gamma.HealthToPoints = function(hp)
	local t = hp;

	if (t < 2)
		// 1 HP enemies are pretty normal ... just return 1.
		return max(0, t);
	end

	// Big enemies / bosses need a more
	// sophisticated calculation.

	// Adjust for time spent in MF2_FRET.
	// 1 second is the longest a vanilla enemy/boss
	// spends in fret, so that makes it easy.
	t = $1 + (hp - 1);

	if (t > 8)
		// Round to nearest 5sec after a certain
		// threshold, for nicer looking numbers.
		t = Gamma.RoundedInt(FixedDiv($1*FRACUNIT, 5*FRACUNIT)) * 5;
	end

	return t;
end

Gamma.ComboPoints = function(mo)
	local hookVal = Gamma.GetPointsValue(mo);

	if (hookVal != nil)
		return hookVal;
	end

	if (mo.flags & (MF_ENEMY|MF_BOSS))
	and (mo.health <= 1)
		// Combo points depends on health
		return Gamma.HealthToPoints(mo.info.spawnhealth);
	end

	// Does not add to combo
	return 0;
end

//
// Per-player combos
//

Gamma.GiveTime = function(points, combo, player)
	if (points == nil or points <= 0)
		return;
	end

	// Killed an enemy? Add some time!
	combo.numGotten = $1 + 1;

	local sec = Gamma.ComboToSeconds(combo.numGotten) * points;
	sec = Gamma.TetherTimeScale($1, player);

	if (Gamma.TimerRunning(player) == true)
		if (combo.addTime <= 0)
			combo.startTime = leveltime;
			combo.offset = Gamma.ComboOffset(player.mo.gammaVars.comboList, combo.startTime);
		end

		combo.addTime = $1 + sec;
		combo.waughmp = Gamma.WAUGHMPTIME * 2;
	end

	Gamma.GiveCoopTime(sec, player);
end

Gamma.GiveTimeFromMobj = function(mobj, points, combo, player)
	Gamma.GiveTime(points, combo, player);
	Gamma.TetherDiscardMobj(mobj, player);
end

Gamma.GiveArbitraryTime = function(seconds, player, noCoop)
	-- Give time without an existing combo
	-- being necessary.

	if (seconds == nil or seconds <= 0)
		return;
	end

	seconds = Gamma.TetherTimeScale($1, player);

	if (Gamma.TimerRunning(player) == true)
		if (player.mo and player.mo.valid)
		and (player.mo.skin == "gamma")
		and (player.mo.gammaVars != nil)
			-- Create a fake combo, so it shows as
			-- a time bubble on the HUD.
			local i = 1;
			while (player.mo.gammaVars.comboList[i] != nil)
				i = $1+1;
			end

			local combo = Gamma.NewCombo();
			combo.addTime = seconds;
			combo.fired = true;
			player.mo.gammaVars.comboList[i] = combo;
		end
	end

	if not (noCoop)
		-- Give everyone else time too.
		Gamma.GiveCoopTime(seconds, player);
	end
end

Gamma.ComboOffset = function(combos, ourTime)
	if (ourTime <= 0)
		return 0;
	end

	local offset = 0;

	for k,v in pairs(combos)
		if (v == nil)
			continue;
		end

		if (v.cashIn <= Gamma.CASHININTERP)
			continue;
		end

		if (v.startTime <= 0)
			continue;
		end

		local theirTime = v.startTime
		if (ourTime > theirTime)
			offset = $1 + 40;
		end
	end

	return offset;
end

//
// Coop combos
//

Gamma.InCoop = function()
	return (
		(multiplayer or netgame)
		and (gametyperules & GTR_FRIENDLY)
	);
end

Gamma.GiveCoopTime = function(seconds, player)
	if (seconds == nil or seconds <= 0)
		return;
	end

	if (Gamma.InCoop() == false)
		return;
	end

	if (Gamma.CoopTimeTotals[#player] == nil)
		Gamma.CoopTimeTotals[#player] = 0;
	end

	Gamma.CoopTimeTotals[#player] = $1 + seconds;

	for p in players.iterate do
		if (p == player)
			continue;
		end

		if (p.gammaTime == nil)
			continue;
		end

		local pmo = p.mo;
		if not (pmo and pmo.valid)
			continue;
		end

		if (pmo.gammaVars == nil)
			continue;
		end

		local id = nil;

		for i = 1,#(pmo.gammaVars.coopCombos) do
			local v = pmo.gammaVars.coopCombos[i];

			if (v == nil)
			or (v.player == player)
				id = i;
				break;
			end
		end

		if (id == nil)
			pmo.gammaVars.numCoopCombos = $1 + 1;
			id = pmo.gammaVars.numCoopCombos;
		end

		if (pmo.gammaVars.coopCombos[id] == nil)
			pmo.gammaVars.coopCombos[id] = {};
			pmo.gammaVars.coopCombos[id].addTime = 0;
			pmo.gammaVars.coopCombos[id].player = player;
		end

		pmo.gammaVars.coopCombos[id].addTime = $1 + seconds;
		pmo.gammaVars.coopCombos[id].waughmp = Gamma.WAUGHMPTIME*2;
		pmo.gammaVars.coopCombos[id].cashIn = Gamma.CASHINTIME;
		pmo.gammaVars.coopCombos[id].offset = Gamma.CoopComboOffset(pmo.gammaVars.coopCombos, id);
	end
end

Gamma.GiveCoopTimeFromMobj = function(mobj, seconds, player)
	Gamma.GiveCoopTime(seconds, player);
	Gamma.TetherDiscardMobj(mobj);
end

Gamma.CoopComboOffset = function(coopCombos, id)
	local offset = 0;

	for k,v in pairs(coopCombos) do
		if (v == nil)
			continue;
		end

		if (k == id)
			break;
		end

		offset = $1 + 32;
	end

	return offset;
end

//
// Map settings
//

Gamma.PreDefinedTime = function(map)
	// Pre-determined time limits
	// Make the early game feel like Emerald Coast / Windy Valley
	// Make the end game / bonus stages feel like Hot Shelter

	local levelName = string.upper(mapheaderinfo[map].lvlttl);
	local act = mapheaderinfo[map].actnum;

	if (levelName == "GREENFLOWER")
		if (act == 1)
			return 180*TICRATE;
		elseif (act == 2)
			return 180*TICRATE;
		elseif (act == 3)
			return 60*TICRATE;
		end
	elseif (levelName == "TECHNO HILL")
		if (act == 1)
			return 180*TICRATE;
		elseif (act == 2)
			return 180*TICRATE;
		elseif (act == 3)
			return 60*TICRATE;
		end
	elseif (levelName == "DEEP SEA")
		if (act == 1)
			return 180*TICRATE;
		elseif (act == 2)
			return 210*TICRATE;
		elseif (act == 3)
			return 60*TICRATE;
		end
	elseif (levelName == "CASTLE EGGMAN")
		if (act == 1)
			return 180*TICRATE;
		elseif (act == 2)
			return 180*TICRATE;
		elseif (act == 3)
			return 120*TICRATE;
		end
	elseif (levelName == "ARID CANYON")
		if (act == 1)
			return 150*TICRATE;
		elseif (act == 2)
			return 120*TICRATE;
		elseif (act == 3)
			return 60*TICRATE;
		end
	elseif (levelName == "RED VOLCANO")
		if (act == 1)
			return 120*TICRATE;
		end
	elseif (levelName == "EGG ROCK")
		if (act == 1)
			return 150*TICRATE;
		elseif (act == 2)
			return 270*TICRATE;
		end
	elseif (levelName == "BLACK CORE")
		if (act == 1)
			return 90*TICRATE;
		elseif (act == 2)
			return 120*TICRATE;
		elseif (act == 3)
			return 120*TICRATE;
		end
	elseif (levelName == "FROZEN HILLSIDE")
		if (act == 0)
			return 120*TICRATE;
		end
	elseif (levelName == "PIPE TOWERS")
		if (act == 0)
			return 180*TICRATE;
		end
	elseif (levelName == "FOREST FORTRESS")
		if (act == 0)
			return 180*TICRATE;
		end
	elseif (levelName == "HAUNTED HEIGHTS")
		if (act == 0)
			return 180*TICRATE;
		end
	-- Don't care beyond this point :zzz:
	elseif (levelName == "AERIAL GARDEN")
		if (act == 0)
			return 600*TICRATE;
		end
	elseif (levelName == "AZURE TEMPLE")
		if (act == 0)
			return 600*TICRATE;
		end
	elseif (levelName == "TECHNO LEGACY")
		if (act == 0)
			return 180*TICRATE;
		end
	end

	return nil;
end

// Since hardcode doesn't define these.
rawset(_G, "BONUSTYPE_NONE", -1);
rawset(_G, "BONUSTYPE_NORMAL", 0);
rawset(_G, "BONUSTYPE_BOSS", 1);
rawset(_G, "BONUSTYPE_ERZ3", 2);
rawset(_G, "BONUSTYPE_NIGHTS", 3);
rawset(_G, "BONUSTYPE_NIGHTSLINK", 4);

Gamma.MinimumDefaultTime = function(map)
	local bt = mapheaderinfo[map].bonustype;
	local minute = (60 * TICRATE);

	if (bt == BONUSTYPE_BOSS)
		// Default to 1 minutes for boss maps.
		return minute;
	elseif (bt == BONUSTYPE_ERZ3)
		// This is a map with a boss at the end.
		// Match SA1's timer of 3 minutes.
		return minute*3;
	end

	// Use average length timer for standard stages.
	return minute*2;
end

Gamma.HubTitleCheck = function(map)
	local title = string.upper(mapheaderinfo[map].lvlttl);
	local subtitle = string.upper(mapheaderinfo[map].subttl);

	return (
		string.find(title, "HUB") != nil
		or string.find(subtitle, "HUB") != nil
	);
end

Gamma.AutoEnemyPoints = function(mobjType, info)
	// Points func for the automatic time limit.

	// No hook for this, because if you're
	// customizing a map for Gamma, you should
	// just set a real freakin' time limit :P

	if (mobjType == MT_CRUSHSTACEAN)
	or (mobjType == MT_BANPYURA)
		return 3;
	elseif (mobjType == MT_EGGGUARD)
		return 2;
	elseif (mobjType == MT_DRAGONBOMBER)
		return 25;
	elseif (mobjType == MT_TNTBARREL)
		return 1;
	end

	if (info.flags & MF_ENEMY)
		if (info.spawnhealth >= 1000)
			// Likely intended to be invincible
			return 0;
		end

		return Gamma.HealthToPoints(info.spawnhealth);
	end

	return 0;
end

Gamma.DefaultTime = function(map)
	if not (gametyperules & GTR_ALLOWEXIT)
		-- Likely in a ringslinger/Battle-type mode, not a Coop/Race-type mode,
		-- so a time limit will just ruin things.
		return 0;
	end

	if (mapheaderinfo[map] == nil)
		-- Invalid map?
		return 0;
	end

	if (G_IsSpecialStage(map)) or (tutorialmode)
		-- Special modes, don't use a time limit.
		return 0;
	end

	local mapHeaderTimeLimit = tonumber(mapheaderinfo[map].gammatimelimit);
	if (mapHeaderTimeLimit != nil and mapHeaderTimeLimit > 0)
		// Use map's predefined Gamma time limit
		return (mapHeaderTimeLimit * TICRATE);
	elseif (mapheaderinfo[map].countdown > 0)
		// Use the map's original time limit.
		return (mapheaderinfo[map].countdown * TICRATE);
	elseif (mapheaderinfo[map].typeoflevel & TOL_NIGHTS)
		// No time limit in NiGHTS
		return 0;
	end

	local bonus = mapheaderinfo[map].bonustype;
	if (bonus == BONUSTYPE_NONE)
		// Probably not a real level (maybe a hub), so no time limit
		return 0;
	end

	local predefined = Gamma.PreDefinedTime(map);
	if (predefined != nil)
		// Use a predefined time for the vanilla stages.
		return predefined;
	end

	if (Gamma.HubTitleCheck(map) == true)
		// Probably a hub, since it's called a hub.
		return 0;
	end

	// Calculate based on number of starposts in the map.
	local starpostTimeStart = (60 * TICRATE); // Timer start point
	local starpostTimeAdd = (30 * TICRATE); // How much to add for each starpost

	local startPoint = nil;
	local endPoint = nil;
	local starposts = {};

	local enemies = 0;

	local bosses = 0;
	for mt in mapthings.iterate do
		if not (mt and mt.valid)
			continue;
		end

		if (mt.type == 1) // player 1 start
			startPoint = Vector2.New(mt.x, mt.y);
			continue;
		end

		if (mt.type == mobjinfo[MT_SIGN].doomednum)
			endPoint = Vector2.New(mt.x, mt.y);
			continue;
		end

		if (mt.type == mobjinfo[MT_STARPOST].doomednum)
			local thisStarpostNum = mt.args[0] + 1;

			local data = nil;
			for i = 1,#starposts do
				if (starposts[i].id == thisStarpostNum)
					data = starposts[i];
					break;
				end
			end

			if (data == nil)
				table.insert(starposts, {
					id = thisStarpostNum,
					count = 0,
					point = Vector2.New(0, 0),
				});
				data = starposts[#starposts];
			end

			data.count = $1 + 1;
			data.point.x = $1 + mt.x;
			data.point.y = $1 + mt.y;
			continue;
		end

		local info = nil;
		local mobjType = MT_NULL;
		for i = 0,#mobjinfo-1
			if (mt.type == mobjinfo[i].doomednum)
				mobjType = i;
				info = mobjinfo[i];
				break;
			end
		end

		if (info != nil)
			if (info.flags & MF_BOSS)
				bosses = $1 + 1;
			else
				enemies = $1 + Gamma.AutoEnemyPoints(mobjType, info);
			end
		end

		if (mt.type == mobjinfo[MT_EGGMOBILE4].doomednum)
		or (mt.type == mobjinfo[MT_METALSONIC_RACE].doomednum)
		or (mt.type == mobjinfo[MT_METALSONIC_BATTLE].doomednum)
		or (mt.type == mobjinfo[MT_CYBRAKDEMON].doomednum)
		or (mt.type == mobjinfo[MT_BLACKEGGMAN].doomednum)
			// All of the later bosses are obnoxiously long, count them as 3 bosses
			bosses = $1 + 2;
		end
	end

	if (enemies == 0 and bosses == 0)
		// NOTHING to get time off of.
		// Just have no time limit.
		return 0;
	end

	table.sort(starposts, function(a, b)
		local aID = INT32_MIN;
		local bID = INT32_MIN;
		if (a != nil)
			aID = a.id;
		end
		if (b != nil)
			bID = b.id;
		end
		return (aID < bID);
	end);

	for i = 1,#starposts do
		starposts[i].point.x = $1 / starposts[i].count;
		starposts[i].point.y = $1 / starposts[i].count;
	end

	local avgDist = 0;
	local distEntries = 0;
	for i = 1,#starposts+1 do
		local s = i - 1;
		local e = i;
		local sP = nil;
		local eP = nil;

		if (s == 0)
			sP = startPoint;
		else
			sP = starposts[s].point;
		end

		if (e == #starposts+1)
			eP = endPoint;
		else
			eP = starposts[e].point;
		end

		if (sP != nil and eP != nil)
			avgDist = $1 + R_PointToDist2(sP.x, sP.y, eP.x, eP.y);
			distEntries = $1 + 1;
		end
	end

	local distMul = FRACUNIT;
	if (distEntries > 0)
		avgDist = $1 / distEntries;

		local expectedDist = 7168;
		local factor = FRACUNIT/2;

		/*
		print(
			string.format("AVERAGE STARPOST DIST: %d", avgDist)
		);
		*/

		distMul = FixedDiv(avgDist, expectedDist);
		distMul = FixedMul($1 - FRACUNIT, factor) + FRACUNIT;
		distMul = min(max($1, FRACUNIT/2), FRACUNIT*2);
	end

	local starpostMax = 0;
	for i = 1,#starposts do
		local starpostValue = FRACUNIT;
		for i = 2,starposts[i].count do
			starpostValue = $1 - ($1 / 3);
		end
		starpostValue = $1 * starposts[i].count;

		//print(string.format("starpost value [%d]: %f", starposts[i].id, starpostValue));

		starpostMax = $1 + starpostValue;
	end

	local enemyMul = FRACUNIT;
	if (bonus != BONUSTYPE_BOSS)
		local expectedEnemyDensity = 24 * FRACUNIT;
		local enemiesPerStarpost = max(FixedDiv(enemies * FRACUNIT, starpostMax + FRACUNIT), FRACUNIT);
		local factor = FRACUNIT/4;

		/*
		print(
			string.format("ENEMIES: %d", enemies),
			string.format("STARPOSTS: %f", starpostMax),
			string.format("ENEMY DENSITY: %f", enemiesPerStarpost)
		);
		*/

		// Scale the timer based on the enemy density
		enemyMul = FixedDiv(expectedEnemyDensity, enemiesPerStarpost);
		enemyMul = FixedMul($1 - FRACUNIT, factor) + FRACUNIT;
		enemyMul = min(max($1, FRACUNIT/2), FRACUNIT*2);
	end

	/*
	print(
		string.format("* ENEMY MULTIPLIER: %f", enemyMul),
		string.format("* DISTANCE MULTIPLIER: %f", distMul)
	);
	*/

	if (enemyMul != FRACUNIT or distMul != FRACUNIT)
		starpostMax = FixedMul(starpostMax, FixedMul(enemyMul, distMul));
	end

	// Each boss is considered a starpost
	starpostMax = $1 + (max(0, bosses) * FRACUNIT);
	starpostMax = Gamma.RoundedInt($1);

	local finalTime = starpostTimeStart + (starpostTimeAdd * starpostMax);
	return min(max(finalTime, Gamma.MinimumDefaultTime(map)), 600*TICRATE);
end

Gamma.MidMapTime = function(player)
	if (Gamma.TimeAttackRules())
		-- Time Attack has different behavior.
		return nil;
	end

	local time = Gamma.MapTimeLimit;

	if (Gamma.MapTimeLimit == 0)
		return nil;
	end

	if (Gamma.InCoop() == true)
		local playerID = nil;
		if (player and player.valid)
			playerID = #player;
		end

		for k,v in pairs(Gamma.CoopTimeTotals) do
			if (k == playerID)
				continue;
			end

			time = $1 + (v * TICRATE);
		end
	end

	return time;
end

Gamma.InitPlayerTime = function(player)
	if (Gamma.TimeAttackRules())
		-- Time Attack has different behavior.
		player.gammaTime = nil;
		player.gammaTimePause = 0;
		return;
	end

	player.gammaTime = Gamma.MidMapTime(player);

	if (player.gammaTime != nil)
		player.gammaTimePause = Gamma.LEVELSTARTPAUSE;
	end
end

Gamma.TimerRunning = function(player)
	if (gametyperules & GTR_RACE)
	and (leveltime < 4*TICRATE)
		return false;
	end

	if (stoppedclock)
		return false;
	end

	if (player and player.valid)
		if (player.exiting > 0)
		or (player.pflags & PF_FINISHED)
		or (player.pflags & PF_GAMETYPEOVER)
			return false;
		end
	end

	return true;
end

Gamma.RunTimer = function(player, timeWillBeAdded)
	if (player.gammaTimerRan)
		-- Timer could run more than once
		-- in Tether without this check.
		return;
	end

	player.gammaTimerRan = true;

	local isGamma = false;
	if ((player.mo.skin == "gamma") and (player.mo.gammaVars != nil))
		isGamma = true;
	end

	-- This is always ran for Time Bonus purposes.
	Gamma.RunTimeAttackTimer(player);

	if (isGamma == true)
		if (player.mo.gammaVars.addingPause > 0)
			player.mo.gammaVars.addingPause = $1 - 1;
		else
			if (player.mo.gammaVars.addSeconds > 0)
				player.mo.gammaVars.addSeconds = $1 - 1;

				if (player.gammaTime != nil)
					player.gammaTime = $1 + TICRATE;
				end

				local schmeconds = TICRATE / 5; // since full seconds are too much
				player.gammaTimeAttackPause = $1 + schmeconds;

				player.mo.gammaVars.addingPause = 1;
				S_StartSound(player.mo, sfx_gmtick, player);
			end
		end

		if (player.gammaTime != nil)
		and (player.gammaTime < 30*TICRATE)
			if (player.mo.gammaVars.didHurry == false)
				local snd = sfx_gmhry1 + P_RandomKey(2);
				S_StartSound(player.mo, snd);
				player.mo.gammaVars.didHurry = true;
			end

			if (player.mo.gammaVars.prevStarpost != nil)
			and (player.starpostnum > player.mo.gammaVars.prevStarpost)
				// Clutch bonus!
				Gamma.GiveArbitraryTime(30, player, true);
			end
		end

		player.mo.gammaVars.prevStarpost = player.starpostnum;
	end

	if (player.gammaTime != nil) // Stage has time limit?
		if (Gamma.TimerRunning(player))
			Gamma.TimeOutSparks(player);

			if (player.gammaTime <= 0)
			and (timeWillBeAdded != true)
				P_KillMobj(player.mo);
			elseif (player.gammaTime > 0)
			and (player.gammaTimePause == 0)
				player.gammaTime = $1 - 1;
			end
		end

		if (player.gammaTimePause > 0)
			player.gammaTimePause = $1 - 1;
		end
	end
end

Gamma.TimeAttackRules = function()
	return (
		modeattacking
		//or marathonmode
		or (gametyperules & GTR_RACE)
	);
end

Gamma.RunTimeAttackTimer = function(player)
	if (player.gammaTimeAttack == nil)
		player.gammaTimeAttack = player.realtime;
	end

	if (player.gammaTimeAttackPause == nil)
		player.gammaTimeAttackPause = 0;
	end

	if (Gamma.TimerRunning(player))
		if (player.gammaTimeAttackPause > 0)
			player.gammaTimeAttackPause = $1 - 1;
		else
			player.gammaTimeAttack = $1 + 1;
		end
	else
		player.gammaTimeAttackPause = 0; -- Stop HUD blinking
	end

	player.realtime = player.gammaTimeAttack;
end
